Package: gedi2
Type: Package
Title: Gene Expression Decomposition and Integration
Version: 2.3.4
Date: 2026-05-09
Authors@R: c(
    person("Arsham", "Mikaeili Namini", email = "arsham.mikaeilinamini@mail.mcgill.ca", role = c("aut", "cre")),
    person("Hamed", "S.Najafabadi", email = "hamed.najafabadi@mcgill.ca", role = c("aut"))
    )
Description: A memory-efficient implementation for integrating gene expression data
    from single-cell RNA sequencing experiments. Uses a C++ backend with
    thin R wrappers to enable analysis of large-scale single-cell datasets. The
    package supports multiple data modalities including count matrices, paired
    data (splicing, RNA velocity, CITE-seq), and binary indicators. It implements
    a latent variable model with block coordinate descent optimization for
    dimensionality reduction and batch effect correction. Core algorithms are described in Madrigal et al. (2024) <doi:10.1038/s41467-024-50963-0>.
License: MIT + file LICENSE
URL: https://github.com/csglab/gedi2
BugReports: https://github.com/csglab/gedi2/issues
Depends: R (>= 4.0.0)
Imports: Rcpp (>= 1.0.0), R6 (>= 2.5.0), Matrix (>= 1.3.0), ggplot2,
        scales, methods, stats, utils
LinkingTo: Rcpp, RcppEigen
Suggests: hdf5r, uwot, digest, glmnet, Seurat, SeuratObject,
        SingleCellExperiment, testthat (>= 3.0.0)
Config/testthat/edition: 3
SystemRequirements: GNU make
Encoding: UTF-8
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-05-13 17:32:10 UTC; arsham79
Author: Arsham Mikaeili Namini [aut, cre],
  Hamed S.Najafabadi [aut]
Maintainer: Arsham Mikaeili Namini <arsham.mikaeilinamini@mail.mcgill.ca>
Repository: CRAN
Date/Publication: 2026-05-19 07:30:21 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-19 09:49:06 UTC; unix
Archs: gedi2.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(print,gedi_dynamics)
S3method(print,gedi_dynamics_svd)
S3method(print,gedi_imputation)
S3method(print,gedi_pathway_associations)
export(CreateGEDIObject)
export(check_optional_dependencies)
export(gedi_to_seurat)
export(install_optional_dependencies)
export(list_h5_structure)
export(plot_convergence)
export(plot_dispersion)
export(plot_embedding)
export(plot_feature_ratio)
export(plot_features)
export(plot_vector_field)
export(read_h5)
export(read_h5ad)
export(seurat_to_gedi)
export(write_h5ad)
import(Matrix)
import(R6)
importFrom(Matrix,Matrix)
importFrom(Matrix,sparseMatrix)
importFrom(Matrix,t)
importFrom(Rcpp,sourceCpp)
importFrom(methods,as)
importFrom(methods,is)
importFrom(stats,coef)
importFrom(stats,median)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,var)
importFrom(utils,object.size)
importFrom(utils,packageVersion)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)
useDynLib(gedi2, .registration = TRUE)

#' @importFrom utils packageVersion

.onAttach <- function(libname, pkgname) {
  # Get package version from DESCRIPTION
  version <- utils::packageVersion(pkgname)
  
  packageStartupMessage(
    "\n",
    "=======================================================================================================\n",
    "  Welcome to GEDI v", version, " - Developed by the Computational and Statistical Genomics Laboratory\n",
    "  at McGill University, Montreal. Distributed under the MIT License.\n",
    "  For documentation and examples, see the package vignette.\n",
    "-------------------------------------------------------------------------------------------------------\n",
    "  Bienvenue dans GEDI v", version, " - Developpe par le Laboratoire de genomique computationnelle\n",
    "  et statistique de l'Universite McGill, a Montreal. Distribue sous licence MIT.\n",
    "  Pour la documentation et des exemples, consultez la vignette du package.\n",
    "=======================================================================================================\n"
  )

  # Check for optional dependencies
  missing_pkgs <- character(0)
  
  # Check H5 support
  if (!requireNamespace("hdf5r", quietly = TRUE)) {
    missing_pkgs <- c(missing_pkgs, "hdf5r (for H5AD/H5 file reading)")
  }
  
  # Check UMAP support
  if (!requireNamespace("uwot", quietly = TRUE)) {
    missing_pkgs <- c(missing_pkgs, "uwot (for UMAP embeddings)")
  }
  
  # Check digest for strict validation
  if (!requireNamespace("digest", quietly = TRUE)) {
    missing_pkgs <- c(missing_pkgs, "digest (for strict M validation)")
  }
  
  if (length(missing_pkgs) > 0) {
    packageStartupMessage(
      "\nOptional dependencies not installed:\n",
      paste0("  - ", missing_pkgs, collapse = "\n"), "\n",
      "\nInstall them manually if needed, e.g.:\n",
      "  install.packages(c('hdf5r', 'uwot', 'digest'))\n"
    )
  }
}

# .onLoad <- function(libname, pkgname) {
#   # Load the DLL
#   library.dynam(pkgname, pkgname, libname)
# }

.onUnload <- function(libpath) {
  library.dynam.unload("gedi2", libpath)
}


#' List Optional GEDI Dependencies
#'
#' Reports which optional packages are needed and provides install commands.
#' Does **not** install anything automatically.
#'
#' @param which Character vector specifying which dependency groups to query.
#'   Options: "h5" (hdf5r), "umap" (uwot), "validation" (digest), or "all" (default).
#' @param verbose Logical, whether to print messages (default: TRUE)
#'
#' @return A named logical vector indicating which packages are installed (invisibly).
#'
#' @examples
#' # Show which optional packages are missing
#' install_optional_dependencies()
#'
#' @export
install_optional_dependencies <- function(which = "all", verbose = TRUE) {

  # Define package groups
  pkg_groups <- list(
    h5 = c("hdf5r"),
    umap = c("uwot"),
    validation = c("digest")
  )

  # Determine which packages to check
  if ("all" %in% which) {
    packages <- unlist(pkg_groups, use.names = FALSE)
  } else {
    invalid <- setdiff(which, names(pkg_groups))
    if (length(invalid) > 0) {
      stop("Invalid dependency group(s): ", paste(invalid, collapse = ", "), "\n",
           "Valid options: ", paste(names(pkg_groups), collapse = ", "), ", all",
           call. = FALSE)
    }
    packages <- unlist(pkg_groups[which], use.names = FALSE)
  }

  installed <- vapply(packages, function(pkg) {
    requireNamespace(pkg, quietly = TRUE)
  }, logical(1))
  to_install <- packages[!installed]

  if (length(to_install) > 0) {
    if (verbose) {
      message("The following optional packages are not installed:")
      message("  ", paste(to_install, collapse = ", "))
      message("\nTo install them, run:")
      message("  install.packages(c(", paste0("'", to_install, "'", collapse = ", "), "))")
    }
  } else {
    if (verbose) message("All requested optional packages are already installed.")
  }

  invisible(installed)
}


#' Check GEDI Optional Dependencies
#'
#' Check which optional dependencies are installed and display their status.
#'
#' @return Named logical vector indicating which optional packages are installed
#'
#' @examples
#' check_optional_dependencies()
#'
#' @export
check_optional_dependencies <- function() {
  
  deps <- c(
    "hdf5r" = "H5AD/H5 file reading",
    "uwot" = "UMAP embeddings",
    "digest" = "Strict M validation (cryptographic hash)"
  )
  
  status <- vapply(names(deps), function(pkg) {
    requireNamespace(pkg, quietly = TRUE)
  }, logical(1))
  
  message("=== GEDI Optional Dependencies ===\n")
  for (i in seq_along(deps)) {
    pkg <- names(deps)[i]
    desc <- deps[i]
    installed <- status[i]
    
    status_msg <- if (installed) "\u2713 Installed" else "\u2717 Not installed"
    message(sprintf("%-12s %-40s %s", pkg, paste0("(", desc, ")"), status_msg))
  }
  
  if (!all(status)) {
    message("\nTo install missing dependencies: install_optional_dependencies()")
  }
  
  invisible(status)
}

# tests/testthat/test-check-deps.R
# -----------------------------------------------------------
# Tests for check_optional_dependencies(). No model needed.
# -----------------------------------------------------------

test_that("check_optional_dependencies returns a named logical vector", {
  result <- check_optional_dependencies()
  expect_type(result, "logical")
  expect_true(length(result) > 0)
  expect_true(all(c("hdf5r", "uwot", "digest") %in% names(result)))
})

#' Write GEDI model to H5AD file
#'
#' Exports a trained GEDI model to H5AD (AnnData) format for interoperability
#' with Python tools like scanpy. The file contains expression data, embeddings,
#' metadata, and GEDI-specific parameters.
#'
#' @param model GEDI R6 object (must be trained)
#' @param file_path Character. Path where the H5AD file should be written.
#' @param X_slot Character. Which expression data to save in the main X slot:
#'   \itemize{
#'     \item "imputed": Imputed expression (default, requires M matrix)
#'     \item "projection": ZDB projection (always available)
#'     \item "original": Original M matrix (requires M parameter)
#'   }
#' @param M Optional. Original count matrix to save when X_slot="original" or
#'   include_raw=TRUE. Must match the dimensions used during model training.
#' @param include_embeddings Logical. Include PCA/UMAP in obsm if cached. Default TRUE.
#' @param include_raw Logical. If TRUE, saves original M in raw.X (requires M parameter).
#'   Default FALSE.
#' @param compression Integer. Gzip compression level (0-9). Default 6.
#' @param verbose Logical. Print progress messages. Default TRUE.
#'
#' @return Invisibly returns the file path
#'
#' @details
#' The H5AD file structure contains:
#' \itemize{
#'   \item X: Main expression matrix (based on X_slot parameter)
#'   \item obs: Cell metadata (sample IDs, colData)
#'   \item var: Gene metadata (gene IDs)
#'   \item obsm: Cell embeddings (X_gedi, X_pca, X_umap)
#'   \item varm: Gene loadings (gedi_Z, gedi_Q_mean)
#'   \item uns: Model parameters and metadata
#'   \item raw.X: Original counts (if include_raw=TRUE)
#' }
#'
#' The function handles the technical details of HDF5/AnnData compatibility:
#' \itemize{
#'   \item Writes sparse matrices in CSR format
#'   \item Transposes dense matrices for Python's row-major layout
#'   \item Creates scalar string attributes (not arrays) for AnnData compatibility
#'   \item Validates M matrix identity using fingerprints
#' }
#'
#' @examples
#' \donttest{
#' if (requireNamespace("hdf5r", quietly = TRUE) &&
#'     requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'
#'   # Write to a temporary file (CRAN policy: never write to the user's
#'   # working directory in examples).
#'   tmp <- tempfile(fileext = ".h5ad")
#'   write_h5ad(model, tmp)
#'   write_h5ad(model, tmp, X_slot = "projection")
#'   unlink(tmp)
#' }
#' }
#'
#' @importFrom Matrix t
#' @export
write_h5ad <- function(model,
                       file_path,
                       X_slot = c("imputed", "projection", "original"),
                       M = NULL,
                       include_embeddings = TRUE,
                       include_raw = FALSE,
                       compression = 6,
                       verbose = TRUE) {
  if (verbose) message("[write_h5ad] Starting H5AD file writing process...")

  # Validate dependencies
  if (!requireNamespace("hdf5r", quietly = TRUE)) {
    stop(
      "[write_h5ad] ERROR: Package 'hdf5r' is required but not installed.\n",
      "  Install with: install.packages('hdf5r')"
    )
  }

  if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop(
      "[write_h5ad] ERROR: Package 'Matrix' is required but not installed.\n",
      "  Install with: install.packages('Matrix')"
    )
  }

  # Validate model
  if (!inherits(model, "GEDI")) {
    stop("[write_h5ad] ERROR: model must be a GEDI R6 object")
  }

  if (is.null(model$params)) {
    stop("[write_h5ad] ERROR: Model not trained. Run model$train() first.")
  }

  X_slot <- match.arg(X_slot)

  # Get model components early to check obs type
  aux <- model$aux

  # Validate M if provided
  if (!is.null(M)) {
    # Check if model has M fingerprint (only for M/M_paired obs types)
    if (aux$obs.type %in% c("M", "M_paired")) {
      # Access private environment to get M fingerprint
      model_env <- environment(model$train)
      private_env <- model_env$private

      if (!is.null(private_env$.M_fingerprint)) {
        if (verbose) message("[write_h5ad] Validating M matrix identity...")
        validate_M_identity(M, private_env$.M_fingerprint)
        if (verbose) message("[write_h5ad]   M matrix validated successfully")
      } else {
        warning("[write_h5ad] No M fingerprint stored in model. Skipping validation.")
      }
    }
  }

  # Check if file exists
  if (file.exists(file_path)) {
    warning("[write_h5ad] File already exists and will be overwritten: ", file_path)
    file.remove(file_path)
  }

  # Create H5 file
  if (verbose) message("[write_h5ad] Creating H5AD file: ", file_path)
  h5file <- tryCatch(
    {
      hdf5r::H5File$new(file_path, mode = "w")
    },
    error = function(e) {
      stop(
        "[write_h5ad] ERROR: Failed to create H5AD file.\n",
        "  File: ", file_path, "\n",
        "  Error: ", conditionMessage(e)
      )
    }
  )

  on.exit(
    {
      if (!is.null(h5file)) {
        h5file$close_all()
        if (verbose) message("[write_h5ad] H5 file closed successfully")
      }
    },
    add = TRUE
  )

  # Get model components
  metadata <- model$metadata
  params <- model$params
  aux <- model$aux

  J <- aux$J
  N <- aux$N
  K <- aux$K

  if (verbose) {
    message("[write_h5ad] Model dimensions: ", J, " genes x ", N, " cells")
    message("[write_h5ad] Latent factors: K = ", K)
  }

  # =========================================================================
  # STEP 1: Prepare and write main expression matrix (X)
  # =========================================================================

  if (verbose) message("[write_h5ad] Preparing main expression matrix (X)...")

  X_matrix <- NULL
  X_description <- ""

  if (X_slot == "imputed") {
    # Check if imputation is available
    if (aux$obs.type %in% c("M", "M_paired")) {
      if (verbose) message("[write_h5ad]   Computing imputed expression...")
      X_matrix <- model$imputed$Y()
      X_description <- "GEDI imputed expression"
    } else {
      warning("[write_h5ad] Imputed expression not available (no M matrix). Using projection instead.")
      X_matrix <- model$projections$ZDB
      X_description <- "GEDI projection (ZDB)"
    }
  } else if (X_slot == "projection") {
    if (verbose) message("[write_h5ad]   Using ZDB projection...")
    X_matrix <- model$projections$ZDB
    X_description <- "GEDI projection (ZDB)"
  } else if (X_slot == "original") {
    if (is.null(M)) {
      stop(
        "[write_h5ad] ERROR: X_slot='original' requires M parameter.\n",
        "  Usage: write_h5ad(model, file, X_slot='original', M=count_matrix)"
      )
    }

    if (verbose) message("[write_h5ad]   Using original M matrix...")

    # Handle paired M case
    if (is.list(M)) {
      if (verbose) message("[write_h5ad]   Detected paired M matrices - using M[[1]]...")
      X_matrix <- M[[1]]
      X_description <- "Original counts (M[[1]] from paired data)"
    } else {
      X_matrix <- M
      X_description <- "Original counts"
    }

    # Validate dimensions
    if (nrow(X_matrix) != J || ncol(X_matrix) != N) {
      stop(
        "[write_h5ad] ERROR: M dimensions (", nrow(X_matrix), " x ", ncol(X_matrix),
        ") don't match model dimensions (", J, " x ", N, ")"
      )
    }
  }

  # Transpose to cells x genes (AnnData format)
  X_matrix <- Matrix::t(X_matrix)

  if (verbose) message("[write_h5ad]   Writing X matrix (", nrow(X_matrix), " x ", ncol(X_matrix), ")...")
  .write_h5ad_sparse_matrix(h5file, "X", X_matrix, compression, verbose)

  # =========================================================================
  # STEP 2: Write cell metadata (obs)
  # =========================================================================

  if (verbose) message("[write_h5ad] Writing cell metadata (obs)...")

  obs_df <- data.frame(
    row.names = metadata$cellIDs,
    sample = metadata$Samples,
    stringsAsFactors = FALSE
  )

  # Add colData if available
  if (!is.null(metadata$colData)) {
    obs_df <- cbind(obs_df, metadata$colData)
  }

  .write_h5ad_dataframe(h5file, "obs", obs_df, verbose)

  # =========================================================================
  # STEP 3: Write gene metadata (var)
  # =========================================================================

  if (verbose) message("[write_h5ad] Writing gene metadata (var)...")

  var_df <- data.frame(
    row.names = metadata$geneIDs,
    gene_ids = metadata$geneIDs,
    stringsAsFactors = FALSE
  )

  .write_h5ad_dataframe(h5file, "var", var_df, verbose)

  # =========================================================================
  # STEP 4: Write multi-dimensional cell annotations (obsm)
  # =========================================================================

  if (verbose) message("[write_h5ad] Writing cell embeddings (obsm)...")

  obsm_group <- h5file$create_group("obsm")

  # Always include GEDI embeddings (DB)
  # DB is K x N, need cells x K for AnnData
  # HDF5 stores in column-major (R), Python reads row-major (C)
  # Write transposed so Python reads correct dimensions
  DB <- Matrix::t(model$projections$DB) # K x N -> N x K
  .write_h5ad_dense_matrix(obsm_group, "X_gedi", t(as.matrix(DB)), verbose)

  # Include PCA and UMAP if cached and requested
  if (include_embeddings) {
    cache_status <- model$cache_status()

    if (cache_status["pca"]) {
      if (verbose) message("[write_h5ad]   Including PCA coordinates...")
      pca <- model$get_pca()
      # PCA needs transpose for Python compatibility
      .write_h5ad_dense_matrix(obsm_group, "X_pca", t(pca), verbose)
    }

    # Check if UMAP is available (would be in cache if computed)
    # Note: We don't have a direct cache flag for UMAP, so we skip for now
    # Users can compute UMAP separately and add it
  }

  # =========================================================================
  # STEP 5: Write multi-dimensional gene annotations (varm)
  # =========================================================================

  if (verbose) message("[write_h5ad] Writing gene loadings (varm)...")

  varm_group <- h5file$create_group("varm")

  # Write Z matrix (gene factor loadings)
  # Z is J x K, need to transpose for Python compatibility
  .write_h5ad_dense_matrix(varm_group, "gedi_Z", t(params$Z), verbose)

  # Write mean Q across samples
  Q_mean <- Reduce("+", params$Qi) / length(params$Qi)
  # Q is J x K, need to transpose for Python compatibility
  .write_h5ad_dense_matrix(varm_group, "gedi_Q_mean", t(Q_mean), verbose)

  # =========================================================================
  # STEP 6: Write unstructured metadata (uns)
  # =========================================================================

  if (verbose) message("[write_h5ad] Writing model parameters (uns)...")

  uns_group <- h5file$create_group("uns")
  gedi_group <- uns_group$create_group("gedi")

  # Model parameters
  gedi_group[["K"]] <- K
  gedi_group[["sigma2"]] <- params$sigma2
  gedi_group[["mode"]] <- aux$mode
  gedi_group[["orthoZ"]] <- aux$orthoZ
  gedi_group[["adjustD"]] <- aux$adjustD
  gedi_group[["obs_type"]] <- aux$obs.type

  # Write D and S vectors
  .write_h5ad_vector(gedi_group, "D", params$D)
  .write_h5ad_vector(gedi_group, "S", params$S)

  # Write o vector
  .write_h5ad_vector(gedi_group, "o", params$o)

  # Metadata
  gedi_group[["n_genes"]] <- J
  gedi_group[["n_cells"]] <- N
  gedi_group[["n_samples"]] <- aux$numSamples

  # Sample names
  if (length(metadata$sampleIDs) > 0) {
    .write_h5ad_string_array(gedi_group, "sample_names", metadata$sampleIDs)
  }

  # Store description of X slot
  uns_group[["X_description"]] <- X_description

  # =========================================================================
  # STEP 7: Write priors if available
  # =========================================================================

  priors <- model$priors
  if (!is.null(priors$inputC) || !is.null(priors$inputH)) {
    if (verbose) message("[write_h5ad] Writing prior matrices...")

    priors_group <- gedi_group$create_group("priors")

    if (!is.null(priors$inputC)) {
      # C is J x P, transpose for Python compatibility
      .write_h5ad_dense_matrix(priors_group, "C", t(priors$inputC), verbose)
      priors_group[["P"]] <- aux$P
    }

    if (!is.null(priors$inputH)) {
      # H is J x L, transpose for Python compatibility
      .write_h5ad_dense_matrix(priors_group, "H", t(priors$inputH), verbose)
      priors_group[["L"]] <- aux$L
    }
  }

  # =========================================================================
  # STEP 8: Write raw counts if requested
  # =========================================================================

  if (include_raw) {
    if (is.null(M)) {
      warning(
        "[write_h5ad] include_raw=TRUE requires M parameter. Skipping raw.X.\n",
        "  Usage: write_h5ad(model, file, include_raw=TRUE, M=count_matrix)"
      )
    } else {
      if (verbose) message("[write_h5ad] Writing raw counts (raw.X)...")

      # Create raw group
      raw_group <- h5file$create_group("raw")

      # Prepare raw matrix
      raw_matrix <- if (is.list(M)) M[[1]] else M

      # Validate dimensions
      if (nrow(raw_matrix) != J || ncol(raw_matrix) != N) {
        warning("[write_h5ad] M dimensions don't match model dimensions. Skipping raw.X.")
      } else {
        # Transpose to cells x genes (AnnData format)
        raw_matrix <- Matrix::t(raw_matrix)

        # Write raw.X
        .write_h5ad_sparse_matrix(raw_group, "X", raw_matrix, compression, verbose)

        # Write raw.var (same as main var)
        .write_h5ad_dataframe(raw_group, "var", var_df, verbose = FALSE)
      }
    }
  }

  # =========================================================================
  # STEP 9: Set top-level attributes
  # =========================================================================

  # Note: encoding-type and encoding-version attributes removed
  # They cause issues with AnnData due to being stored as arrays
  # AnnData can still read the file based on structure alone

  if (verbose) {
    file_size_mb <- file.info(file_path)$size / 1024^2
    message("[write_h5ad] SUCCESS: H5AD file written")
    message("[write_h5ad] File size: ", round(file_size_mb, 2), " MB")
    message("[write_h5ad] Path: ", file_path)
  }

  invisible(file_path)
}


#' Write sparse matrix to H5AD file (CSR format)
#' @keywords internal
#' @noRd
.write_h5ad_sparse_matrix <- function(h5file, name, matrix, compression = 6, verbose = TRUE) {
  # Convert to dgCMatrix if not already
  if (!inherits(matrix, "dgCMatrix")) {
    matrix <- Matrix::Matrix(matrix, sparse = TRUE)
  }

  # Convert to CSR format (AnnData standard)
  # dgCMatrix is CSC (column-compressed), we need CSR (row-compressed)
  # Transpose, convert, then adjust
  matrix_csr <- as(Matrix::t(as(Matrix::t(matrix), "CsparseMatrix")), "RsparseMatrix")

  # Create group for sparse matrix
  mat_group <- h5file$create_group(name)

  # Get CSR components
  # In R's dgRMatrix: x = data, j = column indices, p = row pointers
  data <- matrix_csr@x
  indices <- matrix_csr@j # 0-based in R sparse matrices
  indptr <- matrix_csr@p # 0-based

  # Write components
  mat_group$create_dataset(
    "data",
    robj = data,
    dtype = hdf5r::h5types$H5T_NATIVE_FLOAT,
    space = hdf5r::H5S$new("simple", dims = length(data), maxdims = length(data)),
    gzip_level = compression
  )

  mat_group$create_dataset(
    "indices",
    robj = as.integer(indices),
    dtype = hdf5r::h5types$H5T_NATIVE_INT32,
    space = hdf5r::H5S$new("simple", dims = length(indices), maxdims = length(indices)),
    gzip_level = compression
  )

  mat_group$create_dataset(
    "indptr",
    robj = as.integer(indptr),
    dtype = hdf5r::h5types$H5T_NATIVE_INT32,
    space = hdf5r::H5S$new("simple", dims = length(indptr), maxdims = length(indptr)),
    gzip_level = compression
  )

  # Write attributes for AnnData compatibility
  hdf5r::h5attr(mat_group, "shape") <- as.integer(c(nrow(matrix), ncol(matrix)))

  # Write string attributes as scalars for Python compatibility
  .write_h5ad_scalar_string_attr(mat_group, "encoding-type", "csr_matrix")
  .write_h5ad_scalar_string_attr(mat_group, "encoding-version", "0.1.0")

  if (verbose) {
    sparsity <- 100 * (1 - length(data) / (nrow(matrix) * ncol(matrix)))
    message(
      "[write_h5ad]     Sparse matrix '", name, "': ",
      nrow(matrix), " x ", ncol(matrix),
      " (", round(sparsity, 1), "% sparse)"
    )
  }
}


#' Write dense matrix to H5AD file
#' @keywords internal
#' @noRd
.write_h5ad_dense_matrix <- function(h5group, name, matrix, verbose = TRUE) {
  # Convert to regular matrix if sparse
  if (inherits(matrix, "sparseMatrix")) {
    matrix <- as.matrix(matrix)
  }

  # Ensure it's a matrix
  if (!is.matrix(matrix)) {
    matrix <- as.matrix(matrix)
  }

  # Create dataset
  h5group$create_dataset(
    name,
    robj = matrix,
    dtype = hdf5r::h5types$H5T_NATIVE_FLOAT
  )

  # Note: encoding attributes removed for AnnData compatibility

  if (verbose) {
    message(
      "[write_h5ad]     Dense matrix '", name, "': ",
      nrow(matrix), " x ", ncol(matrix)
    )
  }
}


#' Write data frame to H5AD file
#' @keywords internal
#' @noRd
.write_h5ad_dataframe <- function(h5file, name, df, verbose = TRUE) {
  df_group <- h5file$create_group(name)

  # Write _index (row names)
  if (!is.null(rownames(df))) {
    .write_h5ad_string_array(df_group, "_index", rownames(df))
  }

  # Write columns
  for (col_name in colnames(df)) {
    col_data <- df[[col_name]]

    if (is.factor(col_data)) {
      # Write as categorical
      col_group <- df_group$create_group(col_name)

      categories <- levels(col_data)
      codes <- as.integer(col_data) - 1L # 0-based indexing

      .write_h5ad_string_array(col_group, "categories", categories)

      col_group$create_dataset(
        "codes",
        robj = codes,
        dtype = hdf5r::h5types$H5T_NATIVE_INT32
      )

      # Note: encoding attributes removed for AnnData compatibility
      hdf5r::h5attr(col_group, "ordered") <- FALSE
    } else if (is.character(col_data)) {
      .write_h5ad_string_array(df_group, col_name, col_data)
    } else if (is.numeric(col_data)) {
      df_group$create_dataset(
        col_name,
        robj = col_data,
        dtype = hdf5r::h5types$H5T_NATIVE_DOUBLE
      )
      # Note: encoding attributes removed for AnnData compatibility
    } else {
      warning(
        "[write_h5ad] Skipping column '", col_name,
        "' with unsupported type: ", class(col_data)[1]
      )
    }
  }

  # Write dataframe encoding attributes for AnnData compatibility
  .write_h5ad_scalar_string_attr(df_group, "_index", "_index")
  .write_h5ad_scalar_string_attr(df_group, "encoding-type", "dataframe")
  .write_h5ad_scalar_string_attr(df_group, "encoding-version", "0.2.0")

  # Write column-order attribute
  if (length(colnames(df)) > 0) {
    # Column-order should be an attribute, not a dataset
    col_order_dtype <- hdf5r::H5T_STRING$new(size = Inf)
    col_order_dtype$set_cset(hdf5r::h5const$H5T_CSET_UTF8)
    space <- hdf5r::H5S$new("simple", dims = length(colnames(df)), maxdims = length(colnames(df)))
    attr <- df_group$create_attr("column-order", dtype = col_order_dtype, space = space)
    attr$write(colnames(df))
    attr$close()
  } else {
    # Empty column-order for dataframes with no columns (just index)
    hdf5r::h5attr(df_group, "column-order") <- numeric(0)
  }

  if (verbose) {
    message(
      "[write_h5ad]   DataFrame '", name, "': ",
      nrow(df), " rows, ", ncol(df), " columns"
    )
  }
}


#' Write string array to H5AD file
#' @keywords internal
#' @noRd
.write_h5ad_string_array <- function(h5group, name, strings) {
  # Convert to character
  strings <- as.character(strings)

  # Create variable-length string datatype
  str_dtype <- hdf5r::H5T_STRING$new(size = Inf)
  str_dtype$set_cset(hdf5r::h5const$H5T_CSET_UTF8)

  # Create dataset
  h5group$create_dataset(
    name,
    robj = strings,
    dtype = str_dtype
  )

  # Note: encoding attributes removed for AnnData compatibility
}


#' Write numeric vector to H5AD file
#' @keywords internal
#' @noRd
.write_h5ad_vector <- function(h5group, name, vec) {
  h5group$create_dataset(
    name,
    robj = as.numeric(vec),
    dtype = hdf5r::h5types$H5T_NATIVE_DOUBLE
  )

  # Note: encoding attributes removed for AnnData compatibility
}


#' Write scalar string attribute to H5 object
#'
#' Writes a string attribute as a true scalar (not 1-element array) for Python compatibility.
#' By default, hdf5r writes string attributes as 1-element arrays which causes issues with AnnData.
#'
#' @keywords internal
#' @noRd
.write_h5ad_scalar_string_attr <- function(h5obj, attr_name, attr_value) {
  # Create variable-length string datatype
  str_type <- hdf5r::H5T_STRING$new(size = Inf)
  str_type$set_cset(hdf5r::h5const$H5T_CSET_UTF8)

  # Create scalar dataspace (this is key - not "simple" with dims=1)
  space <- hdf5r::H5S$new(type = "scalar")

  # Create and write attribute
  attr <- h5obj$create_attr(attr_name, dtype = str_type, space = space)
  attr$write(attr_value)
  attr$close()
}

# ==============================================================================
# GEDI Imputation Accessors - R6 Methods
# User-facing interface for imputation, variance, and dispersion analysis
# ==============================================================================

#' Create Imputation Accessor Object
#'
#' @description
#' Creates accessor object for imputation analysis with Y, variance, and
#' dispersion methods. This provides a clean user interface: model$imputed$Y(M)
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#'
#' @return S3 object of class "gedi_imputation"
#'
#' @keywords internal
#' @noRd
create_imputation_accessor <- function(self, private) {
  
  structure(
    list(
      Y = function(M = NULL, logScale = TRUE, rowCentre = TRUE) {
        get_imputed_Y(self, private, M, logScale, rowCentre)
      },
      
      variance = function(M) {
        get_Y_variance(self, private, M)
      },
      
      dispersion = function(M, subsample = 1e6) {
        get_dispersion(self, private, M, subsample)
      }
    ),
    class = "gedi_imputation"
  )
}


#' Get Imputed Y Expression (with caching)
#'
#' @description
#' Returns imputed gene expression with sample-specific effects removed.
#' Results are cached for faster repeated access.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param M Count matrix (optional - only needed for non-log scale output)
#' @param logScale Logical, if TRUE returns log-scale values (default)
#' @param rowCentre Logical, if TRUE removes global gene offset o (default)
#'
#' @return Dense matrix (J x N) with imputed expression values
#'
#' @keywords internal
#' @noRd
get_imputed_Y <- function(self, private, M = NULL, logScale = TRUE, rowCentre = TRUE) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  obs_type <- self$aux$obs.type
  
  # Create cache key
  cache_key <- paste0("Y_imputed_", logScale, "_", rowCentre)
  
  # Check if cached
  if (!is.null(private$.imputation_cache) &&
      !is.null(private$.imputation_cache[[cache_key]])) {
    log_cached("Imputed Y", private$.verbose)
    return(private$.imputation_cache[[cache_key]])
  }
  
  # For non-log scale output, need M for proper transformation
  if (!logScale && is.null(M)) {
    if (obs_type %in% c("M", "M_paired")) {
      stop("M must be provided when logScale = FALSE for observation type '", 
           obs_type, "'", call. = FALSE)
    }
  }
  
  # Validate M if provided
  if (!is.null(M)) {
    validate_M_identity(M, private$.M_fingerprint)
  }
  
  # Split by sample
  numSamples <- self$aux$numSamples
  Samples <- private$.samples
  sampleNames <- private$.sampleNames
  
  cells_by_sample <- split(seq_along(Samples), Samples)[sampleNames]
  
  # Compute imputed Y for each sample
  Y_imputed_list <- vector("list", numSamples)

  if (private$.verbose == 1) {
    message(sprintf("Computing imputed Y (%d samples)", numSamples))
    pb <- txtProgressBar(min = 0, max = numSamples, style = 3,
                         width = 50, file = stderr())
    on.exit(close(pb), add = TRUE)

    for (i in 1:numSamples) {
      # Reconstruct Yi_fitted from parameters (NO M needed!)
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Remove sample effects
      Y_imputed_list[[i]] <- compute_Y_imputed(
        Yi_fitted = Yi_fitted,
        params = self$params,
        sample_idx = i,
        rowCentre = rowCentre
      )

      setTxtProgressBar(pb, i)
    }

  } else {
    # Silent or debug mode - no progress bar
    for (i in 1:numSamples) {
      # Reconstruct Yi_fitted from parameters (NO M needed!)
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Remove sample effects
      Y_imputed_list[[i]] <- compute_Y_imputed(
        Yi_fitted = Yi_fitted,
        params = self$params,
        sample_idx = i,
        rowCentre = rowCentre
      )
    }
  }

  # Concatenate all samples
  Y_imputed <- do.call(cbind, Y_imputed_list)

  # Add dimension names
  rownames(Y_imputed) <- private$.geneIDs
  colnames(Y_imputed) <- private$.cellIDs
  
  # Transform to non-log scale if requested
  if (!logScale) {
    if (obs_type == "M") {
      Y_imputed <- exp(Y_imputed)
      
    } else if (obs_type == "M_paired") {
      Y_imputed <- 1 / (1 + exp(-Y_imputed))
      
    } else if (obs_type == "Y") {
      Y_imputed <- exp(Y_imputed)
      
    } else {
      stop("Unrecognized observation type: ", obs_type, call. = FALSE)
    }
  }
  
  # Cache result
  if (is.null(private$.imputation_cache)) {
    private$.imputation_cache <- list()
  }
  private$.imputation_cache[[cache_key]] <- Y_imputed

  details <- format_dims(nrow(Y_imputed), ncol(Y_imputed), "genes", "cells")
  log_complete("Imputed Y", details, private$.verbose)
  log_info(sprintf("logScale = %s, rowCentre = %s", logScale, rowCentre), private$.verbose)
  
  return(Y_imputed)
}


#' Get Y Variance (with caching)
#'
#' @description
#' Returns posterior variance of imputed Y. Results are cached.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param M Count matrix (required - must match original)
#'
#' @return Dense matrix (J x N) with variance values, or NULL if not applicable
#'
#' @keywords internal
#' @noRd
get_Y_variance <- function(self, private, M) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if M provided
  if (missing(M) || is.null(M)) {
    stop("M must be provided for variance calculation", call. = FALSE)
  }
  
  # Check cache
  if (!is.null(private$.imputation_cache) &&
      !is.null(private$.imputation_cache$Y_var)) {
    log_cached("Y variance", private$.verbose)
    return(private$.imputation_cache$Y_var)
  }
  
  # Call core computation function
  Y_var <- compute_Y_variance(self, private, M)
  
  # Cache result
  if (!is.null(Y_var)) {
    if (is.null(private$.imputation_cache)) {
      private$.imputation_cache <- list()
    }
    private$.imputation_cache$Y_var <- Y_var

    details <- format_dims(nrow(Y_var), ncol(Y_var), "genes", "cells")
    log_complete("Y variance", details, private$.verbose)
  }
  
  return(Y_var)
}


#' Get Dispersion Analysis
#'
#' @description
#' Analyzes dispersion (variance vs mean relationship) for count data.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param M Count matrix (required - must match original)
#' @param subsample Integer, maximum number of nonzero positions to sample
#'
#' @return Data frame with dispersion statistics
#'
#' @keywords internal
#' @noRd
get_dispersion <- function(self, private, M, subsample = 1e6) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if M provided
  if (missing(M) || is.null(M)) {
    stop("M must be provided for dispersion analysis", call. = FALSE)
  }
  
  # Call core computation function
  dispersion_df <- compute_dispersion(self, private, M, subsample)

  details <- sprintf("%d bins across %d samples", nrow(dispersion_df), self$aux$numSamples)
  log_complete("Dispersion analysis", details, private$.verbose)

  return(dispersion_df)
}


#' Print Method for Imputation Accessor
#'
#' @param x Object of class gedi_imputation
#' @param ... Additional arguments (ignored)
#'
#' @return Invisibly returns \code{x}.
#' @method print gedi_imputation
#' @keywords internal
#' @export
print.gedi_imputation <- function(x, ...) {
  cat("<GEDI Imputation Interface>\n")
  cat("\nAvailable methods:\n\n")
  
  cat("Imputed Expression:\n")
  cat("  $Y(M = NULL, logScale = TRUE, rowCentre = TRUE)\n")
  cat("    Get imputed gene expression with sample effects removed\n")
  cat("    - M: Optional count matrix (needed for non-log scale)\n")
  cat("    - logScale: Return log-transformed values (default: TRUE)\n")
  cat("    - rowCentre: Remove global gene offset (default: TRUE)\n\n")
  
  cat("Variance Analysis:\n")
  cat("  $variance(M)\n")
  cat("    Get posterior variance of imputed Y\n")
  cat("    - M: Count matrix (required, must match original)\n\n")
  
  cat("Dispersion Analysis:\n")
  cat("  $dispersion(M, subsample = 1e6)\n")
  cat("    Analyze variance vs mean relationship\n")
  cat("    - M: Count matrix (required, must match original)\n")
  cat("    - subsample: Max nonzero positions to sample per sample\n\n")
  
  cat("Example usage:\n")
  cat("  # Get imputed expression (no M needed!)\n")
  cat("  Y_imputed <- model$imputed$Y()\n")
  cat("  Y_raw_scale <- model$imputed$Y(M, logScale = FALSE)\n\n")
  
  cat("  # Variance and dispersion (M required)\n")
  cat("  Y_var <- model$imputed$variance(M)\n")
  cat("  disp <- model$imputed$dispersion(M)\n")
  
  invisible(x)
}

# Improved Plotting API with Smart Caching and Model-First Interface
# These wrappers provide a cleaner user experience while maintaining backwards compatibility

#' Get Embedding Coordinates with Smart Caching
#'
#' @param model GEDI model object
#' @param embedding_type Character: "umap", "pca", or a custom Nx2 matrix
#' @param dims Integer vector of length 2 for which dimensions to use (default c(1,2))
#' @param verbose Logical, print messages about computation
#' @return Nx2 matrix of embedding coordinates
#' @keywords internal
.get_embedding <- function(model, embedding_type, dims = c(1, 2), verbose = TRUE) {
  # If it's already a matrix, return it
  if (is.matrix(embedding_type) || is.data.frame(embedding_type)) {
    emb <- as.matrix(embedding_type)
    if (ncol(emb) < 2) {
      stop("Custom embedding must have at least 2 columns", call. = FALSE)
    }
    return(emb[, dims])
  }

  # Handle string embedding types
  if (!is.character(embedding_type)) {
    stop("embedding must be 'umap', 'pca', or a matrix", call. = FALSE)
  }

  embedding_type <- tolower(embedding_type)

  if (embedding_type == "umap") {
    # UMAP can be accessed as field or method
    # As field: uses cached default UMAP
    # As method with no args: also uses cached default UMAP
    umap_emb <- model$embeddings$umap
    return(umap_emb[, dims])
  } else if (embedding_type == "pca") {
    # PCA returns cached value if available
    return(model$embeddings$pca[, dims])
  } else {
    stop("embedding must be 'umap', 'pca', or a matrix", call. = FALSE)
  }
}


#' Get Color Vector from Model
#'
#' @param model GEDI model object
#' @param color_by Character: "sample", metadata column name, or gene name/index
#' @param projection Character: "zdb" or "db" for gene expression
#' @return Vector of length N with color values
#' @keywords internal
.get_color_vector <- function(model, color_by, projection = "zdb") {
  if (is.null(color_by)) {
    return(NULL)
  }

  # If it's already a vector, return it
  if (is.vector(color_by) && length(color_by) > 1) {
    N <- ncol(model$params$Bi[[1]]) # Number of cells
    if (length(color_by) != N) {
      stop("color_by vector must have length N (number of cells)", call. = FALSE)
    }
    return(color_by)
  }

  # Handle string color_by
  if (!is.character(color_by)) {
    stop("color_by must be a character or vector", call. = FALSE)
  }

  # Check if it's "sample"
  if (tolower(color_by) == "sample") {
    return(model$metadata$Samples)
  }

  # Check if it's a metadata column
  if (!is.null(model$metadata$colData)) {
    if (color_by %in% colnames(model$metadata$colData)) {
      return(model$metadata$colData[[color_by]])
    }
  }

  # Check if it's a gene name
  geneIDs <- model$metadata$geneIDs
  if (color_by %in% geneIDs) {
    gene_idx <- match(color_by, geneIDs)

    # Compute projection on-demand for this gene only
    Z <- model$params$Z
    feature_weights <- Z[gene_idx, ] # K x 1

    if (projection == "zdb") {
      expr_values <- compute_feature_projection(
        feature_weights = feature_weights,
        D = model$params$D,
        Bi_list = model$params$Bi,
        verbose = 0
      )
    } else if (projection == "db") {
      DB <- model$projections$DB # K x N
      expr_values <- as.vector(t(DB) %*% feature_weights) # N x 1
    } else {
      stop("projection must be 'zdb' or 'db'", call. = FALSE)
    }

    return(expr_values)
  }

  # If we get here, color_by wasn't found
  stop("color_by '", color_by, "' not found in samples, metadata, or genes", call. = FALSE)
}


#' Plot Embedding with Improved API
#'
#' Simplified interface for plotting embeddings with automatic caching.
#' Model is the first argument, and color_by handles metadata/genes automatically.
#'
#' @param model GEDI model object (or embedding matrix for backwards compatibility)
#' @param embedding Character ("umap", "pca") or Nx2 matrix
#' @param color_by Character: "sample", metadata column, gene name, or NULL
#' @param color Vector for manual coloring (overrides color_by)
#' @param projection Character: "zdb" or "db" for gene expression projection
#' @param color_limits Numeric vector c(low, high) or NULL for auto
#' @param palette Character vector of colors for continuous scale
#' @param randomize Logical, randomize point order
#' @param point_size Numeric, size of points
#' @param alpha Numeric, transparency (0-1)
#' @param raster Logical, use rasterization for large datasets
#' @param xlab Character, x-axis label
#' @param ylab Character, y-axis label
#' @param title Character, plot title
#' @param legend_title Character, legend title
#' @param verbose Logical, print computation messages
#'
#' @return ggplot2 object
#'
#' @examples
#' \donttest{
#' if (requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'   plot_embedding(model, embedding = "pca", color_by = "sample")
#' }
#' }
#'
#' @export
plot_embedding <- function(model,
                           embedding = NULL,
                           color_by = NULL,
                           color = NULL,
                           projection = "zdb",
                           color_limits = NULL,
                           palette = c("blue", "lightgrey", "red"),
                           randomize = TRUE,
                           point_size = 0.3,
                           alpha = 0.9,
                           raster = FALSE,
                           xlab = "Dim 1",
                           ylab = "Dim 2",
                           title = NULL,
                           legend_title = NULL,
                           verbose = TRUE) {
  # Backwards compatibility: if model is a matrix, use old API
  if (is.matrix(model) || is.data.frame(model)) {
    # Old API: plot_embedding(embedding_matrix, color = values)
    embedding_mat <- as.matrix(model)

    if (ncol(embedding_mat) != 2) {
      stop("embedding must have exactly 2 columns", call. = FALSE)
    }

    N <- nrow(embedding_mat)

    # Handle color
    if (is.null(color)) {
      color <- rep(1, N)
      use_color <- FALSE
    } else {
      if (length(color) != N) {
        stop("color must have same length as number of rows in embedding", call. = FALSE)
      }
      use_color <- TRUE
    }

    # Create data frame
    df <- data.frame(
      Dim1 = embedding_mat[, 1],
      Dim2 = embedding_mat[, 2],
      Color = color
    )

    # Randomize once before processing
    if (randomize) {
      rand_idx <- sample(N)
      df <- df[rand_idx, ]
    }

    # Detect if color is numeric or discrete
    is_numeric <- is.numeric(color) && length(unique(color)) > 10

    # Start plot
    p <- ggplot2::ggplot(df, ggplot2::aes(x = Dim1, y = Dim2))

    # Add points
    if (raster && N > 100000) {
      if (!use_color) {
        p <- p + ggplot2::geom_point(size = point_size, alpha = alpha)
      } else {
        p <- p + ggplot2::geom_point(ggplot2::aes(color = Color),
          size = point_size, alpha = alpha
        )
      }
      warning("Rasterization not implemented yet. Using regular points.",
        call. = FALSE
      )
    } else {
      if (!use_color) {
        p <- p + ggplot2::geom_point(size = point_size, alpha = alpha)
      } else {
        p <- p + ggplot2::geom_point(ggplot2::aes(color = Color),
          size = point_size, alpha = alpha
        )
      }
    }

    # Add color scale
    if (use_color) {
      if (is_numeric) {
        # Continuous color scale
        if (is.null(color_limits)) {
          color_limits <- compute_color_limits(color, symmetric = TRUE, quantile = 0.99)
        }

        if (is.null(legend_title)) {
          legend_title <- "Value"
        }

        p <- p + scale_color_gedi_diverging(
          limits = color_limits,
          name = legend_title
        )
      } else {
        # Discrete color scale
        if (is.null(legend_title)) {
          legend_title <- "Group"
        }

        p <- p + scale_color_gedi_discrete(name = legend_title) +
          ggplot2::guides(color = ggplot2::guide_legend(
            override.aes = list(size = 3, alpha = 1)
          ))
      }
    }

    # Add labels and theme
    p <- p +
      ggplot2::labs(x = xlab, y = ylab, title = title) +
      theme_gedi()

    return(p)
  }

  # New API: plot_embedding(model, embedding = "umap", color_by = "sample")
  if (!inherits(model, "GEDI")) {
    stop("First argument must be a GEDI model or embedding matrix", call. = FALSE)
  }

  if (is.null(embedding)) {
    stop("embedding argument is required (e.g., 'umap', 'pca')", call. = FALSE)
  }

  # Get embedding coordinates with smart caching
  embedding_mat <- .get_embedding(model, embedding, dims = c(1, 2), verbose = verbose)

  # Get color vector if color_by is specified (overrides manual color)
  if (!is.null(color_by) && is.null(color)) {
    color <- .get_color_vector(model, color_by, projection = projection)

    # Set legend title if not specified
    if (is.null(legend_title)) {
      if (is.character(color_by)) {
        legend_title <- color_by
      }
    }
  }

  # Set axis labels based on embedding type
  if (xlab == "Dim 1" && ylab == "Dim 2") {
    emb_name <- if (is.character(embedding)) toupper(embedding) else "Embedding"
    xlab <- paste(emb_name, "1")
    ylab <- paste(emb_name, "2")
  }

  # Call the base function with extracted data
  .plot_embedding_base(
    embedding = embedding_mat,
    color = color,
    color_limits = color_limits,
    palette = palette,
    randomize = randomize,
    point_size = point_size,
    alpha = alpha,
    raster = raster,
    xlab = xlab,
    ylab = ylab,
    title = title,
    legend_title = legend_title
  )
}

# tests/testthat/test-plot-embedding.R
# -----------------------------------------------------------
# Tests for plot_embedding() using the backwards-compatible
# (matrix) API — no model or optional packages needed.
# -----------------------------------------------------------

test_that("plot_embedding works with matrix + continuous color (old API)", {
  set.seed(1)
  emb  <- matrix(rnorm(200), ncol = 2)
  vals <- rnorm(100)
  p <- plot_embedding(emb, color = vals)
  expect_s3_class(p, "ggplot")
})

test_that("plot_embedding works with matrix + discrete color (old API)", {
  set.seed(1)
  emb    <- matrix(rnorm(200), ncol = 2)
  groups <- rep(c("A", "B"), each = 50)
  p <- plot_embedding(emb, color = groups)
  expect_s3_class(p, "ggplot")
})

test_that("plot_embedding works with no color (old API)", {
  set.seed(1)
  emb <- matrix(rnorm(200), ncol = 2)
  p <- plot_embedding(emb)
  expect_s3_class(p, "ggplot")
})

test_that("plot_embedding errors on wrong embedding dimensions", {
  emb <- matrix(rnorm(300), ncol = 3)
  expect_error(plot_embedding(emb))
})

test_that("plot_embedding errors on mismatched color length", {
  emb <- matrix(rnorm(200), ncol = 2)
  expect_error(plot_embedding(emb, color = 1:5))
})

# ==============================================================================
# GEDI Progress and Logging Utilities
# Unified interface for progress bars and status messages
# ==============================================================================

#' Create Progress Bar (Method 4: Base R txtProgressBar)
#'
#' @description
#' Creates a simple text-based progress bar for iterative operations.
#' Use this for loops over samples or other countable iterations.
#'
#' @param total Integer, total number of iterations
#' @param label Character, optional label to display (default: "Progress")
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show progress)
#'
#' @return Progress bar object or NULL if verbose = 0
#'
#' @keywords internal
#' @noRd
create_progress_bar <- function(total, label = "Progress", verbose = 1) {
  if (verbose < 1 || total <= 1) {
    return(NULL)
  }

  if (nchar(label) > 0) {
    message(label)
  }

  pb <- txtProgressBar(min = 0, max = total, style = 3, width = 50)
  return(pb)
}


#' Update Progress Bar
#'
#' @description
#' Updates the progress bar to the current iteration.
#'
#' @param pb Progress bar object (from create_progress_bar)
#' @param current Integer, current iteration number
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
update_progress_bar <- function(pb, current) {
  if (!is.null(pb)) {
    setTxtProgressBar(pb, current)
  }
  invisible(NULL)
}


#' Close Progress Bar
#'
#' @description
#' Closes the progress bar and prints a newline.
#'
#' @param pb Progress bar object (from create_progress_bar)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
close_progress_bar <- function(pb) {
  if (!is.null(pb)) {
    close(pb)
  }
  invisible(NULL)
}


#' Log Stage Message (Method 7: Progress with custom messages)
#'
#' @description
#' Prints a stage message for multi-stage operations.
#' Format: "  [Stage X/Y] Message"
#'
#' @param stage Integer, current stage number
#' @param total Integer, total number of stages
#' @param message Character, description of this stage
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
log_stage <- function(stage, total, message, verbose = 1) {
  if (verbose >= 1) {
    message(sprintf("  [%d/%d] %s", stage, total, message))
  }
  invisible(NULL)
}


#' Log Start Message
#'
#' @description
#' Prints a start message for an operation.
#' Format: "Computing [operation]..."
#'
#' @param operation Character, name of the operation
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
log_start <- function(operation, verbose = 1) {
  if (verbose >= 1) {
    message("Computing ", operation, "...")
  }
  invisible(NULL)
}


#' Log Complete Message
#'
#' @description
#' Prints a completion message for an operation.
#' Format: "[OK] [operation] computed: [details]"
#'
#' @param operation Character, name of the operation
#' @param details Character, optional details about dimensions/size
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
log_complete <- function(operation, details = NULL, verbose = 1) {
  if (verbose >= 1) {
    if (is.null(details)) {
      message("[DONE] ", operation, " complete")
    } else {
      message("[DONE] ", operation, " computed: ", details)
    }
  }
  invisible(NULL)
}


#' Log Info Message
#'
#' @description
#' Prints an informational message.
#' Format: "  [info]"
#'
#' @param message Character, the message to display
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show)
#' @param indent Logical, whether to indent the message (default: TRUE)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
log_info <- function(message, verbose = 1, indent = TRUE) {
  if (verbose >= 1) {
    if (indent) {
      message("  ", message)
    } else {
      message(message)
    }
  }
  invisible(NULL)
}


#' Log Cached Message
#'
#' @description
#' Prints a message indicating cached data is being used.
#' Format: "[CACHED] Using cached [item]"
#'
#' @param item Character, name of the cached item
#' @param verbose Integer, verbosity level (0 = silent, 1+ = show)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
log_cached <- function(item, verbose = 1) {
  if (verbose >= 1) {
    message("[CACHED] Using cached ", item)
  }
  invisible(NULL)
}


#' Format Dimensions Message
#'
#' @description
#' Creates a formatted string for matrix dimensions.
#' Format: "X rows x Y cols" or "X genes x Y cells"
#'
#' @param nrow Integer, number of rows
#' @param ncol Integer, number of columns
#' @param row_label Character, label for rows (default: "rows")
#' @param col_label Character, label for columns (default: "cols")
#'
#' @return Character string with formatted dimensions
#'
#' @keywords internal
#' @noRd
format_dims <- function(nrow, ncol, row_label = "rows", col_label = "cols") {
  sprintf("%d %s x %d %s", nrow, row_label, ncol, col_label)
}

# tests/testthat/helper-synthetic-data.R
# Shared synthetic data generator for all test files.
# Sourced automatically by testthat before any test-*.R file.

#' Create a small synthetic negative-binomial count matrix with sample labels
#'
#' @param n_genes Number of genes (rows)
#' @param n_cells Number of cells (columns)
#' @param n_samples Number of biological samples
#' @param seed Random seed for reproducibility
#' @return A list with components: M (sparse dgCMatrix), sample_labels (factor)
make_synthetic_data <- function(n_genes = 200,
                                n_cells = 400,
                                n_samples = 4,
                                seed = 42) {
  set.seed(seed)
  cells_per_sample <- n_cells / n_samples
  stopifnot(n_cells %% n_samples == 0)

  sample_labels <- factor(rep(paste0("S", seq_len(n_samples)),
                              each = cells_per_sample))

  gene_means  <- runif(n_genes, min = 0.5, max = 8)
  dispersion  <- 0.5

  counts <- matrix(
    stats::rnbinom(n_genes * n_cells,
                   size = dispersion,
                   mu   = rep(gene_means, n_cells)),
    nrow = n_genes,
    ncol = n_cells
  )
  rownames(counts) <- paste0("Gene_", seq_len(n_genes))
  colnames(counts) <- paste0("Cell_", seq_len(n_cells))

  M <- Matrix::Matrix(counts, sparse = TRUE)

  list(M = M, sample_labels = sample_labels)
}

#' Train a small GEDI model on synthetic data (cached per session)
#'
#' @param K Latent factors
#' @param iterations Training iterations
#' @return Trained GEDI model object
make_trained_model <- function(K = 5, iterations = 20) {
  dat   <- make_synthetic_data()
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = K,
    mode    = "Bsphere",
    verbose = 0
  )
  model$train(iterations = iterations)
  model
}

# tests/testthat/test-plot-features.R
# -----------------------------------------------------------
# Tests for plot_features() and plot_feature_ratio().
# Require a trained model + custom embedding matrix.
# -----------------------------------------------------------

test_that("plot_features returns a ggplot with integer indices", {
  model <- make_trained_model(K = 5, iterations = 10)
  # Use a random 2-column matrix as embedding to avoid uwot dependency
  set.seed(1)
  emb <- matrix(rnorm(400 * 2), ncol = 2)

  p <- plot_features(model, features = c(1, 2, 3), embedding = emb)
  expect_s3_class(p, "ggplot")
})

test_that("plot_features returns a ggplot with gene names", {
  model <- make_trained_model(K = 5, iterations = 10)
  set.seed(1)
  emb <- matrix(rnorm(400 * 2), ncol = 2)
  genes <- model$metadata$geneIDs[1:2]

  p <- plot_features(model, features = genes, embedding = emb)
  expect_s3_class(p, "ggplot")
})

test_that("plot_features errors on invalid gene name", {
  model <- make_trained_model(K = 5, iterations = 10)
  set.seed(1)
  emb <- matrix(rnorm(400 * 2), ncol = 2)
  expect_error(plot_features(model, features = "NONEXISTENT_GENE", embedding = emb))
})

test_that("plot_feature_ratio returns a ggplot with integer indices", {
  model <- make_trained_model(K = 5, iterations = 10)
  set.seed(1)
  emb <- matrix(rnorm(400 * 2), ncol = 2)

  p <- plot_feature_ratio(model, gene1 = 1, gene2 = 2, embedding = emb)
  expect_s3_class(p, "ggplot")
})

test_that("plot_feature_ratio errors on untrained model", {
  dat <- make_synthetic_data()
  model <- CreateGEDIObject(
    Samples = dat$sample_labels, M = dat$M,
    K = 5, mode = "Bsphere", verbose = 0
  )
  set.seed(1)
  emb <- matrix(rnorm(400 * 2), ncol = 2)
  expect_error(plot_feature_ratio(model, gene1 = 1, gene2 = 2, embedding = emb))
})

# tests/testthat/test-plot-convergence.R
# -----------------------------------------------------------
# Tests for plot_convergence(). Requires a trained model.
# -----------------------------------------------------------

test_that("plot_convergence returns a ggplot (faceted layout)", {
  model <- make_trained_model(K = 5, iterations = 20)
  p <- plot_convergence(model, layout = "faceted")
  expect_s3_class(p, "ggplot")
})

test_that("plot_convergence returns a list of ggplots (separate layout)", {
  model <- make_trained_model(K = 5, iterations = 20)
  plots <- plot_convergence(model, layout = "separate")
  expect_type(plots, "list")
  expect_true(length(plots) > 0)
  # Each element should be a ggplot
  for (p in plots) {
    expect_s3_class(p, "ggplot")
  }
})

test_that("plot_convergence errors on untrained model", {
  dat <- make_synthetic_data()
  model <- CreateGEDIObject(
    Samples = dat$sample_labels, M = dat$M,
    K = 5, mode = "Bsphere", verbose = 0
  )
  expect_error(plot_convergence(model))
})

# ==============================================================================
# Seurat Integration Functions
# Convert between Seurat and GEDI objects
# ==============================================================================

#' Convert Seurat Object to GEDI Model
#'
#' @description
#' Extracts count data from a Seurat object and creates a GEDI model.
#' Automatically validates that the data contains raw counts (not normalized).
#' Handles both Seurat v4 and v5, including split layers in v5.
#'
#' @param seurat_object Seurat object
#' @param assay Character, which assay to use (default: "RNA")
#' @param slot Character, which slot/layer to extract (default: "counts").
#'   For Seurat v5 with split layers (e.g., counts.CTRL, counts.STIM),
#'   this will automatically detect and combine all matching layers.
#' @param sample_column Character, column name in meta.data for sample labels
#'   (default: "orig.ident")
#' @param subset_samples Character vector, subset to specific samples (default: NULL = all)
#' @param K Integer, number of latent factors (default: 10)
#' @param mode Character, normalization mode: "Bl2" or "Bsphere" (default: "Bl2")
#' @param C Gene-level prior matrix (genes x pathways) (default: NULL)
#' @param H Sample-level covariate matrix (covariates x samples) (default: NULL)
#' @param validate_counts Logical, whether to validate data appears to be counts
#'   (default: TRUE)
#' @param use_variable_features Logical, whether to subset to highly variable features
#'   (default: TRUE). If TRUE, uses genes from VariableFeatures(seurat_object).
#' @param verbose Logical, whether to print progress messages (default: TRUE)
#' @param ... Additional arguments passed to CreateGEDIObject()
#'
#' @return GEDI R6 object
#'
#' @examples
#' \donttest{
#' library(Seurat)
#'
#' # Load example data
#' pbmc <- SeuratObject::pbmc_small
#'
#' # Basic usage
#' gedi_model <- seurat_to_gedi(
#'   seurat_object = pbmc,
#'   sample_column = "orig.ident",
#'   K = 15
#' )
#'
#' # Train the model
#' gedi_model$train(iterations = 10)
#' }
#'
#' @export
seurat_to_gedi <- function(seurat_object,
                           assay = "RNA",
                           slot = "counts",
                           sample_column = "orig.ident",
                           subset_samples = NULL,
                           K = 10,
                           mode = "Bl2",
                           C = NULL,
                           H = NULL,
                           validate_counts = TRUE,
                           use_variable_features = TRUE,
                           verbose = TRUE,
                           ...) {
  # Check if Seurat is available
  if (!requireNamespace("Seurat", quietly = TRUE)) {
    stop("Package 'Seurat' is required. Install with: install.packages('Seurat')",
      call. = FALSE
    )
  }

  # Validate inputs
  if (!inherits(seurat_object, "Seurat")) {
    stop("seurat_object must be a Seurat object", call. = FALSE)
  }

  if (!assay %in% names(seurat_object@assays)) {
    stop("Assay '", assay, "' not found. Available: ",
      paste(names(seurat_object@assays), collapse = ", "),
      call. = FALSE
    )
  }

  # Extract count matrix (handle both Seurat v4 and v5)
  # Seurat v5 can have split layers like counts.CTRL, counts.STIM
  assay_obj <- seurat_object@assays[[assay]]

  # Detect if we have multiple layers for the requested slot (Seurat v5 split layers)
  if (methods::is(assay_obj, "Assay5")) {
    # Seurat v5 - check for split layers
    layer_names <- names(assay_obj@layers)
    slot_pattern <- paste0("^", slot, "\\.") # e.g., "^counts\\."

    # Find all layers matching the slot pattern
    matching_layers <- grep(slot_pattern, layer_names, value = TRUE)

    if (length(matching_layers) > 1) {
      # Multiple split layers found - need to join them
      if (verbose) {
        message(
          "Detected ", length(matching_layers), " split layers: ",
          paste(matching_layers, collapse = ", ")
        )
      }
      if (verbose) message("Joining layers with do.call(cbind)...")

      # Extract each layer and combine
      layer_matrices <- lapply(matching_layers, function(layer_name) {
        Seurat::GetAssayData(seurat_object, assay = assay, layer = layer_name)
      })

      # Combine all layers column-wise
      M <- do.call(cbind, layer_matrices)
    } else if (length(matching_layers) == 1) {
      # Single layer with pattern name (e.g., just "counts.sample1")
      M <- Seurat::GetAssayData(seurat_object, assay = assay, layer = matching_layers[1])
    } else {
      # No pattern match - try direct slot name
      M <- tryCatch(
        Seurat::GetAssayData(seurat_object, assay = assay, layer = slot),
        error = function(e) {
          stop("No layers found matching '", slot, "' in Seurat v5 assay. ",
            "Available layers: ", paste(layer_names, collapse = ", "),
            call. = FALSE
          )
        }
      )
    }

    if (slot == "data") {
      warning("Using 'data' layer instead of 'counts'. ",
        "GEDI expects raw counts for proper modeling.",
        call. = FALSE
      )
    }
  } else {
    # Seurat v4 or earlier - use traditional slot access
    if (slot == "counts") {
      M <- tryCatch(
        Seurat::GetAssayData(seurat_object, assay = assay, layer = "counts"),
        error = function(e) {
          Seurat::GetAssayData(seurat_object, assay = assay, slot = "counts")
        }
      )
    } else if (slot == "data") {
      M <- tryCatch(
        Seurat::GetAssayData(seurat_object, assay = assay, layer = "data"),
        error = function(e) {
          Seurat::GetAssayData(seurat_object, assay = assay, slot = "data")
        }
      )
      warning("Using 'data' slot instead of 'counts'. ",
        "GEDI expects raw counts for proper modeling.",
        call. = FALSE
      )
    } else {
      stop("slot must be 'counts' or 'data'", call. = FALSE)
    }
  }

  # Subset to highly variable features if requested
  if (use_variable_features) {
    hvg <- Seurat::VariableFeatures(seurat_object)

    if (length(hvg) == 0) {
      warning("use_variable_features = TRUE but no variable features found in Seurat object. ",
        "Run FindVariableFeatures() first, or set use_variable_features = FALSE. ",
        "Using all genes.",
        call. = FALSE
      )
    } else {
      # Subset to HVGs that exist in the matrix
      hvg_in_matrix <- intersect(hvg, rownames(M))

      if (length(hvg_in_matrix) == 0) {
        stop("None of the variable features are found in the matrix. ",
          "Check that VariableFeatures match the gene names in the assay.",
          call. = FALSE
        )
      }

      if (verbose) message("Subsetting to ", length(hvg_in_matrix), " highly variable features")
      M <- M[hvg_in_matrix, , drop = FALSE]
    }
  }

  # Validate that data appears to be counts (uses deterministic check to avoid RNG)
  if (validate_counts && slot == "counts") {
    # Use deterministic sampling: take evenly spaced indices instead of random sample
    n_total <- length(M@x) # Number of non-zero elements in sparse matrix
    if (n_total > 0) {
      # Sample up to 10000 elements evenly spaced
      n_sample <- min(10000, n_total)
      sample_indices <- as.integer(seq(1, n_total, length.out = n_sample))
      sample_values <- M@x[sample_indices]

      # Check if values are integers or very close to integers
      if (length(sample_values) > 0) {
        decimal_parts <- abs(sample_values - round(sample_values))
        non_integer_frac <- mean(decimal_parts > 1e-6)

        if (non_integer_frac > 0.01) {
          warning("Data does not appear to be raw counts (contains many non-integer values). ",
            "GEDI expects raw count data. Set validate_counts=FALSE to skip this check.",
            call. = FALSE
          )
        }
      }
    }
  }

  # Extract sample labels
  if (!sample_column %in% colnames(seurat_object@meta.data)) {
    stop("Column '", sample_column, "' not found in meta.data. Available: ",
      paste(head(colnames(seurat_object@meta.data), 10), collapse = ", "),
      call. = FALSE
    )
  }

  Samples <- as.character(seurat_object@meta.data[[sample_column]])

  # Subset to specific samples if requested
  if (!is.null(subset_samples)) {
    keep_cells <- Samples %in% subset_samples
    if (sum(keep_cells) == 0) {
      stop("No cells found for samples: ", paste(subset_samples, collapse = ", "),
        call. = FALSE
      )
    }

    M <- M[, keep_cells]
    Samples <- Samples[keep_cells]
    if (verbose) {
      message(
        "Subset to ", sum(keep_cells), " cells from ",
        length(subset_samples), " samples"
      )
    }
  }

  # Extract metadata (exclude sample_column to avoid duplication)
  colData <- seurat_object@meta.data
  if (!is.null(subset_samples)) {
    colData <- colData[keep_cells, , drop = FALSE]
  }

  # Remove sample column from colData (already in Samples)
  colData[[sample_column]] <- NULL

  # Create GEDI object
  if (verbose) message("Creating GEDI object from Seurat data...")
  if (verbose) message("  Genes: ", nrow(M))
  if (verbose) message("  Cells: ", ncol(M))
  if (verbose) message("  Samples: ", length(unique(Samples)))

  gedi_model <- CreateGEDIObject(
    Samples = Samples,
    M = M,
    colData = colData,
    K = K,
    mode = mode,
    C = C,
    H = H,
    ...
  )

  if (verbose) message("GEDI object created successfully!")

  return(gedi_model)
}


#' Convert GEDI Model to Seurat Object
#'
#' @description
#' Creates a Seurat object from a trained GEDI model, including imputed data,
#' projections, and embeddings.
#'
#' @param model GEDI model object (trained)
#' @param M Original count matrix (optional). If not provided, will use
#'   back-transformed imputed values as approximate counts.
#' @param project Character, project name for Seurat object (default: "GEDI")
#' @param assay Character, name for the main assay (default: "RNA")
#' @param use_imputed Logical, whether to add imputed data as separate assay
#'   (default: TRUE)
#' @param add_projections Logical, whether to add ZDB and DB projections as
#'   separate assays (default: TRUE)
#' @param add_embeddings Logical, whether to add UMAP and PCA embeddings if
#'   available (default: TRUE)
#' @param min_cells Integer, filter genes with counts in < min_cells (default: 0)
#' @param min_features Integer, filter cells with < min_features genes (default: 0)
#' @param verbose Logical, whether to print progress messages (default: TRUE)
#'
#' @return Seurat object with:
#'   \itemize{
#'     \item RNA assay: Original or back-transformed counts
#'     \item imputed assay: GEDI imputed expression (if use_imputed = TRUE)
#'     \item ZDB/DB/ADB assays: GEDI projections (if add_projections = TRUE)
#'       \itemize{
#'         \item ZDB: Batch-corrected gene expression (genes x cells)
#'         \item DB: Cell embeddings in latent factor space (K x cells)
#'         \item ADB: Pathway activities (pathways x cells) - only if C matrix provided
#'       }
#'     \item umap/pca reductions: Embeddings (if add_embeddings = TRUE and cached)
#'     \item meta.data: Sample labels and colData from GEDI model
#'   }
#'
#' @examples
#' \donttest{
#' # Load example data
#' pbmc <- SeuratObject::pbmc_small
#'
#' # Train GEDI model
#' gedi_model <- seurat_to_gedi(pbmc, K = 15)
#' gedi_model$train(iterations = 10)
#'
#' # Convert back to Seurat
#' seurat_obj <- gedi_to_seurat(
#'   gedi_model,
#'   use_imputed = TRUE,
#'   add_projections = TRUE,
#'   add_embeddings = TRUE
#' )
#'
#' # Now can use Seurat functions
#' library(Seurat)
#' DefaultAssay(seurat_obj) <- "imputed"
#' seurat_obj <- FindVariableFeatures(seurat_obj)
#' }
#'
#' @export
gedi_to_seurat <- function(model,
                           M = NULL,
                           project = "GEDI",
                           assay = "RNA",
                           use_imputed = TRUE,
                           add_projections = TRUE,
                           add_embeddings = TRUE,
                           min_cells = 0,
                           min_features = 0,
                           verbose = TRUE) {
  # Check if Seurat is available
  if (!requireNamespace("Seurat", quietly = TRUE)) {
    stop("Package 'Seurat' is required. Install with: install.packages('Seurat')",
      call. = FALSE
    )
  }

  # Validate model is trained
  if (is.null(model$params)) {
    stop("Model not trained. Run model$train() first.", call. = FALSE)
  }

  if (verbose) message("Converting GEDI model to Seurat object...")

  # Get metadata
  gene_ids <- model$metadata$geneIDs
  cell_ids <- model$metadata$cellIDs
  samples <- model$metadata$Samples

  # Prepare count matrix
  if (is.null(M)) {
    if (verbose) message("  No original counts provided, using back-transformed imputed values...")
    # Get imputed log-expression and back-transform to counts (approximate)
    Y_imputed <- model$imputed$Y()
    M <- Matrix::Matrix(expm1(Y_imputed), sparse = TRUE)
  } else {
    if (verbose) message("  Using provided count matrix...")
    # Ensure M has correct dimensions and ordering
    if (!identical(dim(M), c(length(gene_ids), length(cell_ids)))) {
      stop("M dimensions don't match model (expected ", length(gene_ids),
        " x ", length(cell_ids), ")",
        call. = FALSE
      )
    }
  }

  # Set dimnames
  rownames(M) <- gene_ids
  colnames(M) <- cell_ids

  # Create Seurat object
  if (verbose) message("  Creating Seurat object...")
  seurat_obj <- Seurat::CreateSeuratObject(
    counts = M,
    project = project,
    assay = assay,
    min.cells = min_cells,
    min.features = min_features
  )

  # Add sample information to meta.data
  seurat_obj@meta.data$Sample <- samples

  # Add colData if available
  if (!is.null(model$metadata$colData)) {
    colData <- model$metadata$colData
    # Match to cells in Seurat object (in case filtering occurred)
    common_cells <- intersect(cell_ids, colnames(seurat_obj))
    if (length(common_cells) > 0) {
      colData_matched <- colData[match(common_cells, cell_ids), , drop = FALSE]
      rownames(colData_matched) <- common_cells

      # Add each column to meta.data
      for (col in colnames(colData_matched)) {
        seurat_obj@meta.data[[col]] <- colData_matched[[col]]
      }
    }
  }

  # Add imputed data as separate assay
  if (use_imputed) {
    if (verbose) message("  Adding imputed assay...")
    Y_imputed <- model$imputed$Y()
    rownames(Y_imputed) <- gene_ids
    colnames(Y_imputed) <- cell_ids

    # Filter to cells in Seurat object
    common_cells <- intersect(cell_ids, colnames(seurat_obj))
    Y_imputed <- Y_imputed[, common_cells, drop = FALSE]

    # Create imputed assay (use log-transformed data as 'data' slot)
    imputed_assay <- Seurat::CreateAssayObject(
      data = Matrix::Matrix(Y_imputed, sparse = TRUE)
    )
    seurat_obj[["imputed"]] <- imputed_assay
  }

  # Add projections as separate assays
  if (add_projections) {
    if (verbose) message("  Adding projection assays...")

    # ZDB projection (J genes x N cells)
    ZDB <- model$projections$ZDB
    rownames(ZDB) <- gene_ids
    colnames(ZDB) <- cell_ids
    common_cells <- intersect(cell_ids, colnames(seurat_obj))
    ZDB <- ZDB[, common_cells, drop = FALSE]

    zdb_assay <- Seurat::CreateAssayObject(
      data = Matrix::Matrix(ZDB, sparse = TRUE)
    )
    seurat_obj[["ZDB"]] <- zdb_assay

    # DB projection (K factors x N cells)
    DB <- model$projections$DB
    K <- nrow(DB)
    rownames(DB) <- paste0("LV", 1:K)
    colnames(DB) <- cell_ids
    DB <- DB[, common_cells, drop = FALSE]

    db_assay <- Seurat::CreateAssayObject(
      data = Matrix::Matrix(DB, sparse = TRUE)
    )
    seurat_obj[["DB"]] <- db_assay

    # ADB projection (P pathways x N cells) - only if model has C matrix
    if (model$aux$P > 0) {
      ADB <- model$projections$ADB
      P <- nrow(ADB)
      # Row names should already be pathway names from C matrix
      colnames(ADB) <- cell_ids
      ADB <- ADB[, common_cells, drop = FALSE]

      adb_assay <- Seurat::CreateAssayObject(
        data = Matrix::Matrix(ADB, sparse = TRUE)
      )
      seurat_obj[["ADB"]] <- adb_assay
    }
  }

  # Add embeddings
  if (add_embeddings) {
    cache_status <- model$cache_status()

    # Add UMAP if cached
    if (cache_status["umap"]) {
      if (verbose) message("  Adding UMAP embedding...")
      umap_coords <- model$embeddings$umap
      common_cells <- intersect(cell_ids, colnames(seurat_obj))
      umap_coords <- umap_coords[match(common_cells, cell_ids), , drop = FALSE]
      rownames(umap_coords) <- common_cells
      colnames(umap_coords) <- c("UMAP_1", "UMAP_2")

      seurat_obj[["umap"]] <- Seurat::CreateDimReducObject(
        embeddings = umap_coords,
        key = "UMAP_",
        assay = assay
      )
    }

    # Add PCA if cached
    if (cache_status["pca"]) {
      if (verbose) message("  Adding PCA embedding...")
      pca_coords <- model$embeddings$pca
      common_cells <- intersect(cell_ids, colnames(seurat_obj))
      pca_coords <- pca_coords[match(common_cells, cell_ids), , drop = FALSE]
      rownames(pca_coords) <- common_cells
      colnames(pca_coords) <- paste0("PC_", 1:ncol(pca_coords))

      seurat_obj[["pca"]] <- Seurat::CreateDimReducObject(
        embeddings = pca_coords,
        key = "PC_",
        assay = assay
      )
    }
  }

  if (verbose) message("Seurat object created successfully!")
  if (verbose) message("  Assays: ", paste(names(seurat_obj@assays), collapse = ", "))
  if (verbose && length(seurat_obj@reductions) > 0) {
    message("  Reductions: ", paste(names(seurat_obj@reductions), collapse = ", "))
  }

  return(seurat_obj)
}

# ==============================================================================
# GEDI Imputation - R Wrappers
# Lazy-computed imputation methods with caching and validation
# ==============================================================================

#' Compute M Fingerprint for Validation
#'
#' @description
#' Creates a fast, dependency-free fingerprint of the count matrix for identity validation.
#' Checks: dimensions, sparsity, value statistics, sparse indices, and dimnames.
#'
#' @param M Count matrix (sparse or dense), or list of two matrices for M_paired
#' @param obs_type Character, observation type: "M", "M_paired", "Y", or "X"
#'
#' @return List with fingerprint components for validation
#'
#' @keywords internal
#' @noRd
compute_M_fingerprint <- function(M, obs_type) {

  if (obs_type == "M_paired") {
    if (!is.list(M) || length(M) != 2) {
      stop("M must be a list of two matrices for M_paired observation type", call. = FALSE)
    }

    # Compute fingerprints for both matrices
    fp1 <- compute_M_fingerprint_single(M[[1]], "M1")
    fp2 <- compute_M_fingerprint_single(M[[2]], "M2")

    return(list(
      M1 = fp1,
      M2 = fp2,
      obs_type = obs_type
    ))

  } else {
    # Single matrix
    return(list(
      M = compute_M_fingerprint_single(M, "M"),
      obs_type = obs_type
    ))
  }
}


#' Compute Fingerprint for Single Matrix
#'
#' @description
#' Internal helper to compute fingerprint for a single matrix.
#' Does NOT store gene/cell IDs - only dimensions and statistics.
#'
#' @param M Count matrix (sparse or dense)
#' @param label Character label for this matrix (e.g., "M", "M1", "M2")
#'
#' @return List with fingerprint components
#'
#' @keywords internal
#' @noRd
compute_M_fingerprint_single <- function(M, label) {

  # Fast fingerprint with no dependencies - single validation mode
  # Checks: dimensions, sparsity, values, indices, and dimnames

  dims <- dim(M)
  is_sparse <- inherits(M, "sparseMatrix")

  # Extract dimnames
  row_names <- rownames(M)
  col_names <- colnames(M)

  # Compute fast checksums (sparse-aware)
  if (is_sparse) {
    # Sparse matrix - use internal structure directly
    values_sum <- sum(M@x)                    # sum of values
    values_sq <- sum(M@x^2)                   # sum of squares
    nnz <- length(M@x)                        # number of non-zeros
    indices_check <- sum(M@i %% 1000)         # row index checksum (mod to prevent overflow)

  } else {
    # Dense matrix
    values_sum <- sum(M)
    values_sq <- sum(M^2)
    nnz <- sum(M != 0)
    indices_check <- 0  # Not applicable for dense
  }

  # Dimnames fingerprint (fast hash using string length and first/last chars)
  rownames_fp <- if (!is.null(row_names)) {
    paste0(
      length(row_names), "_",
      if (length(row_names) > 0) substr(row_names[1], 1, 3) else "",
      if (length(row_names) > 1) substr(row_names[length(row_names)], 1, 3) else "",
      "_", sum(nchar(row_names))
    )
  } else { "NULL" }

  colnames_fp <- if (!is.null(col_names)) {
    paste0(
      length(col_names), "_",
      if (length(col_names) > 0) substr(col_names[1], 1, 3) else "",
      if (length(col_names) > 1) substr(col_names[length(col_names)], 1, 3) else "",
      "_", sum(nchar(col_names))
    )
  } else { "NULL" }

  # Composite fingerprint string (all checks combined)
  fingerprint_str <- paste0(
    dims[1], "x", dims[2], "_",      # dimensions
    values_sum, "_",                  # sum of values
    values_sq, "_",                   # sum of squares
    nnz, "_",                         # nnz count
    indices_check, "_",               # index checksum
    rownames_fp, "_",                 # rownames fingerprint
    colnames_fp                       # colnames fingerprint
  )

  return(list(
    nrow = dims[1],
    ncol = dims[2],
    is_sparse = is_sparse,
    fingerprint = fingerprint_str,
    label = label
  ))
}


#' Validate M Identity
#'
#' @description
#' Strict validation that M matrix matches the original used during training.
#' Uses structural checks, statistical fingerprints, and random sampling.
#' Does NOT check gene/cell name order - only dimensions and values.
#'
#' @param M Count matrix (sparse or dense), or list for M_paired
#' @param fingerprint Stored fingerprint from model setup
#'
#' @return TRUE if validation passes, otherwise stops with informative error
#'
#' @keywords internal
#' @noRd
validate_M_identity <- function(M, fingerprint) {
  
  obs_type <- fingerprint$obs_type
  
  if (obs_type == "M_paired") {
    if (!is.list(M) || length(M) != 2) {
      stop("M must be a list of two matrices for M_paired observation type", call. = FALSE)
    }
    
    validate_M_single(M[[1]], fingerprint$M1, "M[[1]]")
    validate_M_single(M[[2]], fingerprint$M2, "M[[2]]")
    
  } else {
    validate_M_single(M, fingerprint$M, "M")
  }
  
  return(TRUE)
}


#' Validate Single Matrix
#'
#' @description
#' Internal helper to validate a single matrix against its fingerprint.
#'
#' @param M Count matrix (sparse or dense)
#' @param fp Stored fingerprint for this matrix
#' @param label Character label for error messages
#'
#' @return TRUE if valid, stops with error otherwise
#'
#' @keywords internal
#' @noRd
validate_M_single <- function(M, fp, label) {

  # Quick dimension check first
  if (nrow(M) != fp$nrow) {
    stop(label, " has wrong number of genes: expected ", fp$nrow,
         ", got ", nrow(M), call. = FALSE)
  }
  if (ncol(M) != fp$ncol) {
    stop(label, " has wrong number of cells: expected ", fp$ncol,
         ", got ", ncol(M), call. = FALSE)
  }

  # Check sparsity type
  is_sparse <- inherits(M, "sparseMatrix")
  if (is_sparse != fp$is_sparse) {
    stop(label, " sparsity type changed: expected ",
         ifelse(fp$is_sparse, "sparse", "dense"),
         ", got ", ifelse(is_sparse, "sparse", "dense"), call. = FALSE)
  }

  # Recompute fingerprint for current matrix
  current_fp <- compute_M_fingerprint_single(M, label)

  # Compare fingerprints
  if (current_fp$fingerprint != fp$fingerprint) {
    stop(label, " fingerprint mismatch - matrix was modified.\n",
         "  Expected: ", fp$fingerprint, "\n",
         "  Got: ", current_fp$fingerprint, call. = FALSE)
  }

  return(TRUE)
}


#' Compute Y Fitted from Parameters
#'
#' @description
#' Reconstructs Yi_fitted = ZDBi + QiDBi + si + o + oi from model parameters.
#' This is the model's prediction of log-expression (or log-ratio for M_paired).
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param sample_idx Integer, which sample to compute (1 to numSamples)
#'
#' @return Dense matrix (J x Ni) - fitted Yi for sample i
#'
#' @keywords internal
#' @noRd
compute_Y_fitted <- function(self, private, sample_idx) {
  
  # Validate sample index
  numSamples <- self$aux$numSamples
  if (sample_idx < 1 || sample_idx > numSamples) {
    stop("sample_idx must be between 1 and ", numSamples, call. = FALSE)
  }
  
  # Get parameters
  Z <- self$params$Z
  D <- self$params$D
  Bi <- self$params$Bi[[sample_idx]]
  Qi <- self$params$Qi[[sample_idx]]
  si <- self$params$si[[sample_idx]]
  o <- self$params$o
  oi <- self$params$oi[[sample_idx]]
  
  # Compute ZDBi
  ZD <- Z %*% diag(D, nrow = length(D))
  ZDBi <- ZD %*% Bi
  
  # Compute QiDBi
  QiD <- Qi %*% diag(D, nrow = length(D))
  QiDBi <- QiD %*% Bi
  
  # Call C++ predict_Yhat
  Yi_fitted <- predict_Yhat(
    ZDBi = ZDBi,
    QiDBi = QiDBi,
    si = si,
    o = o,
    oi = oi
  )
  
  return(Yi_fitted)
}


#' Compute Imputed Y with Sample Effects Removed
#'
#' @description
#' Removes sample-specific effects from Yi_fitted to get shared biological signal.
#' This is the "imputed" expression that can be compared across samples.
#'
#' @param Yi_fitted Dense matrix (J x Ni) - fitted Yi for sample i
#' @param params List with model parameters
#' @param sample_idx Integer, which sample this is
#' @param rowCentre Logical, if TRUE removes global gene offset o
#'
#' @return Dense matrix (J x Ni) - imputed Y with sample effects removed
#'
#' @keywords internal
#' @noRd
compute_Y_imputed <- function(Yi_fitted, params, sample_idx, rowCentre) {
  
  Qi <- params$Qi[[sample_idx]]
  D <- params$D
  Bi <- params$Bi[[sample_idx]]
  si <- params$si[[sample_idx]]
  oi <- params$oi[[sample_idx]]
  o <- params$o
  
  # Compute QiDBi
  QiD <- Qi %*% diag(D, nrow = length(D))
  QiDBi <- QiD %*% Bi
  
  # Call C++ Yi_resZ to remove sample effects
  if (rowCentre) {
    Y_imputed <- Yi_resZ(
      Yi = Yi_fitted,
      QiDBi = QiDBi,
      si = si,
      o = o,
      oi = oi
    )
  } else {
    # Don't remove o (set to zero vector)
    Y_imputed <- Yi_resZ(
      Yi = Yi_fitted,
      QiDBi = QiDBi,
      si = si,
      o = rep(0, length(o)),
      oi = oi
    )
  }
  
  return(Y_imputed)
}


#' Compute Y Variance from Raw Counts
#'
#' @description
#' Computes posterior variance of imputed Y given fitted Yi and raw counts.
#' Requires the original M matrix.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param M Count matrix (must match original), or list for M_paired
#'
#' @return Dense matrix (J x N) - variance of imputed Y
#'
#' @keywords internal
#' @noRd
compute_Y_variance <- function(self, private, M) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check observation type
  obs_type <- self$aux$obs.type
  if (obs_type == "Y") {
    if (private$.verbose >= 1) message("Observation type is 'Y' (not raw counts). Returning NULL.")
    return(NULL)
  }
  if (obs_type == "X") {
    if (private$.verbose >= 1) message("Observation type is 'X' (binary). Variance calculation not applicable. Returning NULL.")
    return(NULL)
  }
  
  # Validate M
  validate_M_identity(M, private$.M_fingerprint)
  
  # Convert to sparse if needed
  if (obs_type == "M") {
    if (!inherits(M, "sparseMatrix")) {
      warning("Converting M to sparse format for memory efficiency")
      M <- as(M, "dgCMatrix")
    }
  } else if (obs_type == "M_paired") {
    if (!inherits(M[[1]], "sparseMatrix")) {
      warning("Converting M[[1]] to sparse format for memory efficiency")
      M[[1]] <- as(M[[1]], "dgCMatrix")
    }
    if (!inherits(M[[2]], "sparseMatrix")) {
      warning("Converting M[[2]] to sparse format for memory efficiency")
      M[[2]] <- as(M[[2]], "dgCMatrix")
    }
  }
  
  # Split M by sample
  numSamples <- self$aux$numSamples
  Samples <- private$.samples
  sampleNames <- private$.sampleNames
  
  # Get cell indices for each sample (in original order)
  cells_by_sample <- split(seq_along(Samples), Samples)[sampleNames]
  
  # Compute variance for each sample
  Y_var_list <- vector("list", numSamples)

  if (private$.verbose == 1) {
    message(sprintf("Computing Y variance (%d samples)", numSamples))
    pb <- txtProgressBar(min = 0, max = numSamples, style = 3,
                         width = 50, file = stderr())
    on.exit(close(pb), add = TRUE)

    for (i in 1:numSamples) {
      idx <- cells_by_sample[[i]]

      # Get Yi_fitted for this sample (reconstruct from params)
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Get Mi for this sample
      if (obs_type == "M") {
        Mi <- M[, idx, drop = FALSE]
        Y_var_list[[i]] <- Yi_var_M(Yi_fitted, self$params$sigma2)

      } else if (obs_type == "M_paired") {
        M1i <- M[[1]][, idx, drop = FALSE]
        M2i <- M[[2]][, idx, drop = FALSE]
        Y_var_list[[i]] <- Yi_var_M_paired(Yi_fitted, M1i, M2i, self$params$sigma2)
      }

      setTxtProgressBar(pb, i)
    }

  } else {
    # Silent or debug mode - no progress bar
    for (i in 1:numSamples) {
      idx <- cells_by_sample[[i]]

      # Get Yi_fitted for this sample (reconstruct from params)
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Get Mi for this sample
      if (obs_type == "M") {
        Mi <- M[, idx, drop = FALSE]
        Y_var_list[[i]] <- Yi_var_M(Yi_fitted, self$params$sigma2)

      } else if (obs_type == "M_paired") {
        M1i <- M[[1]][, idx, drop = FALSE]
        M2i <- M[[2]][, idx, drop = FALSE]
        Y_var_list[[i]] <- Yi_var_M_paired(Yi_fitted, M1i, M2i, self$params$sigma2)
      }
    }
  }
  
  # Concatenate all samples
  Y_var <- do.call(cbind, Y_var_list)
  
  # Add dimension names
  rownames(Y_var) <- private$.geneIDs
  colnames(Y_var) <- private$.cellIDs
  
  return(Y_var)
}


#' Compute Dispersion Analysis
#'
#' @description
#' Analyzes dispersion (variance vs mean relationship) for count data.
#' Compares predicted variance (from Poisson model) to observed variance.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param M Count matrix (must match original), or list for M_paired
#' @param subsample Integer, maximum number of nonzero positions to sample per sample
#'
#' @return Data frame with dispersion statistics by bin
#'
#' @keywords internal
#' @noRd
compute_dispersion <- function(self, private, M, subsample = 1e6) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check observation type (only for M)
  obs_type <- self$aux$obs.type
  if (obs_type != "M") {
    stop("Dispersion analysis only available for observation type 'M' (single count matrix)",
         call. = FALSE)
  }
  
  # Validate M
  validate_M_identity(M, private$.M_fingerprint)
  
  # Convert to sparse if needed
  if (!inherits(M, "sparseMatrix")) {
    warning("Converting M to sparse format for memory efficiency")
    M <- as(M, "dgCMatrix")
  }
  
  # Split M by sample
  numSamples <- self$aux$numSamples
  Samples <- private$.samples
  sampleNames <- private$.sampleNames
  
  cells_by_sample <- split(seq_along(Samples), Samples)[sampleNames]
  
  # Compute dispersion for each sample
  result_list <- vector("list", numSamples)

  if (private$.verbose == 1) {
    message(sprintf("Computing dispersion (%d samples)", numSamples))
    pb <- txtProgressBar(min = 0, max = numSamples, style = 3,
                         width = 50, file = stderr())
    on.exit(close(pb), add = TRUE)

    for (i in 1:numSamples) {
      idx <- cells_by_sample[[i]]

      # Get Yi_fitted for this sample
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Get Mi for this sample
      Mi <- M[, idx, drop = FALSE]

      # Call C++ dispersion function (sparse-optimized)
      dispersion_data <- compute_dispersion_sparse(Yi_fitted, Mi, as.integer(subsample))

      predicted <- dispersion_data$predicted
      observed <- dispersion_data$observed

      # Compute residual squared
      res_sq <- (observed - predicted)^2

      # Bin by predicted values
      n_bins <- floor(sqrt(length(predicted)))
      bins <- cut(
        predicted,
        breaks = stats::quantile(predicted, seq(0, 1, length = n_bins + 1)),
        include.lowest = TRUE
      )

      # Aggregate by bin
      xy <- stats::aggregate(
        cbind(predicted, res_sq),
        by = list(bin = bins),
        FUN = mean
      )

      colnames(xy) <- c("bin", "Expected_Var", "Observed_Var")
      xy$Sample <- sampleNames[i]

      # Count observations per bin
      xy$n <- stats::aggregate(
        predicted,
        by = list(bin = bins),
        FUN = length
      )$x

      result_list[[i]] <- xy

      setTxtProgressBar(pb, i)
    }

  } else {
    # Silent or debug mode - no progress bar
    for (i in 1:numSamples) {
      idx <- cells_by_sample[[i]]

      # Get Yi_fitted for this sample
      Yi_fitted <- compute_Y_fitted(self, private, i)

      # Get Mi for this sample
      Mi <- M[, idx, drop = FALSE]

      # Call C++ dispersion function (sparse-optimized)
      dispersion_data <- compute_dispersion_sparse(Yi_fitted, Mi, as.integer(subsample))

      predicted <- dispersion_data$predicted
      observed <- dispersion_data$observed

      # Compute residual squared
      res_sq <- (observed - predicted)^2

      # Bin by predicted values
      n_bins <- floor(sqrt(length(predicted)))
      bins <- cut(
        predicted,
        breaks = stats::quantile(predicted, seq(0, 1, length = n_bins + 1)),
        include.lowest = TRUE
      )

      # Aggregate by bin
      xy <- stats::aggregate(
        cbind(predicted, res_sq),
        by = list(bin = bins),
        FUN = mean
      )

      colnames(xy) <- c("bin", "Expected_Var", "Observed_Var")
      xy$Sample <- sampleNames[i]

      # Count observations per bin
      xy$n <- stats::aggregate(
        predicted,
        by = list(bin = bins),
        FUN = length
      )$x

      result_list[[i]] <- xy
    }
  }
  
  # Combine all samples
  result <- do.call(rbind, result_list)
  rownames(result) <- NULL
  
  # Reorder columns
  result <- result[, c("Sample", "Expected_Var", "Observed_Var", "n", "bin")]
  
  return(result)
}


#' Clear Imputation Cache
#'
#' @description
#' Clears cached imputation results.
#'
#' @param private Reference to private environment
#' @param what Character vector specifying which cache entries to clear.
#'   Options: "Y_fitted", "Y_imputed", or NULL to clear all (default: NULL)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
clear_imputation_cache <- function(private, what = NULL) {
  
  if (is.null(private$.imputation_cache)) {
    return(invisible(NULL))
  }
  
  if (is.null(what)) {
    # Clear all imputation cache
    private$.imputation_cache <- list()
    if (private$.verbose > 0) {
      message("Imputation cache cleared")
    }
  } else {
    # Clear specific entries
    if ("Y_imputed" %in% what) {
      # Clear ALL Y_imputed variants
      y_imputed_keys <- grep("^Y_imputed_", names(private$.imputation_cache), value = TRUE)
      for (key in y_imputed_keys) {
        private$.imputation_cache[[key]] <- NULL
      }
      if (private$.verbose > 0) {
        message("Cache cleared: Y_imputed (all variants)")
      }
    }
    
    if ("Y_var" %in% what) {
      private$.imputation_cache$Y_var <- NULL
      if (private$.verbose > 0) {
        message("Cache cleared: Y_var")
      }
    }
    
    invalid <- setdiff(what, c("Y_imputed", "Y_var"))
    if (length(invalid) > 0) {
      warning("Invalid cache entries: ", paste(invalid, collapse = ", "))
    }
  }
  
  invisible(NULL)
}


#' Get Imputation Cache Status
#'
#' @description
#' Returns information about which imputation results are currently cached.
#'
#' @param private Reference to private environment
#'
#' @return Named logical vector indicating which results are cached
#'
#' @keywords internal
#' @noRd
get_imputation_cache_status <- function(private) {
  
  if (is.null(private$.imputation_cache)) {
    return(c(Y_imputed = FALSE, Y_var = FALSE))
  }
  
  # Check for any Y_imputed variant (different logScale/rowCentre combinations)
  has_Y_imputed <- any(grepl("^Y_imputed_", names(private$.imputation_cache)))
  
  status <- c(
    Y_imputed = has_Y_imputed,
    Y_var = !is.null(private$.imputation_cache$Y_var)
  )
  
  return(status)
}


#' Get Imputation Cache Memory Usage
#'
#' @description
#' Estimates memory usage of cached imputation results in MB.
#'
#' @param private Reference to private environment
#'
#' @return Named numeric vector with memory usage in MB for each cached result
#'
#' @keywords internal
#' @noRd
get_imputation_cache_memory <- function(private) {
  
  if (is.null(private$.imputation_cache)) {
    return(c(Y_imputed = 0, Y_var = 0, Total = 0))
  }
  
  # Sum all Y_imputed variants
  y_imputed_keys <- grep("^Y_imputed_", names(private$.imputation_cache), value = TRUE)
  y_imputed_size <- sum(vapply(y_imputed_keys, function(key) {
    if (!is.null(private$.imputation_cache[[key]])) {
      as.numeric(object.size(private$.imputation_cache[[key]]) / 1024^2)
    } else {
      0
    }
  }, numeric(1)))
  
  memory_mb <- c(
    Y_imputed = y_imputed_size,
    Y_var = if (!is.null(private$.imputation_cache$Y_var)) {
      object.size(private$.imputation_cache$Y_var) / 1024^2
    } else 0
  )
  
  memory_mb["Total"] <- sum(memory_mb)
  
  return(memory_mb)
}

#' Read H5AD file and convert to sparse matrix
#'
#' Reads an H5AD file (AnnData format) and extracts the expression matrix
#' as a sparse dgCMatrix suitable for use with gedi R6 object.
#'
#' @param file_path Character. Path to the H5AD file.
#' @param layer Character. The layer to extract from the H5AD file.
#'   Default is NULL, which reads from X (the main expression matrix).
#'   Common alternatives include "counts", "data", "scaled", etc.
#' @param use_raw Logical. If TRUE, reads from the raw.X slot instead of X.
#'   Default is FALSE.
#' @param transpose Logical. If TRUE, transposes the matrix so genes are rows
#'   and cells are columns (gedi format). Default is TRUE.
#' @param return_metadata Logical. If TRUE, returns a list with the expression
#'   matrix, cell metadata (obs), and gene metadata (var). If FALSE, returns
#'   only the expression matrix. Default is FALSE.
#' @param feature_format Character. Which feature identifier to use for gene names
#'   when return_metadata = FALSE. Options: "gene_name" (default, uses var rownames)
#'   or "gene_ids" (uses var$gene_ids column). Default is "gene_name".
#' @param verbose Logical. If TRUE, prints detailed progress messages. Default is FALSE.
#'
#' @return If return_metadata = FALSE, returns a sparse matrix (dgCMatrix) with
#'   genes as rows and cells as columns. If return_metadata = TRUE, returns a
#'   list with components:
#'   \itemize{
#'     \item X: sparse expression matrix (genes x cells)
#'     \item obs: data.frame of cell metadata
#'     \item var: data.frame of gene metadata
#'   }
#'
#' @details
#' This function reads H5AD files, which are the standard format for AnnData
#' objects in Python. Compatible with gedi R6 class for seamless integration.
#'
#' @examples
#' \donttest{
#' # Round-trip: write a tiny H5AD via write_h5ad(), then read it back.
#' if (requireNamespace("hdf5r", quietly = TRUE) &&
#'     requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'
#'   tmp <- tempfile(fileext = ".h5ad")
#'   write_h5ad(model, tmp)
#'   expr_matrix <- read_h5ad(tmp)
#'   unlink(tmp)
#' }
#' }
#'
#' @importFrom Matrix sparseMatrix t Matrix
#' @export
read_h5ad <- function(file_path,
                      layer = NULL,
                      use_raw = FALSE,
                      transpose = TRUE,
                      return_metadata = FALSE,
                      feature_format = "gene_name",
                      verbose = FALSE) {
  # Validate feature_format early
  if (!feature_format %in% c("gene_name", "gene_ids")) {
    stop("[read_h5ad] ERROR: feature_format must be either 'gene_name' or 'gene_ids'")
  }

  if (verbose) message("[read_h5ad] Starting H5AD file reading process...")

  # Check dependencies
  if (!requireNamespace("hdf5r", quietly = TRUE)) {
    stop(
      "[read_h5ad] ERROR: Package 'hdf5r' is required but not installed.\n",
      "  Install with: install.packages('hdf5r')\n",
      "  Or use: gedi2::install_h5_dependencies()"
    )
  }

  if (!requireNamespace("Matrix", quietly = TRUE)) {
    stop(
      "[read_h5ad] ERROR: Package 'Matrix' is required but not installed.\n",
      "  Install with: install.packages('Matrix')"
    )
  }

  # Validate file path
  if (!file.exists(file_path)) {
    stop("[read_h5ad] ERROR: File not found: ", file_path)
  }

  file_size_mb <- file.info(file_path)$size / 1024^2
  if (verbose) message("[read_h5ad] File size: ", round(file_size_mb, 2), " MB")

  # Open H5AD file with error handling
  if (verbose) message("[read_h5ad] Opening H5AD file...")
  h5file <- tryCatch(
    {
      hdf5r::H5File$new(file_path, mode = "r")
    },
    error = function(e) {
      stop(
        "[read_h5ad] ERROR: Failed to open H5AD file.\n",
        "  File: ", file_path, "\n",
        "  Error: ", conditionMessage(e)
      )
    }
  )

  # Ensure file is closed on exit
  on.exit(
    {
      if (!is.null(h5file)) {
        h5file$close()
        if (verbose) message("[read_h5ad] H5 file closed successfully")
      }
    },
    add = TRUE
  )

  # Determine matrix path
  if (use_raw) {
    matrix_path <- "raw/X"
    var_path <- "raw/var"
    if (verbose) message("[read_h5ad] Reading from: raw.X (raw counts)")
  } else if (!is.null(layer)) {
    matrix_path <- paste0("layers/", layer)
    var_path <- "var"
    if (verbose) message("[read_h5ad] Reading from layer: ", layer)
  } else {
    matrix_path <- "X"
    var_path <- "var"
    if (verbose) message("[read_h5ad] Reading from: X (main expression matrix)")
  }

  # Check if matrix path exists
  if (!h5file$exists(matrix_path)) {
    available_paths <- tryCatch(
      {
        h5file$ls(recursive = TRUE)$name
      },
      error = function(e) {
        character(0)
      }
    )
    stop(
      "[read_h5ad] ERROR: Matrix path '", matrix_path, "' not found in H5AD file.\n",
      "  Available paths (first 20): ", paste(head(available_paths, 20), collapse = ", "), "\n",
      "  Use list_h5_structure('", file_path, "') to explore the file structure."
    )
  }

  # Read expression matrix
  if (verbose) message("[read_h5ad] Reading expression matrix from: ", matrix_path)
  expr_matrix <- tryCatch(
    {
      .read_h5ad_matrix(h5file, matrix_path, verbose)
    },
    error = function(e) {
      stop(
        "[read_h5ad] ERROR: Failed to read expression matrix.\n",
        "  Matrix path: ", matrix_path, "\n",
        "  Error: ", conditionMessage(e)
      )
    }
  )

  if (verbose) {
    message(
      "[read_h5ad] Matrix dimensions (before transpose): ",
      nrow(expr_matrix), " x ", ncol(expr_matrix)
    )
    message("[read_h5ad] Matrix class: ", class(expr_matrix)[1])
    message(
      "[read_h5ad] Sparsity: ",
      round(100 * (1 - Matrix::nnzero(expr_matrix) / (nrow(expr_matrix) * ncol(expr_matrix))), 2), "%"
    )
  }

  # Transpose if needed (H5AD stores as cells x genes, gedi needs genes x cells)
  if (transpose) {
    if (verbose) message("[read_h5ad] Transposing matrix to genes x cells format...")
    expr_matrix <- Matrix::t(expr_matrix)
    if (verbose) message("[read_h5ad] New dimensions: ", nrow(expr_matrix), " genes x ", ncol(expr_matrix), " cells")
  }

  # Read metadata if requested
  if (return_metadata) {
    if (verbose) message("[read_h5ad] Reading metadata...")

    # Read cell metadata (obs)
    obs <- NULL
    if (h5file$exists("obs")) {
      if (verbose) message("[read_h5ad]   Reading cell metadata (obs)...")
      obs <- tryCatch(
        {
          .read_h5ad_dataframe(h5file, "obs", verbose)
        },
        error = function(e) {
          warning("[read_h5ad] WARNING: Failed to read obs metadata: ", conditionMessage(e))
          NULL
        }
      )
      if (!is.null(obs) && verbose) {
        message("[read_h5ad]   Cell metadata: ", nrow(obs), " cells, ", ncol(obs), " columns")
      }
    }

    # Read gene metadata (var)
    var <- NULL
    if (h5file$exists(var_path)) {
      if (verbose) message("[read_h5ad]   Reading gene metadata (var)...")
      var <- tryCatch(
        {
          .read_h5ad_dataframe(h5file, var_path, verbose)
        },
        error = function(e) {
          warning("[read_h5ad] WARNING: Failed to read var metadata: ", conditionMessage(e))
          NULL
        }
      )
      if (!is.null(var) && verbose) {
        message("[read_h5ad]   Gene metadata: ", nrow(var), " genes, ", ncol(var), " columns")
      }
    }

    # Set dimension names
    .set_h5ad_dimnames(expr_matrix, obs, var, verbose)

    if (verbose) {
      message("[read_h5ad] SUCCESS: H5AD file read complete")
      message("[read_h5ad] Final matrix: ", nrow(expr_matrix), " genes x ", ncol(expr_matrix), " cells")
    }

    return(list(
      X = expr_matrix,
      obs = obs,
      var = var
    ))
  } else {
    # Read var to get gene names based on feature_format
    var <- tryCatch(
      {
        .read_h5ad_dataframe(h5file, var_path, verbose = FALSE)
      },
      error = function(e) {
        NULL
      }
    )

    # Set gene names
    if (!is.null(var) && nrow(var) == nrow(expr_matrix)) {
      if (feature_format == "gene_name") {
        rownames(expr_matrix) <- rownames(var)
      } else if (feature_format == "gene_ids" && "gene_ids" %in% colnames(var)) {
        rownames(expr_matrix) <- as.character(var[["gene_ids"]])
      } else {
        rownames(expr_matrix) <- rownames(var)
      }
    }

    # Set cell names from obs
    obs <- tryCatch(
      {
        .read_h5ad_dataframe(h5file, "obs", verbose = FALSE)
      },
      error = function(e) {
        NULL
      }
    )

    if (!is.null(obs) && nrow(obs) == ncol(expr_matrix)) {
      colnames(expr_matrix) <- rownames(obs)
    }

    if (verbose) {
      message("[read_h5ad] SUCCESS: H5AD file read complete")
      message("[read_h5ad] Final matrix: ", nrow(expr_matrix), " genes x ", ncol(expr_matrix), " cells")
    }

    return(expr_matrix)
  }
}


#' Set dimension names for H5AD matrix
#' @keywords internal
#' @noRd
.set_h5ad_dimnames <- function(expr_matrix, obs, var, verbose) {
  # Set row names (genes)
  if (!is.null(var) && nrow(var) == nrow(expr_matrix)) {
    if ("_index" %in% colnames(var)) {
      rownames(expr_matrix) <- as.character(var$`_index`)
      if (verbose) message("[read_h5ad]   Set gene names from var$`_index`")
    } else if (!is.null(rownames(var))) {
      rownames(expr_matrix) <- rownames(var)
      if (verbose) message("[read_h5ad]   Set gene names from var rownames")
    }
  }

  # Set column names (cells)
  if (!is.null(obs) && nrow(obs) == ncol(expr_matrix)) {
    if ("_index" %in% colnames(obs)) {
      colnames(expr_matrix) <- as.character(obs$`_index`)
      if (verbose) message("[read_h5ad]   Set cell names from obs$`_index`")
    } else if (!is.null(rownames(obs))) {
      colnames(expr_matrix) <- rownames(obs)
      if (verbose) message("[read_h5ad]   Set cell names from obs rownames")
    }
  }
}


#' Read expression matrix from H5AD file
#' @keywords internal
#' @noRd
.read_h5ad_matrix <- function(h5file, matrix_path, verbose = TRUE) {
  matrix_group <- h5file[[matrix_path]]

  # Check if sparse matrix (CSR/CSC format)
  if (inherits(matrix_group, "H5Group")) {
    if (matrix_group$exists("data") &&
      matrix_group$exists("indices") &&
      matrix_group$exists("indptr")) {
      if (verbose) message("[read_h5ad]   Matrix format: Sparse (CSR)")

      # Read sparse matrix components
      data <- tryCatch(matrix_group[["data"]]$read(), error = function(e) {
        stop("Failed to read sparse matrix 'data' component: ", conditionMessage(e))
      })
      indices <- tryCatch(matrix_group[["indices"]]$read(), error = function(e) {
        stop("Failed to read sparse matrix 'indices' component: ", conditionMessage(e))
      })
      indptr <- tryCatch(matrix_group[["indptr"]]$read(), error = function(e) {
        stop("Failed to read sparse matrix 'indptr' component: ", conditionMessage(e))
      })

      # Get shape
      shape <- tryCatch(matrix_group$attr_open("shape")$read(), error = function(e) {
        stop("Failed to read matrix shape attribute: ", conditionMessage(e))
      })
      nrows <- shape[1]
      ncols <- shape[2]

      if (verbose) {
        message("[read_h5ad]   Sparse matrix shape: ", nrows, " x ", ncols)
        message("[read_h5ad]   Non-zero elements: ", length(data))
      }

      # Convert from 0-based to 1-based indexing
      indices <- as.integer(indices) + 1L
      indptr <- as.integer(indptr) + 1L

      # Build row indices from indptr (CSR format)
      i_indices <- integer(length(data))
      counter <- 1
      for (row in 1:(length(indptr) - 1)) {
        start_idx <- indptr[row]
        end_idx <- indptr[row + 1] - 1
        if (end_idx >= start_idx) {
          num_elements <- end_idx - start_idx + 1
          i_indices[counter:(counter + num_elements - 1)] <- row
          counter <- counter + num_elements
        }
      }

      # Create sparse matrix
      sparse_mat <- tryCatch(
        {
          Matrix::sparseMatrix(
            i = i_indices,
            j = indices,
            x = as.numeric(data),
            dims = c(nrows, ncols)
          )
        },
        error = function(e) {
          stop("Failed to construct sparse matrix: ", conditionMessage(e))
        }
      )

      return(sparse_mat)
    } else {
      stop("Unknown sparse matrix format. Expected data/indices/indptr components.")
    }
  } else {
    # Dense matrix
    if (verbose) message("[read_h5ad]   Matrix format: Dense")
    dense_mat <- tryCatch(matrix_group$read(), error = function(e) {
      stop("Failed to read dense matrix: ", conditionMessage(e))
    })

    if (verbose) message("[read_h5ad]   Converting to sparse format...")
    return(Matrix::Matrix(dense_mat, sparse = TRUE))
  }
}


#' Read data frame from H5AD file
#' @keywords internal
#' @noRd
.read_h5ad_dataframe <- function(h5file, df_path, verbose = TRUE) {
  if (!h5file$exists(df_path)) {
    if (verbose) message("[read_h5ad]   Path '", df_path, "' not found, skipping")
    return(NULL)
  }

  df_group <- h5file[[df_path]]

  # Read _index (row names)
  row_names <- NULL
  if (df_group$exists("_index")) {
    row_names <- tryCatch(
      {
        as.character(df_group[["_index"]]$read())
      },
      error = function(e) {
        if (verbose) warning("[read_h5ad]   Could not read _index: ", conditionMessage(e))
        NULL
      }
    )
  }

  # Get column names
  all_items <- tryCatch(df_group$ls()$name, error = function(e) character(0))
  col_names <- setdiff(all_items, "_index")

  if (length(col_names) == 0) {
    if (verbose) message("[read_h5ad]   No columns found in ", df_path)
    return(data.frame())
  }

  # Read columns
  df_list <- list()
  for (col_name in col_names) {
    tryCatch(
      {
        col_group <- df_group[[col_name]]

        # Handle categorical data
        if (inherits(col_group, "H5Group") && col_group$exists("categories")) {
          categories <- as.character(col_group[["categories"]]$read())
          codes <- as.integer(col_group[["codes"]]$read())
          df_list[[col_name]] <- factor(categories[codes + 1], levels = categories)
        } else {
          col_data <- col_group$read()
          df_list[[col_name]] <- col_data
        }
      },
      error = function(e) {
        if (verbose) warning("[read_h5ad]   Could not read column '", col_name, "': ", conditionMessage(e))
      }
    )
  }

  # Create data frame
  if (length(df_list) == 0) {
    return(data.frame())
  }

  df <- tryCatch(
    {
      as.data.frame(df_list, stringsAsFactors = FALSE)
    },
    error = function(e) {
      if (verbose) warning("[read_h5ad]   Error creating data frame: ", conditionMessage(e))
      return(data.frame())
    }
  )

  # Set row names
  if (!is.null(row_names) && nrow(df) == length(row_names)) {
    rownames(df) <- row_names
  }

  return(df)
}


#' Read 10X Genomics H5 file
#'
#' Reads a 10X Genomics HDF5 file (CellRanger v2/v3 format) and converts
#' it to a sparse dgCMatrix suitable for use with gedi R6 object.
#'
#' @param file_path Character. Path to the 10X H5 file.
#' @param feature_format Character. Which feature identifier to use for gene names.
#'   Options: "gene_name" (default, uses feature names) or "gene_ids" (uses feature IDs).
#' @param unique.features Logical. Make feature names unique. Default TRUE.
#' @param verbose Logical. Print progress messages. Default FALSE.
#'
#' @return Sparse matrix (dgCMatrix) with genes as rows and cells as columns.
#'
#' @examples
#' \donttest{
#' # read_h5() expects a 10x Genomics filtered_feature_bc_matrix.h5 file.
#' # The example body runs only when such a file is on disk and hdf5r is
#' # installed; otherwise it is silently skipped.
#' h5_file <- "filtered_feature_bc_matrix.h5"
#' if (file.exists(h5_file) &&
#'     requireNamespace("hdf5r", quietly = TRUE)) {
#'   expr_matrix <- read_h5(h5_file)
#'   expr_matrix_ids <- read_h5(h5_file, feature_format = "gene_ids")
#' }
#' }
#'
#' @importFrom Matrix sparseMatrix
#' @export
read_h5 <- function(file_path,
                    feature_format = "gene_name",
                    unique.features = TRUE,
                    verbose = FALSE) {
  if (isFALSE(requireNamespace("hdf5r", quietly = TRUE))) {
    stop("[read_h5] ERROR: Package 'hdf5r' required. Install with: install.packages('hdf5r')")
  }

  if (!file.exists(file_path)) {
    stop("[read_h5] ERROR: File not found: ", file_path)
  }

  # Validate feature_format
  if (!feature_format %in% c("gene_name", "gene_ids")) {
    stop("[read_h5] ERROR: feature_format must be either 'gene_name' or 'gene_ids'")
  }

  if (verbose) message("[read_h5] Opening H5 file...")
  h5file <- hdf5r::H5File$new(filename = file_path, mode = "r")
  on.exit(h5file$close_all(), add = TRUE)

  genomes <- names(h5file)

  # Determine feature slot based on version
  if (hdf5r::existsGroup(h5file, "matrix")) {
    # CellRanger v3+
    genome <- "matrix"
    if (feature_format == "gene_name") {
      feature_slot <- "features/name"
    } else {
      feature_slot <- "features/id"
    }
  } else {
    # CellRanger v2
    genome <- genomes[1]
    if (feature_format == "gene_name") {
      feature_slot <- "gene_names"
    } else {
      feature_slot <- "genes"
    }
  }

  if (verbose) message("[read_h5] Reading sparse matrix components...")

  # Read sparse matrix components
  counts <- h5file[[paste0(genome, "/data")]][]
  indices <- h5file[[paste0(genome, "/indices")]][]
  indptr <- h5file[[paste0(genome, "/indptr")]][]
  shp <- h5file[[paste0(genome, "/shape")]][]
  features <- h5file[[paste0(genome, "/", feature_slot)]][]
  barcodes <- h5file[[paste0(genome, "/barcodes")]][]

  if (verbose) {
    message("[read_h5] Matrix shape: ", shp[1], " x ", shp[2])
    message("[read_h5] Non-zero elements: ", length(counts))
  }

  # Build sparse matrix (CSC format, transpose representation)
  sparse_mat <- Matrix::sparseMatrix(
    i = as.integer(indices) + 1L,
    p = as.integer(indptr),
    x = as.numeric(counts),
    dims = as.integer(shp),
    repr = "T"
  )

  # Set feature names
  if (unique.features) {
    features <- make.unique(names = features)
  }
  rownames(sparse_mat) <- features
  colnames(sparse_mat) <- barcodes

  if (verbose) {
    message("[read_h5] SUCCESS: H5 file read complete")
    message("[read_h5] Final matrix: ", nrow(sparse_mat), " genes x ", ncol(sparse_mat), " cells")
  }

  return(sparse_mat)
}


#' Auto-detect gene and cell names paths in 10X H5 file
#' @keywords internal
#' @noRd
.detect_h5_paths <- function(h5file, verbose = FALSE) {
  # Common paths for gene names in 10X files
  gene_paths <- c(
    "matrix/features/name",
    "features/name",
    "matrix/features/id",
    "features/id",
    "gene_names",
    "genes"
  )

  # Common paths for cell names
  cell_paths <- c(
    "matrix/barcodes",
    "barcodes",
    "cell_names",
    "cells"
  )

  # Find first existing gene path
  gene_names_path <- NULL
  for (path in gene_paths) {
    exists <- tryCatch(h5file$exists(path), error = function(e) FALSE)
    if (exists) {
      gene_names_path <- path
      if (verbose) message("[read_h5] Auto-detected gene names path: ", path)
      break
    }
  }

  # Find first existing cell path
  cell_names_path <- NULL
  for (path in cell_paths) {
    exists <- tryCatch(h5file$exists(path), error = function(e) FALSE)
    if (exists) {
      cell_names_path <- path
      if (verbose) message("[read_h5] Auto-detected cell names path: ", path)
      break
    }
  }

  return(list(
    gene_names_path = gene_names_path,
    cell_names_path = cell_names_path
  ))
}


#' Read matrix from H5 file
#' @keywords internal
#' @noRd
.read_h5_matrix <- function(h5file, matrix_path, as_sparse = TRUE, verbose = TRUE) {
  matrix_group <- h5file[[matrix_path]]

  # Check for 10X sparse matrix format
  if (inherits(matrix_group, "H5Group")) {
    if (matrix_group$exists("data") &&
      matrix_group$exists("indices") &&
      matrix_group$exists("indptr")) {
      if (verbose) message("[read_h5]   Matrix format: 10X Sparse (CSC)")

      # Read components
      data <- tryCatch(matrix_group[["data"]]$read(), error = function(e) {
        stop("Failed to read 'data': ", conditionMessage(e))
      })
      indices <- tryCatch(matrix_group[["indices"]]$read(), error = function(e) {
        stop("Failed to read 'indices': ", conditionMessage(e))
      })
      indptr <- tryCatch(matrix_group[["indptr"]]$read(), error = function(e) {
        stop("Failed to read 'indptr': ", conditionMessage(e))
      })
      shape <- tryCatch(matrix_group[["shape"]]$read(), error = function(e) {
        stop("Failed to read 'shape': ", conditionMessage(e))
      })

      nrows <- shape[1]
      ncols <- shape[2]

      if (verbose) {
        message("[read_h5]   Sparse matrix shape: ", nrows, " x ", ncols)
        message("[read_h5]   Non-zero elements: ", length(data))
      }

      # Convert 0-based to 1-based indexing
      indices <- as.integer(indices) + 1L
      indptr <- as.integer(indptr) + 1L

      # Build column indices from indptr (CSC format)
      j_indices <- integer(length(data))
      counter <- 1
      for (col in 1:(length(indptr) - 1)) {
        start_idx <- indptr[col]
        end_idx <- indptr[col + 1] - 1
        if (end_idx >= start_idx) {
          num_elements <- end_idx - start_idx + 1
          j_indices[counter:(counter + num_elements - 1)] <- col
          counter <- counter + num_elements
        }
      }

      # Create sparse matrix
      sparse_mat <- tryCatch(
        {
          Matrix::sparseMatrix(
            i = indices,
            j = j_indices,
            x = as.numeric(data),
            dims = c(nrows, ncols)
          )
        },
        error = function(e) {
          stop("Failed to construct sparse matrix: ", conditionMessage(e))
        }
      )

      return(sparse_mat)
    } else {
      stop("Unknown H5 matrix format. Expected data/indices/indptr for sparse matrix.")
    }
  } else {
    # Dense matrix
    if (verbose) message("[read_h5]   Matrix format: Dense")
    dense_mat <- tryCatch(matrix_group$read(), error = function(e) {
      stop("Failed to read dense matrix: ", conditionMessage(e))
    })

    if (as_sparse) {
      if (verbose) message("[read_h5]   Converting to sparse format...")
      return(Matrix::Matrix(dense_mat, sparse = TRUE))
    } else {
      return(dense_mat)
    }
  }
}


#' List structure of H5 or H5AD file
#'
#' Helper function to explore H5/H5AD file structure.
#'
#' @param file_path Character. Path to the H5 or H5AD file.
#' @param recursive Logical. List all nested groups. Default TRUE.
#'
#' @return data.frame with file structure information
#'
#' @examples
#' \donttest{
#' # Round-trip: write a tiny H5AD via write_h5ad() then list its structure.
#' if (requireNamespace("hdf5r", quietly = TRUE) &&
#'     requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'   tmp <- tempfile(fileext = ".h5ad")
#'   write_h5ad(model, tmp)
#'   list_h5_structure(tmp)
#'   unlink(tmp)
#' }
#' }
#'
#' @export
list_h5_structure <- function(file_path, recursive = TRUE) {
  if (!requireNamespace("hdf5r", quietly = TRUE)) {
    stop(
      "[list_h5_structure] ERROR: Package 'hdf5r' required.\n",
      "  Install with: install.packages('hdf5r')"
    )
  }

  if (!file.exists(file_path)) {
    stop("[list_h5_structure] ERROR: File not found: ", file_path)
  }

  h5file <- tryCatch(
    {
      hdf5r::H5File$new(file_path, mode = "r")
    },
    error = function(e) {
      stop(
        "[list_h5_structure] ERROR: Failed to open file.\n",
        "  File: ", file_path, "\n",
        "  Error: ", conditionMessage(e)
      )
    }
  )

  on.exit(h5file$close(), add = TRUE)

  structure_info <- tryCatch(
    {
      h5file$ls(recursive = recursive)
    },
    error = function(e) {
      stop(
        "[list_h5_structure] ERROR: Failed to list file contents.\n",
        "  Error: ", conditionMessage(e)
      )
    }
  )

  message("")
  message("Structure of: ", basename(file_path))
  message(paste(rep("=", nchar(basename(file_path)) + 14), collapse = ""))
  message("Full path: ", file_path)
  message(paste(rep("-", nchar(file_path) + 11), collapse = ""), "\n")

  print(structure_info)
  message("")

  return(invisible(structure_info))
}

# ==============================================================================
# GEDI Projections - R Wrappers
# Lazy-computed projection methods with caching
# ==============================================================================


# Projections Accessor R6 Class (for lazy evaluation)
# ==============================================================================

ProjectionsAccessor <- R6Class(
  "ProjectionsAccessor",
  public = list(
    initialize = function(gedi_self, gedi_private) {
      private$.gedi_self <- gedi_self
      private$.gedi_private <- gedi_private
    },
    
    print = function() {
      cat("<GEDI Projections Accessor>\n")
      cat("\nAvailable projections (lazy-computed):\n")
      cat("  $ZDB  - Shared manifold projection (J x N)\n")
      cat("  $DB   - Latent factor embedding (K x N)\n")
      if (private$.gedi_self$aux$P > 0) {
        cat("  $ADB  - Pathway activity projection (P x N)\n")
      }
      cat("\nAccess with: model$projections$ZDB\n")
      invisible(self)
    }
  ),
  private = list(
    .gedi_self = NULL,
    .gedi_private = NULL
  ),
  active = list(
    ZDB = function(value) {
      if (!missing(value)) stop("ZDB is read-only", call. = FALSE)
      compute_ZDB(private$.gedi_self, private$.gedi_private)
    },
    
    DB = function(value) {
      if (!missing(value)) stop("DB is read-only", call. = FALSE)
      compute_DB(private$.gedi_self, private$.gedi_private)
    },
    
    ADB = function(value) {
      if (!missing(value)) stop("ADB is read-only", call. = FALSE)
      if (private$.gedi_self$aux$P > 0) {
        compute_ADB(private$.gedi_self, private$.gedi_private)
      } else {
        NULL
      }
    }
  )
)


#' Compute ZDB Projection with Caching
#'
#' @description
#' Computes the shared manifold projection ZDB = Z * diag(D) * B, where B is
#' the concatenation of all sample-specific Bi matrices. Results are cached
#' automatically for subsequent access.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return Dense matrix (J x N) with row/column names set to geneIDs/cellIDs
#'
#' @keywords internal
#' @noRd
compute_ZDB <- function(self, private, force_recompute = FALSE) {
  
  # Check if cached and not forcing recompute
  if (!force_recompute && !is.null(private$.cache$ZDB)) {
    log_cached("ZDB", private$.verbose)
    return(private$.cache$ZDB)
  }

  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }

  log_start("ZDB projection", private$.verbose)

  # Call C++ function
  ZDB <- compute_ZDB_cpp(
    Z = self$params$Z,
    D = self$params$D,
    Bi_list = self$params$Bi,
    verbose = 0  # C++ silent
  )
  
  # Add row/column names
  rownames(ZDB) <- private$.geneIDs
  colnames(ZDB) <- private$.cellIDs
  
  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$ZDB <- ZDB

  details <- format_dims(nrow(ZDB), ncol(ZDB), "genes", "cells")
  log_complete("ZDB projection", details, private$.verbose)

  return(ZDB)
}


#' Compute DB Projection with Caching
#'
#' @description
#' Computes the latent factor embedding DB = diag(D) * B. This is the
#' K-dimensional representation of cells in latent factor space.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return Dense matrix (K x N) with row names "LV1", "LV2", ... and
#'   column names set to cellIDs
#'
#' @keywords internal
#' @noRd
compute_DB <- function(self, private, force_recompute = FALSE) {
  
  # Check cache
  if (!force_recompute && !is.null(private$.cache$DB)) {
    log_cached("DB", private$.verbose)
    return(private$.cache$DB)
  }

  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }

  log_start("DB projection", private$.verbose)

  # Call C++ function
  DB <- compute_DB_cpp(
    D = self$params$D,
    Bi_list = self$params$Bi,
    verbose = 0  # C++ silent
  )

  # Add row/column names
  rownames(DB) <- paste0("LV", seq_len(self$aux$K))
  colnames(DB) <- private$.cellIDs

  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$DB <- DB

  details <- format_dims(nrow(DB), ncol(DB), "factors", "cells")
  log_complete("DB projection", details, private$.verbose)

  return(DB)
}


#' Compute ADB Projection with Caching
#'
#' @description
#' Computes the pathway activity projection ADB = C.rotation * A * diag(D) * B.
#' Only available when gene-level prior matrix C was provided during model setup.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return Dense matrix (P x N) with row names from original C matrix column
#'   names and column names set to cellIDs
#'
#' @keywords internal
#' @noRd
compute_ADB <- function(self, private, force_recompute = FALSE) {
  
  # Check cache
  if (!force_recompute && !is.null(private$.cache$ADB)) {
    log_cached("ADB", private$.verbose)
    return(private$.cache$ADB)
  }

  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }

  # Check if C prior was provided
  if (self$aux$P == 0) {
    stop("Cannot compute ADB: no gene-level prior (C) was provided during setup.",
         call. = FALSE)
  }

  # Check if aux_static exists
  if (is.null(private$.aux_static)) {
    stop("Missing auxiliary data. Model may not be properly initialized.",
         call. = FALSE)
  }

  log_start("ADB projection", private$.verbose)

  # Call C++ function
  ADB <- compute_ADB_cpp(
    C_rotation = private$.aux_static$C.rotation,
    A = self$params$A,
    D = self$params$D,
    Bi_list = self$params$Bi,
    verbose = 0  # C++ silent
  )

  # Add row/column names
  rownames(ADB) <- colnames(private$.aux_static$inputC)
  colnames(ADB) <- private$.cellIDs

  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$ADB <- ADB

  details <- format_dims(nrow(ADB), ncol(ADB), "pathways", "cells")
  log_complete("ADB projection", details, private$.verbose)

  return(ADB)
}


#' Compute Differential Expression (Cell-Space JxN)
#'
#' @description
#' Computes the cell-specific differential expression effect (J x N).
#' This is the R wrapper for the C++ function getDiffExp_cpp().
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param contrast Numeric vector of length \code{nrow(H)} (the number of
#'   original covariates) specifying the contrast in the user-facing covariate
#'   space; internally projected into the compressed L-space via
#'   \code{H.rotation \%*\% contrast}.
#' @param include_O Logical, if TRUE adds the global offset effect (diffO)
#'
#' @return Dense matrix (J x N) of differential expression values
#'
#' @keywords internal
#' @noRd
compute_diffQ <- function(self, private, contrast, include_O = FALSE) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if H prior was provided
  if (self$aux$L == 0) {
    stop("Cannot compute diffQ: no sample-level prior (H) was provided during setup.",
         call. = FALSE)
  }

  # Check if aux_static exists
  if (is.null(private$.aux_static)) {
    stop("Missing auxiliary data. Model may not be properly initialized.",
         call. = FALSE)
  }

  # Validate contrast: must live in the original covariate space
  # (= ncol(H.rotation) = number of rows of the user-supplied H matrix).
  # H.rotation projects num_covariates -> L, so contrast is num_covariates-long.
  num_cov <- ncol(private$.aux_static$H.rotation)
  if (!is.numeric(contrast) || length(contrast) != num_cov) {
    stop("contrast must be a numeric vector of length ", num_cov,
         " (number of original H covariates)",
         call. = FALSE)
  }

  # Call C++ function (renamed to match old getDiffExp.gedi logic)
  diffExp <- getDiffExp_cpp(
    Rk_list = self$params$Rk,
    H_rotation = private$.aux_static$H.rotation,
    contrast = contrast,
    D = self$params$D,
    Bi_list = self$params$Bi,
    verbose = private$.verbose
  )
  
  # Add diffO if requested
  if (include_O) {
    diffO_vec <- compute_diffO(self, private, contrast) # Calls the wrapper below
    
    # Add diffO to each column (broadcast)
    diffExp <- sweep(diffExp, 1, diffO_vec, "+")
  }
  
  # Add row/column names
  rownames(diffExp) <- private$.geneIDs
  colnames(diffExp) <- private$.cellIDs
  
  return(diffExp)
}


#' Compute Differential O (Global Offset) Effect
#'
#' @description
#' Computes the global differential effect on gene-specific offsets (J x 1).
#' This is the R wrapper for the C++ function getDiffO_cpp().
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param contrast Numeric vector of length \code{nrow(H)} (the number of
#'   original covariates) specifying the contrast in the user-facing covariate
#'   space; internally projected into the compressed L-space via
#'   \code{H.rotation \%*\% contrast}.
#'
#' @return Numeric vector of length J
#'
#' @keywords internal
#' @noRd
compute_diffO <- function(self, private, contrast) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if H prior was provided
  if (self$aux$L == 0) {
    stop("Cannot compute diffO: no sample-level prior (H) was provided during setup.",
         call. = FALSE)
  }

  # Check if aux_static exists
  if (is.null(private$.aux_static)) {
    stop("Missing auxiliary data. Model may not be properly initialized.",
         call. = FALSE)
  }

  # Validate contrast: must live in the original covariate space
  # (= ncol(H.rotation) = number of rows of the user-supplied H matrix).
  num_cov <- ncol(private$.aux_static$H.rotation)
  if (!is.numeric(contrast) || length(contrast) != num_cov) {
    stop("contrast must be a numeric vector of length ", num_cov,
         " (number of original H covariates)",
         call. = FALSE)
  }

  # Call C++ function
  diffO <- getDiffO_cpp(
    Ro = self$params$Ro,
    H_rotation = private$.aux_static$H.rotation,
    contrast = contrast,
    verbose = private$.verbose
  )
  
  # Add names
  names(diffO) <- private$.geneIDs
  
  return(diffO)
}


#' Compute Differential Expression (Alias)
#'
#' @description
#' Convenience alias for compute_diffQ.
#'
#' @keywords internal
#' @noRd
compute_diffExp <- function(self, private, contrast, include_O = FALSE) {
  # This is just an alias for compute_diffQ
  compute_diffQ(self, private, contrast, include_O)
}


#' Clear Projection Cache
#'
#' @description
#' Clears cached projection results.
#'
#' @keywords internal
#' @noRd
clear_projection_cache <- function(private, what = NULL) {
  
  if (is.null(private$.cache)) {
    return(invisible(NULL))
  }
  
  if (is.null(what)) {
    # Clear all cache
    private$.cache <- list()
    if (private$.verbose > 0) {
      message("Projection cache cleared")
    }
  } else {
    # Clear specific entries
    valid_entries <- c("ZDB", "DB", "ADB")
    invalid <- setdiff(what, valid_entries)
    if (length(invalid) > 0) {
      warning("Invalid cache entries: ", paste(invalid, collapse = ", "))
    }
    
    for (entry in intersect(what, valid_entries)) {
      private$.cache[[entry]] <- NULL
      if (private$.verbose > 0) {
        message(sprintf("Cache cleared: %s", entry))
      }
    }
  }
  
  invisible(NULL)
}


#' Get Projection Cache Status
#'
#' @description
#' Returns information about which projections are currently cached.
#'
#' @keywords internal
#' @noRd
get_cache_status <- function(private) {
  
  if (is.null(private$.cache)) {
    return(c(ZDB = FALSE, DB = FALSE, ADB = FALSE))
  }
  
  status <- c(
    ZDB = !is.null(private$.cache$ZDB),
    DB = !is.null(private$.cache$DB),
    ADB = !is.null(private$.cache$ADB)
  )
  
  return(status)
}


#' Get Projection Cache Memory Usage
#'
#' @description
#' Estimates memory usage of cached projections in MB.
#'
#' @keywords internal
#' @noRd
get_cache_memory <- function(private) {
  
  if (is.null(private$.cache)) {
    return(c(ZDB = 0, DB = 0, ADB = 0, Total = 0))
  }
  
  memory_mb <- c(
    ZDB = if (!is.null(private$.cache$ZDB)) {
      object.size(private$.cache$ZDB) / 1024^2
    } else 0,
    DB = if (!is.null(private$.cache$DB)) {
      object.size(private$.cache$DB) / 1024^2
    } else 0,
    ADB = if (!is.null(private$.cache$ADB)) {
      object.size(private$.cache$ADB) / 1024^2
    } else 0
  )
  
  memory_mb["Total"] <- sum(memory_mb)
  
  return(memory_mb)
}

# ==============================================================================
# GEDI v2: Gene Expression Data Integration
# Memory-efficient R6 wrapper around C++ core with projection support
# ==============================================================================

# Source projection functions
# These will be in the same package, so we don't need explicit sourcing

newLogger <- function(name, level = 1) {
  structure(
    list(
      name = name,
      level = level,
      info = function(msg) {
        if (level >= 1) message("[", name, "] ", msg)
      },
      debug = function(msg) {
        if (level >= 2) message("[", name, "] DEBUG: ", msg)
      },
      warn = function(msg) {
        warning("[", name, "] WARNING: ", msg, call. = FALSE)
      },
      error = function(msg) {
        stop("[", name, "] ERROR: ", msg, call. = FALSE)
      }
    ),
    class = "gedi_logger"
  )
}

generateRsvdO <- function(m, n, k, p = 10, sdist = "normal") {
  if (m < n) {
    temp <- m
    m <- n
    n <- temp
  }

  l <- min(k + p, n)

  O <- switch(sdist,
    normal = matrix(stats::rnorm(l * n), n, l),
    unif = matrix(stats::runif(l * n), n, l),
    rademacher = matrix(sample(c(-1, 1), l * n, replace = TRUE), n, l),
    stop("sdist must be: normal, unif, or rademacher", call. = FALSE)
  )

  return(O)
}

GEDI <- R6Class(
  classname = "GEDI",
  private = list(
    .cppPtr = NULL,
    .lastResult = NULL,
    .verbose = 1,
    .logger = NULL,
    .isInitialized = FALSE,

    # Metadata
    .geneIDs = NULL,
    .cellIDs = NULL,
    .sampleNames = NULL,
    .samples = NULL,

    # Static auxiliary data (rotation matrices, priors)
    .aux_static = NULL,

    # M fingerprint for validation
    .M_fingerprint = NULL,

    # Projection cache
    .cache = NULL,

    # Imputation cache
    .imputation_cache = NULL,
    prepareData = function(Samples, Y, X, M, colData, C, H, K, mode, adjustD, orthoZ,
                           Z_shrinkage, A_shrinkage, Qi_shrinkage, Rk_shrinkage,
                           oi_shrinkage, o_shrinkage, si_shrinkage, fixed_si,
                           rsvd_p, rsvd_sdist) {
      if (is.null(M) && is.null(Y) && is.null(X)) {
        private$.logger$error("One of Y, M, or X must be provided")
      }

      # For computing hyperparameters, we need Y_log_dense temporarily
      # But we won't pass it to C++ for M/M_paired/X cases
      Y_log_dense <- NULL
      obs_type <- NULL
      geneIDs <- NULL
      cellIDs <- NULL

      # ===================================================================
      # STEP 1: Determine observation type
      # ===================================================================

      if (!is.null(M)) {
        if (is.list(M)) {
          if (length(M) != 2) {
            private$.logger$error("M list must contain exactly 2 matrices")
          }
          if (!identical(dim(M[[1]]), dim(M[[2]]))) {
            private$.logger$error("M[[1]] and M[[2]] must have identical dimensions")
          }

          M1_dense <- as.matrix(M[[1]])
          M2_dense <- as.matrix(M[[2]])
          Y_log_dense <- log((M1_dense + 1) / (M2_dense + 1))

          obs_type <- "M_paired"
          geneIDs <- rownames(M[[1]])
          cellIDs <- colnames(M[[1]])
        } else {
          Y_log_dense <- as.matrix(log1p(M))

          obs_type <- "M"
          geneIDs <- rownames(M)
          cellIDs <- colnames(M)
        }
      } else if (!is.null(X)) {
        Y_log_dense <- as.matrix(X)
        Y_log_dense[which(X == 1)] <- 1
        Y_log_dense[which(X == 0)] <- -1
        Y_log_dense[which(is.na(X))] <- 0

        obs_type <- "X"
        geneIDs <- rownames(X)
        cellIDs <- colnames(X)
      } else {
        Y_log_dense <- as.matrix(Y)

        obs_type <- "Y"
        geneIDs <- rownames(Y)
        cellIDs <- colnames(Y)
      }

      # ===================================================================
      # STEP 2: Store dimensions
      # ===================================================================

      J <- nrow(Y_log_dense)
      J_vec <- rep(1, J)
      N <- ncol(Y_log_dense)
      N_vec <- rep(1, N)

      if (N != length(Samples)) {
        private$.logger$error("Number of cells doesn't match Samples length")
      }

      # Validate K
      if (!is.numeric(K) || length(K) != 1 || K < 1) {
        private$.logger$error("K must be a positive integer")
      }
      if (K > min(J, N)) {
        private$.logger$error(sprintf("K (%d) must be less than min(n_genes=%d, n_cells=%d)", K, J, N))
      }

      # ===================================================================
      # STEP 3: Process prior matrix C (gene-level)
      # ===================================================================

      inputC <- NULL
      C_matrix <- matrix(nrow = 0, ncol = 0)
      C_rotation <- NULL
      P <- 0

      if (!is.null(C)) {
        inputC <- as.matrix(C)
        if (J != nrow(inputC)) {
          private$.logger$error("C matrix must have same number of rows as data")
        }

        minCL2 <- min(colSums(inputC^2))
        svdC <- svd(inputC)
        P <- sum(cumsum(rev(svdC$d^2)) > minCL2 * 0.1)

        C_matrix <- svdC$u[, 1:P, drop = FALSE]

        # Handle P = 1 edge case where diag() doesn't work as expected
        if (P == 1) {
          C_rotation <- svdC$v[, 1, drop = FALSE] / svdC$d[1]
        } else {
          C_rotation <- svdC$v[, 1:P, drop = FALSE] %*% diag(1 / svdC$d[1:P], nrow = P)
        }

        if (private$.verbose >= 1) {
          private$.logger$info(sprintf(
            "C matrix: %d pathways -> %d components",
            ncol(inputC), P
          ))
        }
      }

      # ===================================================================
      # STEP 4: Store K and related parameters
      # ===================================================================

      diag_K <- diag(K)
      BcolL2 <- 1
      BrowL2 <- 1

      # ===================================================================
      # STEP 5: Organize cells by sample
      # ===================================================================

      unique_samples <- unique(Samples)
      num_samples <- length(unique_samples)

      cells_by_sample <- split(seq_along(Samples), Samples)[unique_samples]
      Ni <- vapply(cells_by_sample, length, integer(1))
      Ni_vec <- lapply(Ni, function(n) rep(1, n))

      cells_all <- unlist(cells_by_sample, use.names = FALSE)
      cellIDs <- cellIDs[cells_all]

      if (!is.null(colData)) {
        colData <- colData[cells_all, , drop = FALSE]
      }

      # ===================================================================
      # STEP 6: Process prior matrix H (sample-level)
      # ===================================================================

      inputH <- NULL
      H_matrix <- matrix(nrow = 0, ncol = 0)
      H_rotation <- NULL
      L <- 0

      if (!is.null(H)) {
        inputH <- as.matrix(H)
        inputH <- inputH[, match(unique_samples, colnames(inputH)), drop = FALSE]

        if (ncol(inputH) != num_samples) {
          private$.logger$error("H matrix must have columns for all samples")
        }

        minHL2 <- min(rowSums(inputH^2))
        svdH <- svd(t(inputH))
        L <- sum(cumsum(rev(svdH$d^2)) > minHL2 * 0.1)

        H_matrix <- t(svdH$u[, 1:L, drop = FALSE])
        H_rotation <- t(svdH$v[, 1:L, drop = FALSE] %*% diag(1 / svdH$d[1:L], nrow = L))

        if (private$.verbose >= 1) {
          private$.logger$info(sprintf(
            "H matrix: %d covariates -> %d components",
            nrow(inputH), L
          ))
        }
      }

      # ===================================================================
      # STEP 7: Initialize hyperparameters (compute s_0 and o_0)
      # ===================================================================

      s_0 <- NULL
      o_0 <- NULL

      if (is.null(M)) {
        if (is.null(X)) {
          # Y case - use dense functions
          s_0 <- compute_s_0_dense(J_vec, Y_log_dense, J)
          Yp <- compute_Yp_dense(Y_log_dense, J_vec, s_0)
          o_0 <- compute_o_0_dense(Yp, N_vec, N)
        } else {
          # X case
          s_0 <- rep(0, N)
          o_0 <- rep(0, J)
        }
      } else if (is.list(M)) {
        # Paired M case - use sparse functions
        s1_0 <- compute_s_0(J_vec, M[[1]], J)
        Mp1 <- compute_Mp(M[[1]], J_vec, s1_0)
        o1_0 <- compute_o_0(Mp1, N_vec, N)

        s2_0 <- compute_s_0(J_vec, M[[2]], J)
        Mp2 <- compute_Mp(M[[2]], J_vec, s2_0)
        o2_0 <- compute_o_0(Mp2, N_vec, N)

        s_0 <- log(s1_0 / s2_0)
        o_0 <- log(o1_0 / o2_0)
      } else {
        # Single M case - use sparse functions
        s_0_counts <- compute_s_0(J_vec, M, J)
        Mp <- compute_Mp(M, J_vec, s_0_counts)
        o_0_counts <- compute_o_0(Mp, N_vec, N)

        s_0 <- log(s_0_counts)
        o_0 <- log(o_0_counts)
      }

      # Split s_0 by sample
      si_0 <- split(s_0, Samples)[unique_samples]
      if (!is.na(fixed_si)) {
        si_0 <- lapply(Ni, function(n) rep(fixed_si, n))
      }
      is_si_fixed <- !is.na(fixed_si)

      # Initialize sigma2_0
      if (obs_type != "X") {
        Yp_full <- Y_log_dense - VecVecProduct(J_vec, s_0) - o_0
        # Fix integer overflow: convert N and J to numeric before multiplication
        sigma2_0 <- sum(Yp_full^2) / (as.numeric(N) * as.numeric(J) + 1)
      } else {
        sigma2_0 <- 1
      }

      # Other hyperparameters
      S_o <- 1 / N / o_shrinkage
      S_si <- 1 / J / si_shrinkage
      S_Z <- 1 / Z_shrinkage
      S_A <- 1 / A_shrinkage
      S_R <- 1 / Rk_shrinkage
      S_oi <- 1 / Ni / oi_shrinkage
      S_oi_mean <- num_samples / N / oi_shrinkage
      S_Qi <- N / Ni / Qi_shrinkage
      S_Qi_mean <- num_samples / Qi_shrinkage

      # Generate O matrix for randomized SVD initialization
      O_matrix <- generateRsvdO(m = J, n = N, k = K, p = rsvd_p, sdist = rsvd_sdist)

      if (private$.verbose >= 1) {
        private$.logger$info(sprintf("Initial sigma2 = %.6f", sigma2_0))
      }

      # ===================================================================
      # STEP 8: Initialize D based on mode
      # ===================================================================

      if (mode == "Bl2") {
        D_init <- rep(BrowL2, K)
      } else if (mode == "Bsphere") {
        D_init <- sqrt(BrowL2) * rep(sqrt(K / N), K)
      } else {
        private$.logger$error("mode must be 'Bl2' or 'Bsphere'")
      }

      # ===================================================================
      # STEP 9: Initialize model parameters
      # ===================================================================

      params <- list(
        Bi = lapply(Ni, function(n) matrix(0, nrow = K, ncol = n)),
        Qi = lapply(1:num_samples, function(i) matrix(0, nrow = J, ncol = K)),
        si = si_0,
        oi = lapply(1:num_samples, function(i) rep(0, J)),
        o = o_0,
        Z = matrix(0, nrow = J, ncol = K),
        U = matrix(0, nrow = J, ncol = K),
        S = rep(1, K),
        D = D_init,
        sigma2 = sigma2_0,
        A = if (P > 0) matrix(0, nrow = P, ncol = K) else matrix(nrow = 0, ncol = 0),
        Rk = if (L > 0) lapply(1:K, function(k) matrix(0, nrow = J, ncol = L)) else list(),
        Ro = if (L > 0) matrix(0, nrow = J, ncol = L) else matrix(nrow = 0, ncol = 0)
      )

      # ===================================================================
      # STEP 10: Initialize auxiliary variables
      # ===================================================================

      aux <- list(
        # DBi optimization: Store smaller K x Ni instead of J x Ni matrices
        DBi = lapply(Ni, function(n) matrix(0, nrow = K, ncol = n)),
        Qi_hat = if (L > 0) lapply(1:num_samples, function(i) matrix(0, nrow = J, ncol = K)) else list(),
        oi_hat = if (L > 0) lapply(1:num_samples, function(i) rep(0, J)) else list(),
        J = J,
        N = N,
        K = K,
        P = P,
        L = L,
        numSamples = num_samples,
        diag_K = diag_K,
        J_vec = J_vec,
        Ni = Ni,
        Ni_vec = Ni_vec,
        obs.type = obs_type,
        mode = mode,
        orthoZ = orthoZ,
        adjustD = adjustD,
        is_si_fixed = is_si_fixed,
        C = C_matrix,
        H = H_matrix,
        inputC = inputC,
        C.rotation = C_rotation,
        inputH = inputH,
        H.rotation = H_rotation,
        BcolL2 = BcolL2,
        BrowL2 = BrowL2,
        colData = colData,
        ite = 0
      )

      # ===================================================================
      # STEP 10.5: Store static auxiliary data for projections
      # ===================================================================

      private$.aux_static <- list(
        C.rotation = C_rotation,
        H.rotation = H_rotation,
        inputC = inputC,
        inputH = inputH,
        colData = colData
      )

      # ===================================================================
      # STEP 11: Prepare target matrices
      # OPTION 2: Don't compute Yi in R - C++ will compute it from M!
      # ===================================================================

      if (private$.verbose >= 1) {
        private$.logger$info("Preparing target matrices (C++ will compute Yi)...")
      }

      target <- list()

      if (obs_type == "M") {
        # Pass ONLY raw sparse M - C++ will compute Yi = log1p(M)
        target$Mi <- lapply(cells_by_sample, function(idx) M[, idx, drop = FALSE])
        target$Yi <- vector("list", num_samples) # Empty - C++ fills
        target$M1i <- vector("list", num_samples)
        target$M2i <- vector("list", num_samples)
        target$Xi <- vector("list", num_samples)
      } else if (obs_type == "M_paired") {
        # Pass ONLY raw sparse M1 and M2 - C++ computes Yi = log((M1+1)/(M2+1))
        target$M1i <- lapply(cells_by_sample, function(idx) M[[1]][, idx, drop = FALSE])
        target$M2i <- lapply(cells_by_sample, function(idx) M[[2]][, idx, drop = FALSE])
        target$Yi <- vector("list", num_samples) # Empty - C++ fills
        target$Mi <- vector("list", num_samples)
        target$Xi <- vector("list", num_samples)
      } else if (obs_type == "X") {
        # Pass X - C++ converts to Yi format
        target$Xi <- lapply(cells_by_sample, function(idx) X[, idx, drop = FALSE])
        target$Yi <- vector("list", num_samples) # Empty - C++ fills
        target$Mi <- vector("list", num_samples)
        target$M1i <- vector("list", num_samples)
        target$M2i <- vector("list", num_samples)
      } else {
        # Y type - must compute Yi in R (no M available)
        target$Yi <- lapply(cells_by_sample, function(idx) as.matrix(Y[, idx, drop = FALSE]))
        target$Mi <- vector("list", num_samples)
        target$M1i <- vector("list", num_samples)
        target$M2i <- vector("list", num_samples)
        target$Xi <- vector("list", num_samples)
      }

      # Force garbage collection - Y_log_dense no longer needed
      rm(Y_log_dense)
      invisible(gc())

      # ===================================================================
      # STEP 12: Create hyperparams list
      # ===================================================================

      hyperparams <- list(
        S_Qi = S_Qi,
        S_oi = S_oi,
        S_Z = S_Z,
        S_A = S_A,
        S_R = S_R,
        S_Qi_mean = S_Qi_mean,
        S_oi_mean = S_oi_mean,
        S_si = S_si,
        S_o = S_o,
        o_0 = o_0,
        si_0 = si_0,
        O = O_matrix
      )

      return(list(
        params = params,
        aux = aux,
        target = target,
        hyperparams = hyperparams,
        geneIDs = geneIDs,
        cellIDs = cellIDs,
        sampleNames = unique_samples
      ))
    }
  ),
  public = list(
    setup = function(Samples,
                     M = NULL,
                     Y = NULL,
                     X = NULL,
                     colData = NULL,
                     C = NULL,
                     H = NULL,
                     K = 10,
                     mode = "Bl2",
                     adjustD = TRUE,
                     orthoZ = TRUE,
                     Z_shrinkage = 1,
                     A_shrinkage = 1,
                     Qi_shrinkage = 1,
                     Rk_shrinkage = 1,
                     oi_shrinkage = 1,
                     o_shrinkage = 1,
                     si_shrinkage = 1,
                     fixed_si = NA,
                     rsvd_p = 10,
                     rsvd_sdist = "normal",
                     verbose = 1,
                     num_threads = 0) {
      private$.verbose <- verbose
      private$.logger <- newLogger("GEDI", level = verbose)

      private$.logger$info("Setting up GEDI model...")

      data <- private$prepareData(
        Samples = Samples, Y = Y, X = X, M = M,
        colData = colData, C = C, H = H, K = K,
        mode = mode, adjustD = adjustD, orthoZ = orthoZ,
        Z_shrinkage = Z_shrinkage, A_shrinkage = A_shrinkage,
        Qi_shrinkage = Qi_shrinkage, Rk_shrinkage = Rk_shrinkage,
        oi_shrinkage = oi_shrinkage, o_shrinkage = o_shrinkage,
        si_shrinkage = si_shrinkage, fixed_si = fixed_si,
        rsvd_p = rsvd_p, rsvd_sdist = rsvd_sdist
      )

      private$.geneIDs <- data$geneIDs
      private$.cellIDs <- data$cellIDs
      private$.sampleNames <- data$sampleNames

      # Store Samples vector (in original order)
      private$.samples <- Samples

      # Compute and store M fingerprint
      if (!is.null(M)) {
        private$.M_fingerprint <- compute_M_fingerprint(
          M = M,
          obs_type = data$aux$obs.type
        )

        if (private$.verbose >= 1) {
          private$.logger$info("M fingerprint computed")
        }
      }

      # Create C++ object (C++ will compute Yi from M)
      private$.cppPtr <- GEDI_new(
        params = data$params,
        aux = data$aux,
        target = data$target,
        hyperparams = data$hyperparams,
        verbose = private$.verbose,
        num_threads = num_threads
      )

      # CRITICAL: Force cleanup of R-side data to save memory
      data$target <- NULL
      data$params$Bi <- NULL
      data$params$Qi <- NULL
      data$aux$DBi <- NULL # DBi optimization: cleanup DBi instead of ZDBi/QiDBi
      rm(data)
      gc(full = TRUE)

      private$.logger$info("Setup complete.")

      invisible(self)
    },
    initialize_lvs = function(multimodal = FALSE) {
      if (is.null(private$.cppPtr)) {
        stop("Model not setup. Call CreateGEDIObject() first.", call. = FALSE)
      }

      private$.logger$info("Initializing latent variables...")

      private$.lastResult <- GEDI_initialize(private$.cppPtr, multimodal = multimodal)
      private$.isInitialized <- TRUE

      private$.logger$info("Initialization complete.")

      invisible(self)
    },
    optimize = function(iterations = 50, track_interval = 5) {
      if (!private$.isInitialized) {
        stop("Model not initialized. Call $initialize_lvs() first.", call. = FALSE)
      }

      if (private$.verbose == 1) {
        GEDI_set_verbose(private$.cppPtr, 0L)
        on.exit(GEDI_set_verbose(private$.cppPtr, 1L), add = TRUE)
      }

      private$.logger$info(sprintf("Running optimization (%d iterations)...", iterations))

      private$.lastResult <- GEDI_optimize(
        model_ptr = private$.cppPtr,
        iterations = iterations,
        track_interval = track_interval
      )

      private$.logger$info("Optimization complete.")

      # Clear caches after optimization
      clear_projection_cache(private)
      clear_dimred_cache(private)
      clear_imputation_cache(private)

      invisible(self)
    },
    train = function(iterations = 50, track_interval = 5, multimodal = FALSE) {
      if (is.null(private$.cppPtr)) {
        stop("Model not setup. Call CreateGEDIObject() first.", call. = FALSE)
      }

      if (private$.verbose == 1) {
        # Suppress C++ per-iteration output; show a single R-side message instead
        GEDI_set_verbose(private$.cppPtr, 0L)
        on.exit(GEDI_set_verbose(private$.cppPtr, 1L), add = TRUE)
        message(sprintf("Training GEDI model (%d iterations)...", iterations))
      }

      # Single call — no split, no double data transfer
      private$.lastResult <- GEDI_train(
        model_ptr = private$.cppPtr,
        iterations = iterations,
        track_interval = track_interval,
        multimodal = multimodal
      )

      private$.isInitialized <- TRUE

      private$.logger$info("Training complete.")

      # Clear caches after training
      clear_projection_cache(private)
      clear_dimred_cache(private)
      clear_imputation_cache(private)

      invisible(self)
    },

    # =========================================================================
    # Cache Management Methods
    # =========================================================================

    clear_cache = function(what = NULL) {
      # Clear projection cache
      clear_projection_cache(private, what)

      # Clear dimred cache if "svd" or "pca" specified, or if clearing all
      if (is.null(what) || any(what %in% c("svd", "pca"))) {
        dimred_what <- if (is.null(what)) NULL else intersect(what, c("svd", "pca"))
        clear_dimred_cache(private, dimred_what)
      }

      # Clear imputation cache if specified, or if clearing all
      if (is.null(what) || any(what %in% c("Y_fitted", "Y_imputed", "Y_var"))) {
        imputation_what <- if (is.null(what)) NULL else intersect(what, c("Y_fitted", "Y_imputed", "Y_var"))
        clear_imputation_cache(private, imputation_what)
      }

      invisible(self)
    },
    cache_status = function() {
      proj_status <- get_cache_status(private)
      dimred_status <- get_dimred_cache_status(private)
      imputation_status <- get_imputation_cache_status(private)
      c(proj_status, dimred_status, imputation_status)
    },
    cache_memory = function() {
      proj_memory <- get_cache_memory(private)
      dimred_memory <- get_dimred_cache_memory(private)
      imputation_memory <- get_imputation_cache_memory(private)

      # Combine and recalculate total
      all_memory <- c(
        proj_memory[names(proj_memory) != "Total"],
        dimred_memory[names(dimred_memory) != "Total"],
        imputation_memory[names(imputation_memory) != "Total"]
      )
      all_memory["Total"] <- sum(all_memory)

      return(all_memory)
    },
    set_verbose = function(level = 1) {
      if (!level %in% c(0, 1, 2)) {
        stop("verbose must be 0 (silent), 1 (normal), or 2 (debug)", call. = FALSE)
      }
      private$.verbose <- level
      if (!is.null(private$.logger)) {
        private$.logger$level <- level
      }
      invisible(self)
    },

    # =========================================================================
    # Differential Expression Methods
    # =========================================================================

    diffQ = function(contrast, include_O = FALSE) {
      compute_diffQ(self, private, contrast, include_O)
    },
    diffO = function(contrast) {
      compute_diffO(self, private, contrast)
    },
    diffExp = function(contrast, include_O = FALSE) {
      compute_diffExp(self, private, contrast, include_O)
    },

    # =========================================================================
    # Dimensionality Reduction Methods
    # =========================================================================

    get_svd = function() {
      compute_svd_factorized(self, private)
    },
    get_pca = function() {
      compute_pca(self, private)
    },
    get_umap = function(input = "pca",
                        n_neighbors = 15,
                        min_dist = 0.1,
                        n_components = 2,
                        metric = "euclidean",
                        n_threads = 0,
                        ...) {
      compute_umap(
        self, private, input, n_neighbors,
        min_dist, n_components, metric, n_threads, ...
      )
    },

    # =========================================================================
    # Export Methods
    # =========================================================================

    save_h5ad = function(file_path,
                         X_slot = c("imputed", "projection", "original"),
                         M = NULL,
                         include_embeddings = TRUE,
                         include_raw = FALSE,
                         compression = 6,
                         verbose = TRUE) {
      write_h5ad(
        model = self,
        file_path = file_path,
        X_slot = X_slot,
        M = M,
        include_embeddings = include_embeddings,
        include_raw = include_raw,
        compression = compression,
        verbose = verbose
      )
    },


    # =========================================================================
    # Print Method
    # =========================================================================

    print = function() {
      cat("<GEDI Model>\n")

      if (is.null(private$.cppPtr)) {
        cat("Status: Not setup\n")
        return(invisible(self))
      }

      if (!is.null(private$.lastResult)) {
        aux <- private$.lastResult$aux
        cat(sprintf("Dimensions: %d genes x %d cells\n", aux$J, aux$N))
        cat(sprintf(
          "Samples: %d (%s)\n",
          aux$numSamples,
          paste(head(private$.sampleNames, 3), collapse = ", ")
        ))
        if (aux$numSamples > 3) cat("           ...\n")
        cat(sprintf("Latent factors: K = %d\n", aux$K))
        cat(sprintf("Mode: %s (orthoZ = %s)\n", aux$mode, aux$orthoZ))

        if (aux$P > 0) cat(sprintf("Prior C: %d pathways\n", aux$P))
        if (aux$L > 0) cat(sprintf("Prior H: %d covariates\n", aux$L))

        # Cache info
        proj_cache <- get_cache_status(private)
        dimred_cache <- get_dimred_cache_status(private)
        imputation_cache <- get_imputation_cache_status(private)
        all_cached <- c(proj_cache, dimred_cache, imputation_cache)
        if (any(all_cached)) {
          cached_names <- names(all_cached)[all_cached]
          cat(sprintf("Cached: %s\n", paste(cached_names, collapse = ", ")))
        }
      }

      if (private$.isInitialized) {
        sigma2 <- private$.lastResult$params$sigma2
        cat(sprintf("\nStatus: Fitted (sigma2 = %.6f)\n", sigma2))
      } else {
        cat("\nStatus: Setup complete (not yet initialized)\n")
      }

      invisible(self)
    }
  ),
  active = list(
    params = function(value) {
      if (!missing(value)) stop("params is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }
      private$.lastResult$params
    },
    aux = function(value) {
      if (!missing(value)) stop("aux is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }
      # Merge with geneIDs/cellIDs for backward compatibility
      c(private$.lastResult$aux, list(
        geneIDs = private$.geneIDs,
        cellIDs = private$.cellIDs
      ))
    },
    hyperparams = function(value) {
      if (!missing(value)) stop("hyperparams is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }
      private$.lastResult$hyperparams
    },
    tracking = function(value) {
      if (!missing(value)) stop("tracking is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $optimize() or $train() first.", call. = FALSE)
      }
      private$.lastResult$tracking
    },
    Z = function(value) {
      if (!missing(value)) stop("Z is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }
      private$.lastResult$params$Z
    },

    # =========================================================================
    # Metadata Accessor
    # =========================================================================

    metadata = function(value) {
      if (!missing(value)) stop("metadata is read-only", call. = FALSE)
      if (is.null(private$.geneIDs)) {
        stop("Model not setup. Call CreateGEDIObject() first.", call. = FALSE)
      }

      list(
        geneIDs = private$.geneIDs,
        cellIDs = private$.cellIDs,
        sampleIDs = private$.sampleNames,
        Samples = private$.samples,
        colData = if (!is.null(private$.aux_static)) private$.aux_static$colData else NULL,
        n_genes = if (!is.null(private$.lastResult)) private$.lastResult$aux$J else NULL,
        n_cells = if (!is.null(private$.lastResult)) private$.lastResult$aux$N else NULL,
        n_samples = if (!is.null(private$.lastResult)) private$.lastResult$aux$numSamples else NULL
      )
    },

    # =========================================================================
    # Priors Accessor
    # =========================================================================

    priors = function(value) {
      if (!missing(value)) stop("priors is read-only", call. = FALSE)
      if (is.null(private$.aux_static)) {
        stop("Model not setup. Call CreateGEDIObject() first.", call. = FALSE)
      }

      list(
        inputC = private$.aux_static$inputC,
        C.rotation = private$.aux_static$C.rotation,
        inputH = private$.aux_static$inputH,
        H.rotation = private$.aux_static$H.rotation
      )
    },

    # =========================================================================
    # Projections Accessor (Lazy Computation)
    # =========================================================================

    projections = function(value) {
      if (!missing(value)) stop("projections is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }

      # Return R6 accessor with lazy active bindings
      ProjectionsAccessor$new(self, private)
    },

    # =========================================================================
    # Embeddings Accessor (Lazy Computation)
    # =========================================================================

    embeddings = function(value) {
      if (!missing(value)) stop("embeddings is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }

      # Return R6 accessor with lazy active bindings
      EmbeddingsAccessor$new(self, private)
    },

    # =========================================================================
    # Pathway Associations Accessor
    # =========================================================================

    pathway_associations = function(value) {
      if (!missing(value)) stop("pathway_associations is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }

      # Return accessor object with dense() and sparse() methods
      create_pathway_associations_accessor(self, private)
    },

    # =========================================================================
    # Dynamics Analysis Accessor
    # =========================================================================

    dynamics = function(value) {
      if (!missing(value)) stop("dynamics is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }

      # Return accessor object for vector field and gradient analysis
      create_dynamics_accessor(self, private)
    },

    # =========================================================================
    # Imputation Accessor
    # =========================================================================

    imputed = function(value) {
      if (!missing(value)) stop("imputed is read-only", call. = FALSE)
      if (is.null(private$.lastResult)) {
        stop("No results yet. Run $initialize_lvs() or $train() first.", call. = FALSE)
      }

      # Return accessor object for imputation analysis
      create_imputation_accessor(self, private)
    }
  )
)

#' Create GEDI Object
#'
#' @description
#' Creates and configures a GEDI (Gene Expression Data Integration) model object.
#' This implementation uses Option 2 memory optimization: C++ computes Yi from M,
#' eliminating duplicate Yi storage in R.
#'
#' @param Samples Factor or character vector indicating sample of origin for each cell
#' @param M Raw count matrix (sparse or dense), or list of two matrices for paired data.
#'   C++ will compute Yi = log(M+1) internally - no Yi copy stored in R!
#' @param Y Log-transformed expression matrix (optional if M provided)
#' @param X Binary indicator matrix (optional if M or Y provided)
#' @param colData Optional data.frame with cell metadata
#' @param C Gene-level prior matrix (genes x pathways)
#' @param H Sample-level covariate matrix (covariates x samples)
#' @param K Number of latent factors (default: 10)
#' @param mode Normalization mode: "Bl2" or "Bsphere" (default: "Bl2")
#' @param adjustD Whether to adjust D based on B row norms (default: TRUE)
#' @param orthoZ Whether Z columns should be orthogonal (default: TRUE)
#' @param Z_shrinkage,A_shrinkage,Qi_shrinkage,Rk_shrinkage,oi_shrinkage,o_shrinkage,si_shrinkage
#'   Regularization strengths (default: 1)
#' @param fixed_si Fix cell library sizes at this value, or NA to optimize (default: NA)
#' @param rsvd_p Oversampling parameter for randomized SVD (default: 10)
#' @param rsvd_sdist Random distribution for rSVD: "normal", "unif", or "rademacher"
#'   (default: "normal")
#' @param verbose Verbosity level: 0 (silent), 1 (info), 2 (debug) (default: 1)
#' @param num_threads Number of OpenMP threads (default: 0 = auto)
#'
#' @return GEDI R6 object with memory-efficient architecture
#'
#' @examples
#' \donttest{
#' # Load example data
#' pbmc_small <- SeuratObject::pbmc_small
#'
#' # Basic usage - memory efficient!
#' model <- CreateGEDIObject(
#'   Samples = pbmc_small@meta.data$orig.ident,
#'   M = pbmc_small@assays$RNA@counts, # Only M stored in R; Yi computed in C++
#'   K = 10,
#'   num_threads = 1
#' )
#'
#' # Train the model
#' model$train(iterations = 50)
#'
#' # Access results via active bindings
#' Z <- model$Z
#' params <- model$params
#'
#' # Access projections (computed on demand, cached)
#' zdb <- model$projections$ZDB
#' db <- model$projections$DB
#'
#' # Access imputed expression
#' Y_imputed <- model$imputed$Y() # No M needed!
#' Y_var <- model$imputed$variance(pbmc_small@assays$RNA@counts) # M required
#' }
#'
#' @export
CreateGEDIObject <- function(
  Samples,
  M = NULL,
  Y = NULL,
  X = NULL,
  colData = NULL,
  C = NULL,
  H = NULL,
  K = 40,
  mode = "Bsphere",
  adjustD = TRUE,
  orthoZ = TRUE,
  Z_shrinkage = 1,
  A_shrinkage = 1,
  Qi_shrinkage = 1,
  Rk_shrinkage = 1,
  oi_shrinkage = 1,
  o_shrinkage = 1,
  si_shrinkage = 1,
  fixed_si = NA,
  rsvd_p = 10,
  rsvd_sdist = "normal",
  verbose = 1,
  num_threads = 1
) {
  model <- GEDI$new()

  model$setup(
    Samples = Samples,
    M = M, Y = Y, X = X,
    colData = colData,
    C = C, H = H,
    K = K,
    mode = mode,
    adjustD = adjustD,
    orthoZ = orthoZ,
    Z_shrinkage = Z_shrinkage,
    A_shrinkage = A_shrinkage,
    Qi_shrinkage = Qi_shrinkage,
    Rk_shrinkage = Rk_shrinkage,
    oi_shrinkage = oi_shrinkage,
    o_shrinkage = o_shrinkage,
    si_shrinkage = si_shrinkage,
    fixed_si = fixed_si,
    rsvd_p = rsvd_p,
    rsvd_sdist = rsvd_sdist,
    verbose = verbose,
    num_threads = num_threads
  )

  return(model)
}

# tests/testthat/test-diff-contrast-length.R
# -----------------------------------------------------------------------------
# Regression test for the contrast-length bug in the differential expression
# API (model$diffO / $diffQ / $diffExp).
#
# Before the fix:
#   - R guards (R/gedi_projections.R::compute_diffQ / compute_diffO) and the
#     C++ guards (src/gedi_differential.cpp::getDiff{O,Q,Exp}_cpp) all enforced
#     length(contrast) == L.
#   - The C++ formula computes Ro %*% H_rotation %*% contrast, where
#     H_rotation is L x num_covariates. So contrast must be num_covariates-long.
#   - Whenever L < num_covariates (rank-deficient H), the only call path the
#     guards permitted triggered an Eigen out-of-bounds read (silent UB).
#
# After the fix:
#   - Both guards check length(contrast) == ncol(H.rotation) (= num_covariates).
#   - A length-num_covariates contrast is accepted; a length-L contrast is
#     rejected with a clear error (when L < num_covariates).
#   - The returned values match a pure-R replica of the documented formula.
# -----------------------------------------------------------------------------

# Fixture: 4 samples, 3 covariates that are linearly dependent
#   cov3 = cov1 + cov2  =>  rank(H^T) = 2  =>  L = 2 < num_covariates = 3
make_diff_fixture <- function() {
  set.seed(1234)
  n_genes   <- 60
  n_cells   <- 80
  n_samples <- 4

  M <- Matrix::Matrix(
    matrix(stats::rpois(n_genes * n_cells, lambda = 5), n_genes, n_cells),
    sparse = TRUE
  )
  rownames(M) <- paste0("Gene_", seq_len(n_genes))
  colnames(M) <- paste0("Cell_", seq_len(n_cells))
  samples <- factor(rep(paste0("S", seq_len(n_samples)), each = n_cells / n_samples))

  H <- matrix(
    c(1, 1, 0, 0,
      0, 0, 1, 1,
      1, 1, 1, 1),
    nrow = 3, byrow = TRUE
  )
  rownames(H) <- c("batch", "condition", "combined")
  colnames(H) <- paste0("S", seq_len(n_samples))

  model <- CreateGEDIObject(
    Samples = samples, M = M, K = 3, H = H, verbose = 0
  )
  model$train(iterations = 10, track_interval = 5)

  list(model = model, H = H, num_cov = nrow(H))
}

# -----------------------------------------------------------------------------
test_that("diffO/diffQ accept length-num_covariates contrast (rank-deficient H)", {
  fx <- make_diff_fixture()
  m  <- fx$model

  # Sanity: fixture really has L < num_covariates
  expect_lt(m$aux$L, fx$num_cov)
  expect_equal(ncol(m$priors$H.rotation), fx$num_cov)

  # length-num_covariates contrast must be accepted by all three methods
  contrast_full <- c(batch = 1, condition = -1, combined = 0)
  expect_no_error(diffO_v   <- m$diffO(contrast_full))
  expect_no_error(diffQ_m   <- m$diffQ(contrast_full))
  expect_no_error(diffExp_m <- m$diffExp(contrast_full))

  expect_length(diffO_v, length(m$metadata$geneIDs))
  expect_equal(dim(diffQ_m),   c(length(m$metadata$geneIDs),
                                 length(m$metadata$cellIDs)))
  expect_equal(dim(diffExp_m), c(length(m$metadata$geneIDs),
                                 length(m$metadata$cellIDs)))
})

# -----------------------------------------------------------------------------
test_that("diffO/diffQ reject length-L contrast when L < num_covariates", {
  fx <- make_diff_fixture()
  m  <- fx$model

  # length-L contrast is the OLD (buggy) contract; must now be rejected with
  # a message that mentions the correct expected length.
  short <- rep(1, m$aux$L)
  expect_error(m$diffO(short),   regexp = as.character(fx$num_cov))
  expect_error(m$diffQ(short),   regexp = as.character(fx$num_cov))
  expect_error(m$diffExp(short), regexp = as.character(fx$num_cov))
})

# -----------------------------------------------------------------------------
test_that("diffO matches the documented formula Ro %*% H.rotation %*% contrast", {
  fx <- make_diff_fixture()
  m  <- fx$model

  contrast <- c(0.5, -0.3, 0.7)
  Ro       <- m$params$Ro            # J x L
  H_rot    <- m$priors$H.rotation    # L x num_cov

  expected <- as.vector(Ro %*% H_rot %*% contrast)
  observed <- as.vector(m$diffO(contrast))

  expect_equal(observed, expected, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
test_that("diffQ matches sum_k Rk %*% H.rotation %*% contrast for each gene-factor pair", {
  fx <- make_diff_fixture()
  m  <- fx$model

  contrast <- c(0.5, -0.3, 0.7)
  H_rot    <- m$priors$H.rotation    # L x num_cov
  H_c      <- as.vector(H_rot %*% contrast)   # length L
  Rk_list  <- m$params$Rk
  D        <- m$params$D
  Bi_list  <- m$params$Bi
  K        <- m$aux$K
  J        <- length(m$metadata$geneIDs)
  N        <- length(m$metadata$cellIDs)

  # Rebuild diffExp = sum_k (Rk %*% H_c) %o% (D[k] * B[k, :])
  B_full <- do.call(cbind, Bi_list)              # K x N
  DB     <- diag(D, nrow = K) %*% B_full         # K x N

  expected <- matrix(0, nrow = J, ncol = N)
  for (k in seq_len(K)) {
    eff_k    <- Rk_list[[k]] %*% H_c             # J x 1
    expected <- expected + eff_k %*% DB[k, , drop = FALSE]
  }

  observed <- m$diffExp(contrast)
  expect_equal(unname(observed), unname(expected), tolerance = 1e-10)
})

# tests/testthat/test-CreateGEDIObject.R
# -----------------------------------------------------------
# Tests for CreateGEDIObject(): object creation, validation,
# training, and basic output structure.
# -----------------------------------------------------------

# ---- shared fixture ----
dat <- make_synthetic_data(n_genes = 200, n_cells = 400,
                           n_samples = 4, seed = 42)

# ===========================================================
# 1. Object creation — correct class and metadata
# ===========================================================
test_that("CreateGEDIObject returns an R6 object with correct class", {
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = 10,
    mode    = "Bsphere",
    verbose = 0
  )
  expect_s3_class(model, "R6")
  # print method should work without error
  expect_output(print(model))
})

test_that("CreateGEDIObject preserves gene and cell IDs before training", {
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = 10,
    mode    = "Bsphere",
    verbose = 0
  )
  meta <- model$metadata
  expect_equal(meta$geneIDs, rownames(dat$M))
  expect_equal(meta$cellIDs, colnames(dat$M))
  expect_equal(levels(meta$sampleIDs), levels(dat$sample_labels))
})

# ===========================================================
# 2. Input validation — bad inputs should error
# ===========================================================
test_that("CreateGEDIObject errors on mismatched sample labels length", {
  bad_labels <- factor(rep("S1", 10))
  expect_error(
    CreateGEDIObject(Samples = bad_labels, M = dat$M, K = 10, verbose = 0)
  )
})

test_that("CreateGEDIObject errors when K >= min(n_genes, n_cells)", {
  expect_error(
    CreateGEDIObject(Samples = dat$sample_labels, M = dat$M, K = 500,
                     verbose = 0)
  )
})

test_that("CreateGEDIObject errors on missing data matrix", {
  expect_error(
    CreateGEDIObject(Samples = dat$sample_labels, K = 10, verbose = 0)
  )
})

# ===========================================================
# 3. Mode selection
# ===========================================================
test_that("Both Bsphere and Bl2 modes create models without error", {
  expect_no_error(
    CreateGEDIObject(Samples = dat$sample_labels, M = dat$M,
                     K = 5, mode = "Bsphere", verbose = 0)
  )
  expect_no_error(
    CreateGEDIObject(Samples = dat$sample_labels, M = dat$M,
                     K = 5, mode = "Bl2", verbose = 0)
  )
})

# ===========================================================
# 4. Training runs and produces expected output
# ===========================================================
test_that("Model trains without error and populates parameters", {
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = 5,
    mode    = "Bsphere",
    verbose = 0
  )
  expect_no_error(model$train(iterations = 10))

  # Core parameter matrices should exist with correct shapes
  Z <- model$params$Z
  expect_true(is.matrix(Z))
  expect_equal(nrow(Z), 200)
  expect_equal(ncol(Z), 5)

  D <- model$params$D
  expect_true(is.numeric(D))
  expect_length(D, 5)
})

test_that("Metadata dimensions are correct after training", {
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = 10,
    mode    = "Bsphere",
    verbose = 0
  )
  model$train(iterations = 10)

  meta <- model$metadata
  expect_equal(meta$n_genes,   200)
  expect_equal(meta$n_cells,   400)
  expect_equal(meta$n_samples, 4)

  expect_equal(model$aux$K,    10)
  expect_equal(model$aux$mode, "Bsphere")
})

test_that("Training populates convergence tracking", {
  model <- CreateGEDIObject(
    Samples = dat$sample_labels,
    M       = dat$M,
    K       = 5,
    mode    = "Bsphere",
    verbose = 0
  )
  model$train(iterations = 20, track_interval = 5)

  tracking <- model$tracking
  expect_true(is.list(tracking))
  expect_true(length(tracking) > 0)
  # Each tracked vector should have entries
  expect_true(length(tracking[[1]]) > 0)
})

# ===========================================================
# 5. Projections (lazy accessors)
# ===========================================================
test_that("Projections ZDB and DB have correct dimensions", {
  model <- make_trained_model(K = 5, iterations = 10)

  zdb <- model$projections$ZDB
  expect_true(is.matrix(zdb) || inherits(zdb, "Matrix"))
  expect_equal(nrow(zdb), 200)  # genes
  expect_equal(ncol(zdb), 400)  # cells

  db <- model$projections$DB
  expect_equal(nrow(db), 5)     # K
  expect_equal(ncol(db), 400)   # cells
})

# ===========================================================
# 6. Imputation accessor
# ===========================================================
test_that("Imputed Y matrix has correct dimensions", {
  model <- make_trained_model(K = 5, iterations = 10)

  Y_imp <- model$imputed$Y()
  expect_true(is.matrix(Y_imp) || inherits(Y_imp, "Matrix"))
  expect_equal(nrow(Y_imp), 200)
  expect_equal(ncol(Y_imp), 400)
})

# ===========================================================
# 7. Reproducibility with same seed
# ===========================================================
test_that("Same seed produces identical synthetic data", {
  dat1 <- make_synthetic_data(seed = 123)
  dat2 <- make_synthetic_data(seed = 123)

  expect_identical(as.matrix(dat1$M), as.matrix(dat2$M))
  expect_identical(dat1$sample_labels, dat2$sample_labels)
})

# ===========================================================
# 8. Original user scenario: 2000 cells, 1000 genes, K=10
# ===========================================================
test_that("Full scenario: NB(1000x2000), 10 samples, K=10, Bsphere", {
  big <- make_synthetic_data(n_genes = 1000, n_cells = 2000,
                             n_samples = 10, seed = 99)

  model <- CreateGEDIObject(
    Samples = big$sample_labels,
    M       = big$M,
    K       = 10,
    mode    = "Bsphere",
    verbose = 0
  )

  # Train for a few iterations
  model$train(iterations = 15)

  # Verify dimensions after training
  expect_equal(model$metadata$n_genes,   1000)
  expect_equal(model$metadata$n_cells,   2000)
  expect_equal(model$metadata$n_samples, 10)
  expect_equal(model$aux$K,              10)
  expect_equal(model$aux$mode,           "Bsphere")

  Z <- model$params$Z
  expect_equal(dim(Z), c(1000, 10))

  # Projections should be computable
  zdb <- model$projections$ZDB
  expect_equal(dim(zdb), c(1000, 2000))
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Create New GEDI Model Object (Internal)
#'
#' Constructs the C++ GEDI model object and returns an external pointer.
#' This function is called internally by the R6 GEDI class during model setup
#' and should not be called directly by users.
#'
#' @param params List containing initialized parameter matrices:
#'   \itemize{
#'     \item Bi: List of cell projection matrices (K x Ni) for each sample
#'     \item Qi: List of sample-specific metagene matrices (J x K)
#'     \item si, oi: Cell and sample-specific offset vectors
#'     \item o: Global gene offset vector (length J)
#'     \item Z: Shared metagene matrix (J x K)
#'     \item U, S, D: SVD components and scaling factors
#'     \item sigma2: Initial variance estimate
#'     \item A: Pathway-latent factor connection matrix (if C provided)
#'     \item Rk, Ro: Sample covariate effect matrices (if H provided)
#'   }
#' @param aux List containing auxiliary variables and dimensions:
#'   \itemize{
#'     \item J, N, K, P, L: Dimensions (genes, cells, factors, pathways, covariates)
#'     \item numSamples: Number of samples
#'     \item obs.type: Observation type ("M", "M_paired", "Y", or "X")
#'     \item mode: Normalization mode ("Bl2" or "Bsphere")
#'     \item orthoZ, adjustD, is_si_fixed: Boolean flags
#'     \item C, H: Prior matrices (gene pathways, sample covariates)
#'     \item diag_K, J_vec, Ni_vec, Ni: Dimension vectors
#'     \item ZDBi, QiDBi: Precomputed product matrices
#'   }
#' @param target List containing target data matrices:
#'   \itemize{
#'     \item Yi: Log-transformed expression by sample (computed by C++ if empty)
#'     \item Mi: Raw count matrix by sample (for obs.type = "M")
#'     \item M1i, M2i: Paired count matrices (for obs.type = "M_paired")
#'     \item Xi: Binary indicator matrix (for obs.type = "X")
#'   }
#' @param hyperparams List containing hyperparameters for regularization:
#'   \itemize{
#'     \item S_Qi, S_oi, S_Z, S_A, S_R: Shrinkage parameters
#'     \item S_Qi_mean, S_oi_mean, S_si, S_o: Mean shrinkage parameters
#'     \item o_0, si_0: Prior mean values
#'     \item O: Random matrix for rSVD initialization
#'   }
#' @param verbose Integer verbosity level (0=silent, 1=info, 2=debug, 4=timing)
#' @param num_threads Number of OpenMP threads (0=auto, uses all available)
#'
#' @return External pointer (SEXP) to C++ GEDI object
#'
#' @details
#' This function creates a stateful C++ GEDI object that stores all model parameters
#' and data. The C++ object performs all heavy computation, while the R6 wrapper
#' remains lightweight (~1 KB). Memory-efficient design: if Yi is not provided,
#' C++ computes it from raw counts M, eliminating duplicate storage in R.
#'
#' The returned external pointer is managed by R's garbage collector and will
#' automatically free the C++ object when no longer referenced.
#'
#' @keywords internal
#' @noRd
GEDI_new <- function(params, aux, target, hyperparams, verbose = 1L, num_threads = 0L) {
    .Call(`_gedi2_GEDI_new`, params, aux, target, hyperparams, verbose, num_threads)
}

#' Set Verbose Level on Existing GEDI Object (Internal)
#'
#' Updates the verbosity level of an already-created GEDI C++ object.
#' This allows R to suppress C++ output when using an R-side progress bar.
#'
#' @param model_ptr External pointer to C++ GEDI object (from GEDI_new)
#' @param verbose Integer verbosity level: 0 (silent), 1 (info), 2+ (debug)
#'
#' @keywords internal
#' @noRd
GEDI_set_verbose <- function(model_ptr, verbose) {
    invisible(.Call(`_gedi2_GEDI_set_verbose`, model_ptr, verbose))
}

#' Initialize Latent Variables (Internal)
#'
#' Initializes latent variables Z, Qi, and Bi using randomized SVD on residual
#' expression data. This is the first step in GEDI model fitting and must be
#' called before optimization.
#'
#' @param model_ptr External pointer to C++ GEDI object (from GEDI_new)
#' @param multimodal Logical, if TRUE skips normalization steps for multi-modal
#'   integration (used internally when combining multiple data modalities)
#'
#' @return List containing initialized model state:
#'   \itemize{
#'     \item params: Updated parameter matrices (Z, Qi, Bi, etc.)
#'     \item aux: Auxiliary variables and dimensions
#'     \item hyperparams: Hyperparameters
#'   }
#'
#' @details
#' Initialization procedure:
#' \enumerate{
#'   \item Solve for initial oi (sample-specific offsets)
#'   \item Compute residual Yp after removing o, oi, si effects
#'   \item Perform randomized SVD on Yp to obtain initial Z
#'   \item Initialize Qi = Z for all samples
#'   \item Solve for initial Bi given Z and Qi
#'   \item Normalize B matrices and update auxiliary products (ZDBi, QiDBi)
#' }
#'
#' Uses randomized SVD (rSVD) for computational efficiency with large matrices.
#' The multimodal flag is used internally for advanced integration workflows.
#'
#' @keywords internal
#' @noRd
GEDI_initialize <- function(model_ptr, multimodal = FALSE) {
    .Call(`_gedi2_GEDI_initialize`, model_ptr, multimodal)
}

#' Optimize GEDI Model via Block Coordinate Descent (Internal)
#'
#' Performs block coordinate descent optimization to fit the GEDI model.
#' Iteratively updates all model parameters (Bi, Z, Qi, oi, si, o, sigma2)
#' until convergence or maximum iterations reached.
#'
#' @param model_ptr External pointer to initialized C++ GEDI object
#' @param iterations Integer, number of optimization iterations to perform
#' @param track_interval Integer, interval for tracking convergence metrics
#'   (e.g., track_interval=5 tracks every 5th iteration; track_interval=1 tracks all)
#'
#' @return List containing optimized model results:
#'   \itemize{
#'     \item params: Optimized parameter matrices (Z, Bi, Qi, sigma2, etc.)
#'     \item aux: Auxiliary variables and dimensions
#'     \item hyperparams: Updated hyperparameters (o_0, S_o, S_si)
#'     \item tracking: Convergence metrics (dZ, dBi, dQi, sigma2 trajectory)
#'     \item convergence: Summary statistics (iterations, total_time_ms, final_sigma2)
#'   }
#'
#' @details
#' Block coordinate descent iteration:
#' \enumerate{
#'   \item Solve Bi for all samples (parallelized with OpenMP)
#'   \item Normalize B matrices
#'   \item Solve Z (orthogonal or regular, depending on orthoZ flag)
#'   \item Solve Qi for all samples (parallelized)
#'   \item Solve oi, si, o (offset parameters)
#'   \item Update Yi from raw counts M (for count data)
#'   \item Solve sigma2 (variance parameter)
#'   \item Update hyperpriors (o_0, S_o, S_si)
#'   \item Solve A, Rk, Ro (if priors C or H are provided)
#'   \item Track convergence metrics at specified intervals
#' }
#'
#' OpenMP parallelization is automatically enabled if available. Progress and
#' timing information printed based on verbosity level.
#'
#' @keywords internal
#' @noRd
GEDI_optimize <- function(model_ptr, iterations, track_interval = 5L) {
    .Call(`_gedi2_GEDI_optimize`, model_ptr, iterations, track_interval)
}

#' Train GEDI Model (Initialize + Optimize) (Internal)
#'
#' Convenience function that performs both initialization and optimization in a
#' single call. Equivalent to calling GEDI_initialize() followed by GEDI_optimize().
#'
#' @param model_ptr External pointer to C++ GEDI object (from GEDI_new)
#' @param iterations Integer, number of optimization iterations to perform
#' @param track_interval Integer, interval for tracking convergence metrics
#' @param multimodal Logical, if TRUE skips normalization during initialization
#'   (for multi-modal integration workflows)
#'
#' @return List containing trained model results (same format as GEDI_optimize):
#'   \itemize{
#'     \item params: Optimized parameter matrices
#'     \item aux: Auxiliary variables
#'     \item hyperparams: Updated hyperparameters
#'     \item tracking: Convergence metrics
#'     \item convergence: Summary statistics
#'   }
#'
#' @details
#' This is the recommended way to fit a GEDI model in a single step. Internally:
#' \enumerate{
#'   \item Calls initialize() to set up latent variables via rSVD
#'   \item Calls optimize() to perform block coordinate descent
#'   \item Returns the final optimized model state
#' }
#'
#' For more control over the fitting process, use GEDI_initialize() and
#' GEDI_optimize() separately. The multimodal parameter is for advanced use cases.
#'
#' @keywords internal
#' @noRd
GEDI_train <- function(model_ptr, iterations, track_interval = 5L, multimodal = FALSE) {
    .Call(`_gedi2_GEDI_train`, model_ptr, iterations, track_interval, multimodal)
}

getDiffO_cpp <- function(Ro, H_rotation, contrast, verbose = 0L) {
    .Call(`_gedi2_getDiffO_cpp`, Ro, H_rotation, contrast, verbose)
}

getDiffQ_cpp <- function(Rk_list, H_rotation, contrast, verbose = 0L) {
    .Call(`_gedi2_getDiffQ_cpp`, Rk_list, H_rotation, contrast, verbose)
}

getDiffExp_cpp <- function(Rk_list, H_rotation, contrast, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_getDiffExp_cpp`, Rk_list, H_rotation, contrast, D, Bi_list, verbose)
}

compute_svd_factorized_cpp <- function(Z, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_svd_factorized_cpp`, Z, D, Bi_list, verbose)
}

run_factorized_svd_cpp <- function(Z, projDB, verbose = 0L) {
    .Call(`_gedi2_run_factorized_svd_cpp`, Z, projDB, verbose)
}

#' Compute Residual After Removing Sample Effects (Internal)
#'
#' Removes sample-specific effects (QiDBi, si, oi) and global offset (o) from Yi.
#' Used to extract the shared biological signal (ZDBi component).
#'
#' @param Yi Dense matrix (J x Ni) - fitted log-expression for sample i
#' @param QiDBi Dense matrix (J x Ni) - sample-specific metagene projections
#' @param si Vector (Ni) - cell-specific library size offsets for sample i
#' @param o Vector (J) - global gene-specific offsets
#' @param oi Vector (J) - sample-specific gene offsets
#'
#' @return Dense matrix (J x Ni) - residual Yi after removing sample effects
#'
#' @details
#' Computes: Yi - QiDBi - (si (x) 1^T) - (o + oi) (x) 1^T
#' This leaves only the ZDBi component (shared biological signal).
#'
#' @keywords internal
#' @noRd
Yi_resZ <- function(Yi, QiDBi, si, o, oi) {
    .Call(`_gedi2_Yi_resZ`, Yi, QiDBi, si, o, oi)
}

#' Predict Yi from Model Parameters (Internal)
#'
#' Reconstructs the model's prediction of Yi (log-expression) from all components.
#'
#' @param ZDBi Dense matrix (J x Ni) - shared metagene projections
#' @param QiDBi Dense matrix (J x Ni) - sample-specific metagene projections
#' @param si Vector (Ni) - cell-specific library size offsets
#' @param o Vector (J) - global gene-specific offsets
#' @param oi Vector (J) - sample-specific gene offsets
#'
#' @return Dense matrix (J x Ni) - predicted Yi
#'
#' @details
#' Computes: Y_hati = ZDBi + QiDBi + (si (x) 1^T) + (o + oi) (x) 1^T
#' This is the full model prediction for log-expression.
#'
#' @keywords internal
#' @noRd
predict_Yhat <- function(ZDBi, QiDBi, si, o, oi) {
    .Call(`_gedi2_predict_Yhat`, ZDBi, QiDBi, si, o, oi)
}

#' Variance of Yi for Single Count Matrix (Internal)
#'
#' Computes posterior variance of Yi given the fitted values and model variance.
#' For observation type "M" (single count matrix).
#'
#' @param Yi Dense matrix (J x Ni) - fitted log-expression
#' @param sigma2 Scalar - model variance parameter
#'
#' @return Dense matrix (J x Ni) - posterior variance at each position
#'
#' @details
#' Variance formula: Var(Yi | Mi, model) = 1 / (exp(Yi) + 1/sigma2)
#'
#' This comes from the Poisson-lognormal model where:
#' - Mi ~ Poisson(exp(Yi))
#' - Yi ~ N(Y_hati, sigma2)
#'
#' The posterior variance decreases where:
#' - Counts are high (exp(Yi) large)
#' - Model uncertainty is low (sigma2 small)
#'
#' @keywords internal
#' @noRd
Yi_var_M <- function(Yi, sigma2) {
    .Call(`_gedi2_Yi_var_M`, Yi, sigma2)
}

#' Variance of Yi for Paired Count Matrices (Internal)
#'
#' Computes posterior variance of Yi for paired count data (e.g., CITE-seq, ATAC+RNA).
#' For observation type "M_paired".
#'
#' @param Yi Dense matrix (J x Ni) - fitted log-ratio: log((M1+1)/(M2+1))
#' @param M1i Sparse matrix (J x Ni) - first count matrix (e.g., RNA)
#' @param M2i Sparse matrix (J x Ni) - second count matrix (e.g., protein)
#' @param sigma2 Scalar - model variance parameter
#'
#' @return Dense matrix (J x Ni) - posterior variance at each position
#'
#' @details
#' Variance formula: Var(Yi | M1i, M2i, model) = 1 / (M * exp(-|Yi|) / (1+exp(-|Yi|))^2 + 1/sigma2)
#' where M = M1i + M2i
#'
#' This comes from the binomial-logistic-normal model where:
#' - M1i ~ Binomial(M1i + M2i, p)
#' - logit(p) = Yi ~ N(Y_hati, sigma2)
#'
#' The variance depends on:
#' - Total counts M (more counts = less variance)
#' - Yi magnitude (variance minimized at Yi = 0, i.e., p = 0.5)
#' - Model uncertainty sigma2
#'
#' @keywords internal
#' @noRd
Yi_var_M_paired <- function(Yi, M1i, M2i, sigma2) {
    .Call(`_gedi2_Yi_var_M_paired`, Yi, M1i, M2i, sigma2)
}

#' Compute Dispersion Statistics (Sparse-Optimized) (Internal)
#'
#' Analyzes the relationship between predicted and observed variance for count data.
#' Only samples nonzero positions from the sparse count matrix for memory efficiency.
#'
#' @param Yi_fitted Dense matrix (J x Ni) - model prediction of log-expression
#' @param Mi Sparse matrix (J x Ni) - observed count matrix
#' @param subsample Integer - maximum number of nonzero positions to sample
#'
#' @return List containing:
#'   - predicted: Vector of predicted counts (lambda = exp(Yi_fitted))
#'   - observed: Vector of observed counts from Mi
#'
#' @details
#' For Poisson model: Mi ~ Poisson(lambda), expected variance = lambda.
#' This function samples nonzero positions and returns predicted vs observed values
#' for downstream dispersion analysis (binning and aggregation done in R).
#'
#' Memory optimization: Works only on sparse nonzero positions (typically 5-10% of matrix).
#' For 30K genes x 10K cells with 5% nonzero: samples from ~15M positions instead of 300M.
#'
#' @keywords internal
#' @noRd
compute_dispersion_sparse <- function(Yi_fitted, Mi, subsample) {
    .Call(`_gedi2_compute_dispersion_sparse`, Yi_fitted, Mi, subsample)
}

#' Sum of Squared Errors for Paired Data (Internal)
#'
#' Computes the SSE term used in sigma2 optimization for M_paired observation type.
#' This is included for completeness but is called during optimization, not imputation.
#'
#' @param Yi Dense matrix (J x Ni) - fitted log-ratio
#' @param M1i Sparse matrix (J x Ni) - first count matrix
#' @param M2i Sparse matrix (J x Ni) - second count matrix
#' @param ZDBi Dense matrix (J x Ni) - shared metagene projections
#' @param QiDBi Dense matrix (J x Ni) - sample-specific metagene projections
#' @param si Vector (Ni) - cell-specific offsets
#' @param o Vector (J) - global gene offsets
#' @param oi Vector (J) - sample-specific gene offsets
#' @param sigma2 Scalar - current variance estimate
#'
#' @return Scalar - SSE contribution from sample i
#'
#' @keywords internal
#' @noRd
Yi_SSE_M_paired <- function(Yi, M1i, M2i, ZDBi, QiDBi, si, o, oi, sigma2) {
    .Call(`_gedi2_Yi_SSE_M_paired`, Yi, M1i, M2i, ZDBi, QiDBi, si, o, oi, sigma2)
}

#' Compute Single Feature Projection (Internal)
#'
#' Projects a single feature (gene) through the GEDI model to get cell-level values.
#' Computes: (feature_weights * D) %*% B, avoiding full ZDB computation.
#'
#' @param feature_weights Vector of length K (factor loadings for this feature)
#' @param D Scaling vector of length K
#' @param Bi_list List of sample-specific cell projection matrices (K x Ni each)
#' @param verbose Integer verbosity level
#'
#' @return Vector of length N (total cells) with projected values
#'
#' @keywords internal
#' @noRd
compute_feature_projection <- function(feature_weights, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_feature_projection`, feature_weights, D, Bi_list, verbose)
}

#' Compute Multi-Feature Projection (Internal)
#'
#' Projects multiple features through the GEDI model simultaneously.
#' Computes: (feature_weights * D) %*% B for F features.
#'
#' @param feature_weights Matrix K x F (factor loadings for F features)
#' @param D Scaling vector of length K
#' @param Bi_list List of sample-specific cell projection matrices (K x Ni each)
#' @param verbose Integer verbosity level
#'
#' @return Matrix N x F with projected values for each feature
#'
#' @keywords internal
#' @noRd
compute_multi_feature_projection <- function(feature_weights, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_multi_feature_projection`, feature_weights, D, Bi_list, verbose)
}

#' Aggregate Vector Field into Bins (Internal)
#'
#' Bins vector field data into a grid and computes mean vectors per bin.
#' Used for cleaner arrow plots without overplotting.
#'
#' @param Dim1 Start x-coordinates (length N)
#' @param Dim2 Start y-coordinates (length N)
#' @param To1 End x-coordinates (length N)
#' @param To2 End y-coordinates (length N)
#' @param color Color values (length N), can be numeric or will be converted
#' @param alpha Alpha values (length N)
#' @param n_bins Number of bins per dimension
#' @param min_per_bin Minimum observations required per bin
#'
#' @return Data frame with columns: Dim1, Dim2, To1, To2, deltaDim1, deltaDim2,
#'   Color, Alpha, n (observations per bin)
#'
#' @keywords internal
#' @noRd
aggregate_vectors <- function(Dim1, Dim2, To1, To2, color, alpha, n_bins, min_per_bin) {
    .Call(`_gedi2_aggregate_vectors`, Dim1, Dim2, To1, To2, color, alpha, n_bins, min_per_bin)
}

#' Compute ZDB Projection (Internal)
#'
#' Computes the shared manifold projection ZDB = Z * diag(D) * B, where B is
#' the concatenation of all sample-specific Bi matrices. This is the main
#' integrated representation of cells in the GEDI latent space.
#'
#' @param Z Shared metagene matrix (J x K), where J = genes, K = latent factors
#' @param D Scaling vector (length K) representing the importance of each factor
#' @param Bi_list List of sample-specific cell projection matrices, where each
#'   Bi is a K x Ni matrix (Ni = number of cells in sample i)
#' @param verbose Integer verbosity level:
#'   \itemize{
#'     \item 0: Silent (no output)
#'     \item 1: Progress bar and summary statistics
#'     \item 2: Detailed per-sample information
#'   }
#'
#' @return Dense matrix ZDB of dimensions J x N, where N = sum(Ni) is the total
#'   number of cells across all samples. Each column represents a cell in the
#'   integrated latent space.
#'
#' @details
#' The ZDB projection represents each cell as a linear combination of shared
#' metagenes (Z), weighted by latent factors and scaled by their importance (D).
#' 
#' Computational strategy:
#' \enumerate{
#'   \item Pre-compute ZD = Z * diag(D) once (saves KxJxnumSamples operations)
#'   \item For each sample i: compute ZD * Bi and concatenate
#'   \item Use Eigen block operations for efficient memory access
#' }
#'
#' Memory: Allocates one J x N dense matrix. For large datasets (e.g., 20k genes,
#' 50k cells), this requires ~8 GB RAM. Consider computing projections on demand
#' or working with subsets if memory is limited.
#'
#' Performance: OpenMP parallelization available if enabled during compilation.
#' Typical speed: ~100-200ms for 20k x 5k dataset on modern CPU.
#'
#' @keywords internal
#' @noRd
compute_ZDB_cpp <- function(Z, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_ZDB_cpp`, Z, D, Bi_list, verbose)
}

#' Compute DB Projection (Internal)
#'
#' Computes the latent factor embedding DB = diag(D) * B, where B is the
#' concatenation of all sample-specific Bi matrices. This represents cells
#' in the K-dimensional latent factor space.
#'
#' @param D Scaling vector (length K) representing factor importance
#' @param Bi_list List of sample-specific cell projection matrices, where each
#'   Bi is a K x Ni matrix
#' @param verbose Integer verbosity level (0 = silent, 1 = progress, 2 = detailed)
#'
#' @return Dense matrix DB of dimensions K x N, where N = sum(Ni). Each column
#'   represents a cell's coordinates in the latent factor space.
#'
#' @details
#' The DB projection is simpler than ZDB as it operates directly in the K-dimensional
#' latent space without projecting through the gene space. This is useful for:
#' \itemize{
#'   \item Visualization (when K is small, e.g., K=2 or K=3)
#'   \item Downstream clustering or classification
#'   \item Understanding factor contributions per cell
#' }
#'
#' Computational strategy:
#' \enumerate{
#'   \item Concatenate all Bi matrices horizontally
#'   \item Apply diagonal scaling: diag(D) * B
#' }
#'
#' Memory: Much smaller than ZDB (K x N vs. J x N). For K=10 and N=50k cells,
#' requires only ~4 MB vs. ~8 GB for ZDB when J=20k.
#'
#' Performance: Very fast (~10-50ms) since K << J typically.
#'
#' @keywords internal
#' @noRd
compute_DB_cpp <- function(D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_DB_cpp`, D, Bi_list, verbose)
}

#' Compute ADB Projection (Internal)
#'
#' Computes the pathway activity projection ADB = C.rotation * A * diag(D) * B.
#' This represents cells in terms of their pathway/gene set activities rather
#' than individual gene expression.
#'
#' @param C_rotation Rotation matrix (num_pathways x P) that maps from the P-dimensional
#'   reduced SVD space back to the original pathway space
#' @param A Pathway-factor connection matrix (P x K) learned during training, where
#'   P is the number of reduced components and K is the number of latent factors
#' @param D Scaling vector (length K) representing factor importance
#' @param Bi_list List of sample-specific cell projection matrices (K x Ni each)
#' @param verbose Integer verbosity level (0 = silent, 1 = progress, 2 = detailed)
#'
#' @return Dense matrix ADB of dimensions num_pathways x N, where num_pathways is the
#'   number of original pathways and N = total cells. Each column represents a cell's
#'   pathway activities.
#'
#' @details
#' The ADB projection is only available when a gene-level prior matrix C was
#' provided during model setup. It enables:
#' \itemize{
#'   \item Pathway-level interpretation of cell states
#'   \item Identifying enriched pathways per cell type
#'   \item Biological interpretation in original pathway space
#' }
#'
#' Mathematical formulation:
#' \deqn{ADB = C.rotation \times A \times diag(D) \times B}
#'
#' where:
#' \itemize{
#'   \item C.rotation (num_pathways x P) maps from reduced space to original pathways
#'   \item A (P x K) captures pathway-factor associations in reduced space
#'   \item D (K) scales the importance of each latent factor
#'   \item B (K x N) contains cell projections in latent space
#' }
#'
#' Computational strategy:
#' \enumerate{
#'   \item Concatenate B from all Bi matrices
#'   \item Compute (C.rotation x A) x (diag(D) x B) efficiently
#'   \item Return in original pathway space for interpretation
#' }
#'
#' Memory: num_pathways x N matrix. For 100 original pathways and N=50k cells, ~40 MB.
#'
#' Performance: Fast due to small P and K dimensions (~50-100ms typical).
#'
#' @keywords internal
#' @noRd
compute_ADB_cpp <- function(C_rotation, A, D, Bi_list, verbose = 0L) {
    .Call(`_gedi2_compute_ADB_cpp`, C_rotation, A, D, Bi_list, verbose)
}

#' Compute Cell-Specific Scaling Factors (Dense)
#'
#' Calculates column-wise means of a dense expression matrix for hyperparameter initialization.
#' This function computes s_0 = (J_vec^T * Y) / J, where each element represents the mean
#' expression across all genes for each cell. Used internally during GEDI model setup.
#'
#' @param J_vec Vector of ones with length equal to number of genes (J)
#' @param Y Dense log-transformed expression matrix (genes x cells)
#' @param J Number of genes (rows in Y)
#'
#' @return Vector of length N (number of cells) containing cell-specific scaling factors
#'
#' @details
#' This is the dense matrix version used when Y matrix is provided directly.
#' For count matrices (M), use the sparse version \code{compute_s_0()} instead.
#' The function uses BLAS-optimized dense matrix-vector multiplication for efficiency.
#'
#' @keywords internal
#' @noRd
compute_s_0_dense <- function(J_vec, Y, J) {
    .Call(`_gedi2_compute_s_0_dense`, J_vec, Y, J)
}

#' Compute Library-Size Normalized Expression (Sparse)
#'
#' Normalizes sparse expression matrix Y by dividing each element by the outer product
#' of J_vec and s_0. This is equivalent to Y / (J_vec * s_0^T).
#'
#' @param Y Sparse expression matrix (genes x cells)
#' @param J_vec Vector of ones with length J (number of genes)
#' @param s_0 Vector of cell-specific scaling factors (length N)
#'
#' @return Sparse matrix Yp with normalized expression values
#'
#' @details
#' Efficiently iterates through non-zero elements only, preserving sparsity.
#' Each Y(i,j) is divided by (J_vec(i) * s_0(j)).
#'
#' @keywords internal
#' @noRd
compute_Yp <- function(Y, J_vec, s_0) {
    .Call(`_gedi2_compute_Yp`, Y, J_vec, s_0)
}

#' Compute Gene-Specific Offsets (Dense)
#'
#' Calculates row-wise means of normalized expression matrix for hyperparameter initialization.
#' Computes o_0 = (Yp * N_vec) / N, where each element represents the mean expression
#' of a gene across all cells after library size normalization.
#'
#' @param Yp Normalized dense expression matrix (genes x cells)
#' @param N_vec Vector of ones with length N (number of cells)
#' @param N Number of cells (columns in Yp)
#'
#' @return Vector of length J (number of genes) containing gene-specific offsets
#'
#' @details
#' Used in conjunction with \code{compute_s_0_dense()} and \code{compute_Yp_dense()}
#' to initialize hyperparameters o_0 and s_0 for the GEDI model.
#'
#' @keywords internal
#' @noRd
compute_o_0_dense <- function(Yp, N_vec, N) {
    .Call(`_gedi2_compute_o_0_dense`, Yp, N_vec, N)
}

#' Compute Library-Size Normalized Expression (Dense)
#'
#' Computes normalized expression Yp by subtracting the outer product of J_vec and s_0 from Y.
#' This is equivalent to Yp = Y - J_vec * s_0^T.
#'
#' @param Y Dense expression matrix (genes x cells)
#' @param J_vec Vector of ones with length J (number of genes)
#' @param s_0 Vector of cell-specific scaling factors (length N)
#'
#' @return Dense matrix Yp with normalized expression values
#'
#' @details
#' Alternative formulation of library size normalization for dense matrices.
#' Uses \code{noalias()} for efficient in-place subtraction without temporary objects.
#'
#' @keywords internal
#' @noRd
compute_Yp_dense <- function(Y, J_vec, s_0) {
    .Call(`_gedi2_compute_Yp_dense`, Y, J_vec, s_0)
}

#' Vector Outer Product
#'
#' Computes the outer product of two vectors: a * b^T, returning a dense matrix.
#' This is used internally for efficient matrix operations in GEDI model fitting.
#'
#' @param a First vector (length m)
#' @param b Second vector (length n)
#'
#' @return Dense matrix of dimensions (m x n) containing the outer product
#'
#' @details
#' Uses BLAS-optimized operations for efficient computation.
#' Equivalent to R's \code{a \%*\% t(b)}.
#'
#' @keywords internal
#' @noRd
VecVecProduct <- function(a, b) {
    .Call(`_gedi2_VecVecProduct`, a, b)
}

#' Dense Matrix-Vector Product
#'
#' Computes the product of a dense matrix A and vector b: A * b.
#' Returns the result as a column vector.
#'
#' @param A Dense matrix (m x n)
#' @param b Vector of length n
#'
#' @return Vector of length m containing the product A * b
#'
#' @details
#' Uses BLAS level-2 optimized matrix-vector multiplication.
#' Equivalent to R's \code{A \%*\% b}.
#'
#' @keywords internal
#' @noRd
MatVecProduct <- function(A, b) {
    .Call(`_gedi2_MatVecProduct`, A, b)
}

#' Sparse Matrix-Vector Product
#'
#' Computes the product of a sparse matrix A and vector b: A * b.
#' Efficiently handles sparse matrices by only operating on non-zero elements.
#'
#' @param A Sparse matrix (m x n)
#' @param b Vector of length n
#'
#' @return Vector of length m containing the product A * b
#'
#' @details
#' Optimized for sparse matrices, avoiding unnecessary multiplications by zero.
#' Useful for large-scale single-cell data where most entries are zero.
#'
#' @keywords internal
#' @noRd
eigenSparseMatVecProduct <- function(A, b) {
    .Call(`_gedi2_eigenSparseMatVecProduct`, A, b)
}

#' Compute Library-Size Normalized Count Matrix (Sparse)
#'
#' Normalizes sparse count matrix M by dividing each element by the outer product
#' of J_vec and s_0. This is equivalent to Mp = M / (J_vec * s_0^T).
#'
#' @param M Sparse raw count matrix (genes x cells)
#' @param J_vec Vector of ones with length J (number of genes)
#' @param s_0 Vector of cell-specific scaling factors (length N)
#'
#' @return Sparse matrix Mp with normalized counts
#'
#' @details
#' Efficiently iterates through non-zero elements only, preserving sparsity structure.
#' Each M(i,j) is divided by (J_vec(i) * s_0(j)).
#' Used for initializing hyperparameters from raw count data.
#'
#' @keywords internal
#' @noRd
compute_Mp <- function(M, J_vec, s_0) {
    .Call(`_gedi2_compute_Mp`, M, J_vec, s_0)
}

#' Compute Cell-Specific Scaling Factors from Counts (Sparse)
#'
#' Calculates column-wise sums of sparse count matrix M for hyperparameter initialization.
#' Computes s_0 = (J_vec^T * M + 0.01) / J, where each element represents the library
#' size (total count) for each cell with a small constant added for numerical stability.
#'
#' @param J_vec Vector of ones with length equal to number of genes (J)
#' @param M Sparse raw count matrix (genes x cells)
#' @param J Number of genes (rows in M)
#'
#' @return Vector of length N (number of cells) containing cell-specific scaling factors
#'
#' @details
#' This is the sparse matrix version used when raw count matrix M is provided.
#' For pre-processed Y matrices, use \code{compute_s_0_dense()} instead.
#' The small constant (0.01) prevents division by zero for cells with very low counts.
#' Uses BLAS-optimized sparse matrix-vector multiplication.
#'
#' @keywords internal
#' @noRd
compute_s_0 <- function(J_vec, M, J) {
    .Call(`_gedi2_compute_s_0`, J_vec, M, J)
}

#' Compute Gene-Specific Offsets from Normalized Counts (Sparse)
#'
#' Calculates row-wise sums of library-size normalized sparse matrix Mp for
#' hyperparameter initialization. Computes o_0 = (Mp * N_vec + 1e-5) / N, where
#' each element represents the mean normalized count of a gene across all cells.
#'
#' @param Mp Sparse library-size normalized count matrix (genes x cells)
#' @param N_vec Vector of ones with length N (number of cells)
#' @param N Number of cells (columns in Mp)
#'
#' @return Vector of length J (number of genes) containing gene-specific offsets
#'
#' @details
#' Used in conjunction with \code{compute_s_0()} and \code{compute_Mp()} to initialize
#' hyperparameters o_0 and s_0 from raw count data. The small constant (1e-5) provides
#' numerical stability for genes with very low expression.
#' Uses BLAS-optimized sparse matrix-vector multiplication.
#'
#' @keywords internal
#' @noRd
compute_o_0 <- function(Mp, N_vec, N) {
    .Call(`_gedi2_compute_o_0`, Mp, N_vec, N)
}

# R/gedi_plot_embedding.R

# Suppress R CMD check NOTEs for ggplot2 NSE variables
utils::globalVariables(c(
  "Dim1", "Dim2", "Color", "Value", "Feature", "Alpha",
  "To1", "To2", "VF_Color",
  "Expected_Var", "Observed_Var", "Sample",
  "Iteration", "RMSD", "Parameter", "Factor", "Group", "sigma2"
))

#' Plot 2D Embedding (Base Function)
#'
#' Base function for plotting 2D embeddings with customizable coloring.
#' Supports both continuous and discrete color variables.
#'
#' @param embedding Matrix (N x 2) with x and y coordinates for each cell
#' @param color Vector of length N for coloring points, or NULL for uniform color
#' @param color_limits Numeric vector c(low, high) for color scale limits,
#'   or NULL to auto-compute from data (uses 99th percentile for symmetric limits)
#' @param palette Character vector of colors for continuous scale (length 3 for diverging)
#' @param randomize Logical, whether to randomize point order before plotting
#' @param point_size Numeric, size of points
#' @param alpha Numeric, transparency of points (0-1)
#' @param raster Logical, use rasterization for large datasets (>100k points)
#' @param xlab Character, x-axis label
#' @param ylab Character, y-axis label
#' @param title Character, plot title
#' @param legend_title Character, legend title (auto-detected if NULL)
#'
#' @return ggplot2 object
#'
#' @keywords internal
.plot_embedding_base <- function(embedding,
                                 color = NULL,
                                 color_limits = NULL,
                                 palette = c("blue", "lightgrey", "red"),
                                 randomize = TRUE,
                                 point_size = 0.3,
                                 alpha = 0.9,
                                 raster = FALSE,
                                 xlab = "Dim 1",
                                 ylab = "Dim 2",
                                 title = NULL,
                                 legend_title = NULL) {
  # Validate inputs
  if (!is.matrix(embedding) && !is.data.frame(embedding)) {
    stop("embedding must be a matrix or data.frame", call. = FALSE)
  }

  if (ncol(embedding) != 2) {
    stop("embedding must have exactly 2 columns", call. = FALSE)
  }

  N <- nrow(embedding)

  # Handle color
  if (is.null(color)) {
    color <- rep(1, N)
    use_color <- FALSE
  } else {
    if (length(color) != N) {
      stop("color must have same length as number of rows in embedding", call. = FALSE)
    }
    use_color <- TRUE
  }

  # Create data frame
  df <- data.frame(
    Dim1 = embedding[, 1],
    Dim2 = embedding[, 2],
    Color = color
  )

  # Randomize ONCE before any processing
  if (randomize) {
    rand_idx <- sample(N)
    df <- df[rand_idx, ]
  }

  # Detect if color is numeric or discrete
  is_numeric <- is.numeric(color) && length(unique(color)) > 10

  # Start plot
  p <- ggplot2::ggplot(df, ggplot2::aes(x = Dim1, y = Dim2))

  # Add points (raster or regular)
  if (raster && N > 100000) {
    if (!use_color) {
      p <- p + ggplot2::geom_point(size = point_size, alpha = alpha)
    } else {
      p <- p + ggplot2::geom_point(ggplot2::aes(color = Color),
        size = point_size, alpha = alpha
      )
    }
    warning("Rasterization not implemented yet. Using regular points.",
      call. = FALSE
    )
  } else {
    if (!use_color) {
      p <- p + ggplot2::geom_point(size = point_size, alpha = alpha)
    } else {
      p <- p + ggplot2::geom_point(ggplot2::aes(color = Color),
        size = point_size, alpha = alpha
      )
    }
  }

  # Add color scale
  if (use_color) {
    if (is_numeric) {
      # Continuous color scale
      if (is.null(color_limits)) {
        color_limits <- compute_color_limits(color, symmetric = TRUE, quantile = 0.99)
      }

      if (is.null(legend_title)) {
        legend_title <- "Value"
      }

      p <- p + scale_color_gedi_diverging(
        limits = color_limits,
        name = legend_title
      )
    } else {
      # Discrete color scale
      if (is.null(legend_title)) {
        legend_title <- "Group"
      }

      p <- p + scale_color_gedi_discrete(name = legend_title) +
        ggplot2::guides(color = ggplot2::guide_legend(
          override.aes = list(size = 3, alpha = 1)
        ))
    }
  }

  # Add labels and theme
  p <- p +
    ggplot2::labs(x = xlab, y = ylab, title = title) +
    theme_gedi()

  return(p)
}


#' Plot Multiple Features on Embedding
#'
#' Efficiently plots multiple gene features on a 2D embedding using faceting.
#' Computes projections on-the-fly without storing full ZDB matrix.
#'
#' @param model GEDI model object
#' @param features Character vector of gene names or integer indices
#' @param embedding Character specifying embedding type ("umap", "pca") or
#'   a custom N x 2 matrix
#' @param projection Character, type of projection to compute ("zdb" or "db")
#' @param color_limits Character ("global" for shared scale, "individual" for
#'   per-facet scale) or numeric vector c(low, high)
#' @param ncol Integer, number of columns in facet layout
#' @param randomize Logical, whether to randomize point order
#' @param point_size Numeric, size of points
#' @param alpha Numeric, transparency of points
#' @param title Character, plot title
#'
#' @return ggplot2 object with faceted features
#'
#' @examples
#' \donttest{
#' if (requireNamespace("SeuratObject", quietly = TRUE) &&
#'     requireNamespace("uwot", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'   plot_features(model, c(1, 2), embedding = "pca")
#' }
#' }
#'
#' @export
plot_features <- function(model,
                          features,
                          embedding = "umap",
                          projection = "zdb",
                          color_limits = "global",
                          ncol = NULL,
                          randomize = TRUE,
                          point_size = 0.2,
                          alpha = 0.9,
                          title = NULL) {
  # Validate model
  if (is.null(model$params)) {
    stop("Model not trained. Run $train() first.", call. = FALSE)
  }

  # Get embedding coordinates
  if (is.character(embedding)) {
    if (embedding == "umap") {
      emb_mat <- model$embeddings$umap
    } else if (embedding == "pca") {
      emb_mat <- model$embeddings$pca[, 1:2]
    } else {
      stop("embedding must be 'umap', 'pca', or a matrix", call. = FALSE)
    }
  } else if (is.matrix(embedding) || is.data.frame(embedding)) {
    emb_mat <- as.matrix(embedding)
    if (ncol(emb_mat) < 2) {
      stop("Custom embedding must have at least 2 columns", call. = FALSE)
    }
    emb_mat <- emb_mat[, 1:2]
  } else {
    stop("embedding must be a character or matrix", call. = FALSE)
  }

  N <- nrow(emb_mat)

  # Check projection type
  if (!projection %in% c("zdb", "db", "adb")) {
    stop("projection must be 'zdb', 'db', or 'adb'", call. = FALSE)
  }

  # For adb, feature names refer to pathway names, not genes
  if (projection == "adb") {
    if (model$aux$P == 0) {
      stop("projection 'adb' requires a model with gene-level prior (C matrix). This model has P=0.", call. = FALSE)
    }

    # Get the pathway names
    pathway_names <- colnames(model$.__enclos_env__$private$.aux_static$inputC)
    if (is.null(pathway_names)) {
      pathway_names <- paste0("Pathway", 1:model$aux$P)
    }

    if (is.character(features)) {
      feature_idx <- match(features, pathway_names)
      if (any(is.na(feature_idx))) {
        missing <- features[is.na(feature_idx)]
        stop("Features (pathways) not found: ", paste(missing, collapse = ", "), call. = FALSE)
      }
      feature_names <- features
    } else {
      feature_idx <- as.integer(features)
      if (any(feature_idx < 1 | feature_idx > length(pathway_names))) {
        stop("Feature indices out of range for pathways", call. = FALSE)
      }
      feature_names <- pathway_names[feature_idx]
    }

    F_count <- length(feature_idx)

    # Evaluate ADB projection specifically for chosen pathways
    ADB <- model$projections$ADB # P x N
    if (!is.null(rownames(ADB))) {
      # If ADB has row names (from C matrix columns), we can extract directly
      projections <- t(ADB[feature_names, , drop = FALSE]) # N x F_count
    } else {
      # Otherwise fall back to integer indices
      projections <- t(ADB[feature_idx, , drop = FALSE]) # N x F_count
    }
  } else {
    # Convert feature names to indices (genes for zdb/db)
    geneIDs <- model$metadata$geneIDs
    if (is.character(features)) {
      feature_idx <- match(features, geneIDs)
      if (any(is.na(feature_idx))) {
        missing <- features[is.na(feature_idx)]
        stop("Features not found: ", paste(missing, collapse = ", "), call. = FALSE)
      }
      feature_names <- features
    } else {
      feature_idx <- as.integer(features)
      if (any(feature_idx < 1 | feature_idx > length(geneIDs))) {
        stop("Feature indices out of range", call. = FALSE)
      }
      feature_names <- geneIDs[feature_idx]
    }

    F_count <- length(feature_idx)

    # Extract feature weights from Z
    Z <- model$params$Z
    feature_weights <- Z[feature_idx, , drop = FALSE] # F x K
    feature_weights <- t(feature_weights) # K x F for C++

    # Compute projections using C++
    if (projection == "zdb") {
      projections <- compute_multi_feature_projection(
        feature_weights = feature_weights,
        D = model$params$D,
        Bi_list = model$params$Bi,
        verbose = 0
      ) # Returns N x F matrix
    } else if (projection == "db") {
      # For DB: just use factor weights directly
      DB <- model$projections$DB # K x N
      projections <- t(DB) %*% feature_weights # N x F
    }
  }

  # Build long-format data frame
  df_list <- vector("list", F_count)
  for (f in 1:F_count) {
    df_list[[f]] <- data.frame(
      Dim1 = emb_mat[, 1],
      Dim2 = emb_mat[, 2],
      Value = projections[, f],
      Feature = feature_names[f]
    )
  }
  df <- do.call(rbind, df_list)
  df$Feature <- factor(df$Feature, levels = feature_names)

  # Randomize ONCE before plotting
  if (randomize) {
    df <- df[sample(nrow(df)), ]
  }

  # Compute color limits
  if (is.character(color_limits)) {
    if (color_limits == "global") {
      lim <- compute_color_limits(df$Value, symmetric = TRUE, quantile = 0.99)
      use_free_scale <- FALSE
    } else if (color_limits == "individual") {
      lim <- NULL
      use_free_scale <- TRUE
    } else {
      stop("color_limits must be 'global', 'individual', or numeric vector", call. = FALSE)
    }
  } else {
    lim <- color_limits
    use_free_scale <- FALSE
  }

  # Determine legend label based on projection type
  legend_label <- switch(projection,
    "zdb" = "Expression",
    "db" = "Factor Activity",
    "adb" = "Pathway Activity",
    "Expression" # default fallback
  )

  # Determine default title based on projection type
  default_title <- switch(projection,
    "zdb" = "Gene Expression",
    "db" = "Latent Factor Activity",
    "adb" = "Pathway Activity",
    "Feature Expression" # default fallback
  )

  # Create plot
  p <- ggplot2::ggplot(df, ggplot2::aes(x = Dim1, y = Dim2, color = Value)) +
    ggplot2::geom_point(size = point_size, alpha = alpha) +
    ggplot2::facet_wrap(~Feature, ncol = ncol)

  # Add color scale
  if (use_free_scale) {
    # Per-facet limits - need to compute per feature
    p <- p + scale_color_gedi_diverging(name = legend_label)
  } else {
    p <- p + scale_color_gedi_diverging(limits = lim, name = legend_label)
  }

  # Add labels and theme
  emb_name <- if (is.character(embedding)) toupper(embedding) else "Embedding"
  p <- p +
    ggplot2::labs(
      x = paste(emb_name, "1"),
      y = paste(emb_name, "2"),
      title = if (is.null(title)) default_title else title
    ) +
    theme_gedi()

  return(p)
}


#' Plot Two-Feature Comparison
#'
#' Compares two features by computing their difference or correlation in
#' the projected space. Mathematically grounded for GEDI's log-space representation.
#'
#' @param model GEDI model object
#' @param gene1 Character or integer, first gene name or index
#' @param gene2 Character or integer, second gene name or index
#' @param comparison Character, type of comparison ("difference" or "correlation")
#' @param embedding Character specifying embedding type ("umap", "pca") or
#'   a custom N x 2 matrix
#' @param projection Character, type of projection ("zdb" or "db")
#' @param color_limits Numeric vector c(low, high) or NULL for auto-compute
#' @param randomize Logical, whether to randomize point order
#' @param point_size Numeric, size of points
#' @param alpha Numeric, transparency of points
#' @param title Character, plot title
#'
#' @return ggplot2 object
#'
#' @details
#' For comparison = "difference":
#' Computes \verb{(Z[gene1,] - Z[gene2,]) * D * B}, equivalent to \verb{ZDB[gene1,] - ZDB[gene2,]}.
#' In log-space, this represents log(gene1/gene2) in the original count space.
#' Positive values indicate gene1 > gene2, negative indicates gene2 > gene1.
#'
#' @examples
#' \donttest{
#' if (requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'   gene1 <- rownames(pbmc_small)[1]
#'   gene2 <- rownames(pbmc_small)[2]
#'   plot_feature_ratio(model, gene1, gene2, comparison = "difference",
#'                      embedding = "pca")
#' }
#' }
#'
#' @export
plot_feature_ratio <- function(model,
                               gene1,
                               gene2,
                               comparison = "difference",
                               embedding = "umap",
                               projection = "zdb",
                               color_limits = NULL,
                               randomize = TRUE,
                               point_size = 0.3,
                               alpha = 0.9,
                               title = NULL) {
  # Validate model
  if (is.null(model$params)) {
    stop("Model not trained. Run $train() first.", call. = FALSE)
  }

  # Validate comparison type
  if (!comparison %in% c("difference", "correlation")) {
    stop("comparison must be 'difference' or 'correlation'", call. = FALSE)
  }

  # Get embedding coordinates
  if (is.character(embedding)) {
    if (embedding == "umap") {
      emb_mat <- model$embeddings$umap
    } else if (embedding == "pca") {
      emb_mat <- model$embeddings$pca[, 1:2]
    } else {
      stop("embedding must be 'umap', 'pca', or a matrix", call. = FALSE)
    }
  } else {
    emb_mat <- as.matrix(embedding[, 1:2])
  }

  # Convert gene names to indices
  geneIDs <- model$metadata$geneIDs

  if (is.character(gene1)) {
    gene1_idx <- match(gene1, geneIDs)
    if (is.na(gene1_idx)) stop("gene1 not found: ", gene1, call. = FALSE)
    gene1_name <- gene1
  } else {
    gene1_idx <- as.integer(gene1)
    gene1_name <- geneIDs[gene1_idx]
  }

  if (is.character(gene2)) {
    gene2_idx <- match(gene2, geneIDs)
    if (is.na(gene2_idx)) stop("gene2 not found: ", gene2, call. = FALSE)
    gene2_name <- gene2
  } else {
    gene2_idx <- as.integer(gene2)
    gene2_name <- geneIDs[gene2_idx]
  }

  # Compute comparison
  if (comparison == "difference") {
    # Compute (Z[gene1,] - Z[gene2,]) * D * B
    Z <- model$params$Z
    weight_diff <- Z[gene1_idx, ] - Z[gene2_idx, ] # K x 1

    ratio_values <- compute_feature_projection(
      feature_weights = weight_diff,
      D = model$params$D,
      Bi_list = model$params$Bi,
      verbose = 0
    )

    legend_title <- paste0(gene1_name, " - ", gene2_name)
  } else if (comparison == "correlation") {
    # Future: 2D scatter or correlation coefficient per cell
    stop("correlation comparison not yet implemented", call. = FALSE)
  }

  # Compute color limits
  if (is.null(color_limits)) {
    color_limits <- compute_color_limits(ratio_values, symmetric = TRUE, quantile = 0.99)
  }

  # Create plot using plot_embedding
  emb_name <- if (is.character(embedding)) toupper(embedding) else "Embedding"

  if (is.null(title)) {
    title <- paste("Feature Comparison:", gene1_name, "vs", gene2_name)
  }

  p <- .plot_embedding_base(
    embedding = emb_mat,
    color = ratio_values,
    color_limits = color_limits,
    randomize = randomize,
    point_size = point_size,
    alpha = alpha,
    xlab = paste(emb_name, "1"),
    ylab = paste(emb_name, "2"),
    title = title,
    legend_title = legend_title
  )

  return(p)
}


# R/gedi_plot_dynamics.R

#' Plot Vector Field from Dynamics Analysis
#'
#' Visualizes vector fields showing cell state transitions. Uses binned
#' aggregation for cleaner visualization without overplotting.
#'
#' @param dynamics_svd Result from model$dynamics$vector_field() or similar
#' @param color Vector of length N for coloring arrows, or NULL
#' @param alpha Vector of length N or scalar for arrow transparency
#' @param n_bins Integer, number of bins per dimension for aggregation
#' @param min_per_bin Integer, minimum observations required per bin
#' @param randomize Logical, whether to randomize data order
#' @param arrow_size Numeric, size of arrow lines
#' @param arrow_length Numeric, length of arrow heads (in cm)
#' @param arrow_color Character, color for arrows (if color is NULL)
#' @param dims Integer vector of length 2, which dimensions to plot
#' @param xlab Character, x-axis label
#' @param ylab Character, y-axis label
#' @param title Character, plot title
#'
#' @return ggplot2 object
#'
#' @examples
#' \donttest{
#' # Build a tiny multi-sample fixture with a sample-level prior H so that
#' # model$dynamics$vector_field() is available.
#' set.seed(1)
#' n_genes <- 80; n_cells <- 60; n_samples <- 3
#' M <- Matrix::Matrix(
#'   matrix(stats::rpois(n_genes * n_cells, 5), n_genes, n_cells),
#'   sparse = TRUE
#' )
#' rownames(M) <- paste0("G", seq_len(n_genes))
#' colnames(M) <- paste0("C", seq_len(n_cells))
#' samples <- factor(rep(paste0("S", seq_len(n_samples)),
#'                       each = n_cells / n_samples))
#' H <- matrix(c(1, 0, 0,  0, 1, 0), nrow = 2, byrow = TRUE)
#' colnames(H) <- paste0("S", seq_len(n_samples))
#' rownames(H) <- c("cond_a", "cond_b")
#'
#' model <- CreateGEDIObject(Samples = samples, M = M, K = 3, H = H,
#'                           verbose = 0)
#' model$train(iterations = 5)
#' vf <- model$dynamics$vector_field(start.cond = c(1, 0),
#'                                   end.cond   = c(0, 1))
#' plot_vector_field(vf)
#' }
#'
#' @export
plot_vector_field <- function(dynamics_svd,
                              color = NULL,
                              alpha = 1,
                              n_bins = 50,
                              min_per_bin = 10,
                              randomize = TRUE,
                              arrow_size = 0.5,
                              arrow_length = 0.15,
                              arrow_color = "black",
                              dims = c(1, 2),
                              xlab = NULL,
                              ylab = NULL,
                              title = NULL) {
  # Validate input
  if (!inherits(dynamics_svd, "gedi_dynamics_svd")) {
    stop("dynamics_svd must be output from model$dynamics functions", call. = FALSE)
  }

  if (length(dims) != 2 || !all(dims %in% 1:ncol(dynamics_svd$v))) {
    stop("dims must be a vector of 2 valid dimension indices", call. = FALSE)
  }

  # Extract start and end points
  v_matrix <- dynamics_svd$v
  indices <- dynamics_svd$indices

  N <- length(indices$embedding)

  # Get start points (original embedding)
  start_idx <- indices$embedding
  Dim1 <- v_matrix[start_idx, dims[1]]
  Dim2 <- v_matrix[start_idx, dims[2]]

  # Get end points (vector field)
  if ("vector_field" %in% names(indices)) {
    # Standard vector field: start at 1:N, end at (N+1):(2N)
    end_idx <- indices$vector_field[indices$vector_field > N]
    To1 <- v_matrix[end_idx, dims[1]]
    To2 <- v_matrix[end_idx, dims[2]]
  } else {
    stop("dynamics_svd does not contain vector_field information", call. = FALSE)
  }

  # Handle color
  if (is.null(color)) {
    color <- rep(1, N)
    use_color <- FALSE
  } else {
    if (length(color) != N) {
      stop("color must have same length as number of cells", call. = FALSE)
    }
    use_color <- TRUE
  }

  # Handle alpha
  if (length(alpha) == 1) {
    alpha <- rep(alpha, N)
  } else if (length(alpha) != N) {
    stop("alpha must be scalar or have length N", call. = FALSE)
  }

  # Call C++ aggregation function
  agg_df <- aggregate_vectors(
    Dim1 = Dim1,
    Dim2 = Dim2,
    To1 = To1,
    To2 = To2,
    color = color,
    alpha = alpha,
    n_bins = n_bins,
    min_per_bin = min_per_bin
  )

  # Randomize if requested
  if (randomize && nrow(agg_df) > 0) {
    agg_df <- agg_df[sample(nrow(agg_df)), ]
  }

  # Detect if color is numeric
  is_numeric <- use_color && is.numeric(color) && length(unique(color)) > 10

  # Create base plot
  if (!use_color || length(unique(agg_df$Color)) == 1) {
    # Uniform color
    p <- ggplot2::ggplot(agg_df, ggplot2::aes(
      x = Dim1, y = Dim2,
      xend = To1, yend = To2,
      alpha = Alpha
    )) +
      ggplot2::geom_segment(
        arrow = ggplot2::arrow(length = ggplot2::unit(arrow_length, "cm")),
        color = arrow_color,
        linewidth = arrow_size
      ) +
      ggplot2::scale_alpha_identity()
  } else if (is_numeric) {
    # Continuous color
    lim <- compute_color_limits(agg_df$Color, symmetric = TRUE, quantile = 0.99)

    p <- ggplot2::ggplot(agg_df, ggplot2::aes(
      x = Dim1, y = Dim2,
      xend = To1, yend = To2,
      color = Color, alpha = Alpha
    )) +
      ggplot2::geom_segment(
        arrow = ggplot2::arrow(length = ggplot2::unit(arrow_length, "cm")),
        linewidth = arrow_size
      ) +
      scale_color_gedi_diverging(limits = lim, name = "Value") +
      ggplot2::scale_alpha_identity()
  } else {
    # Discrete color
    p <- ggplot2::ggplot(agg_df, ggplot2::aes(
      x = Dim1, y = Dim2,
      xend = To1, yend = To2,
      color = factor(Color), alpha = Alpha
    )) +
      ggplot2::geom_segment(
        arrow = ggplot2::arrow(length = ggplot2::unit(arrow_length, "cm")),
        linewidth = arrow_size
      ) +
      scale_color_gedi_discrete(name = "Group") +
      ggplot2::scale_alpha_identity()
  }

  # Add labels
  if (is.null(xlab)) xlab <- paste("LV", dims[1])
  if (is.null(ylab)) ylab <- paste("LV", dims[2])
  if (is.null(title)) {
    if (!is.null(dynamics_svd$metadata$method)) {
      title <- paste("Vector Field:", dynamics_svd$metadata$method)
    } else {
      title <- "Vector Field"
    }
  }

  p <- p +
    ggplot2::labs(x = xlab, y = ylab, title = title) +
    theme_gedi()

  return(p)
}


# R/gedi_plot_diagnostics.R

#' Plot Dispersion Analysis
#'
#' Visualizes the relationship between expected and observed variance
#' for count data. Useful for assessing model fit quality.
#'
#' @param dispersion_df Data frame from compute_dispersion() with columns:
#'   Expected_Var, Observed_Var, Sample, n, bin
#' @param show_identity Logical, whether to show y=x identity line
#' @param point_size Numeric, size of points
#' @param alpha Numeric, transparency of points
#' @param title Character, plot title
#'
#' @return ggplot2 object
#'
#' @examples
#' \donttest{
#' if (requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   M <- pbmc_small@assays$RNA@counts
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = M,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 5)
#'   disp <- model$imputed$dispersion(M)
#'   plot_dispersion(disp)
#' }
#' }
#'
#' @export
plot_dispersion <- function(dispersion_df,
                            show_identity = TRUE,
                            point_size = 0.1,
                            alpha = 0.5,
                            title = "Dispersion Analysis") {
  # Validate input
  required_cols <- c("Expected_Var", "Observed_Var", "Sample")
  if (!all(required_cols %in% colnames(dispersion_df))) {
    stop("dispersion_df must contain columns: ",
      paste(required_cols, collapse = ", "),
      call. = FALSE
    )
  }

  # Create plot
  p <- ggplot2::ggplot(
    dispersion_df,
    ggplot2::aes(x = Expected_Var, y = Observed_Var, color = Sample)
  ) +
    ggplot2::geom_point(size = point_size, alpha = alpha) +
    ggplot2::scale_x_log10() +
    ggplot2::scale_y_log10()

  # Add identity line
  if (show_identity) {
    p <- p + ggplot2::geom_abline(
      intercept = 0, slope = 1,
      linetype = "dashed", color = "gray30"
    )
  }

  # Add labels and theme
  p <- p +
    scale_color_gedi_discrete(name = "Sample") +
    ggplot2::labs(
      x = "Expected Variance",
      y = "Observed Variance",
      title = title
    ) +
    theme_gedi()

  return(p)
}


#' Plot Training Convergence
#'
#' Visualizes convergence of model parameters during training.
#' Supports multiple layout styles for different use cases.
#'
#' @param model GEDI model object
#' @param layout Character, layout style:
#'   \itemize{
#'     \item "faceted": Single plot with facet_wrap (default, best for reports)
#'     \item "separate": List of individual plots (for interactive exploration)
#'     \item "compact": Two-panel plot (global vs sample-specific)
#'   }
#' @param params Character vector, which parameters to include. Options:
#'   "Z", "A", "o", "Bi", "Qi", "oi", "si", "Rk", "Ro", "sigma2".
#'   NULL means all available.
#' @param log_scale Logical, use log10 scale for y-axis
#' @param smooth Logical, add smooth trend line
#' @param title Character, plot title
#'
#' @return ggplot2 object (for "faceted" or "compact") or list of ggplot2 objects (for "separate")
#'
#' @examples
#' \donttest{
#' if (requireNamespace("SeuratObject", quietly = TRUE)) {
#'   pbmc_small <- SeuratObject::pbmc_small
#'   model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M       = pbmc_small@assays$RNA@counts,
#'     K       = 3,
#'     verbose = 0
#'   )
#'   model$train(iterations = 10, track_interval = 2)
#'   plot_convergence(model)
#'   plots <- plot_convergence(model, layout = "separate")
#' }
#' }
#'
#' @export
plot_convergence <- function(model,
                             layout = c("faceted", "separate", "compact"),
                             params = NULL,
                             log_scale = TRUE,
                             smooth = FALSE,
                             title = "Training Convergence") {
  layout <- match.arg(layout)

  # Validate model
  if (is.null(model$tracking)) {
    stop("No tracking data available. Model may not have been trained with tracking.",
      call. = FALSE
    )
  }

  tracking <- model$tracking
  aux <- model$aux

  # Determine which parameters are available
  available_params <- c()
  if (!all(is.na(tracking$dZ))) available_params <- c(available_params, "Z")
  if (!all(is.na(tracking$dA))) available_params <- c(available_params, "A")
  if (!all(is.na(tracking$do))) available_params <- c(available_params, "o")
  if (!all(is.na(tracking$dRo))) available_params <- c(available_params, "Ro")
  if (!all(is.na(tracking$sigma2))) available_params <- c(available_params, "sigma2")

  # Check sample-specific parameters
  has_Bi <- !all(sapply(tracking$dBi, function(x) all(is.na(x))))
  has_Qi <- !all(sapply(tracking$dQi, function(x) all(is.na(x))))
  has_oi <- !all(sapply(tracking$doi, function(x) all(is.na(x))))
  has_si <- !all(sapply(tracking$dsi, function(x) all(is.na(x))))
  has_Rk <- !all(sapply(tracking$dRk, function(x) all(is.na(x))))

  if (has_Bi) available_params <- c(available_params, "Bi")
  if (has_Qi) available_params <- c(available_params, "Qi")
  if (has_oi) available_params <- c(available_params, "oi")
  if (has_si) available_params <- c(available_params, "si")
  if (has_Rk) available_params <- c(available_params, "Rk")

  # Filter params if specified
  if (!is.null(params)) {
    params <- intersect(params, available_params)
    if (length(params) == 0) {
      stop("None of the requested parameters are available in tracking data",
        call. = FALSE
      )
    }
  } else {
    params <- available_params
  }

  # Build data frames for each parameter type
  plot_list <- list()

  # Global parameters (Z, A, o, Ro, sigma2)
  if (any(c("Z", "A", "o", "Ro") %in% params)) {
    global_df <- data.frame()
    l <- length(tracking$dZ)

    if ("Z" %in% params) {
      global_df <- rbind(global_df, data.frame(
        Iteration = 1:l,
        RMSD = tracking$dZ,
        Parameter = "Z"
      ))
    }
    if ("A" %in% params) {
      global_df <- rbind(global_df, data.frame(
        Iteration = 1:l,
        RMSD = tracking$dA,
        Parameter = "A"
      ))
    }
    if ("o" %in% params) {
      global_df <- rbind(global_df, data.frame(
        Iteration = 1:l,
        RMSD = tracking$do,
        Parameter = "o"
      ))
    }
    if ("Ro" %in% params) {
      global_df <- rbind(global_df, data.frame(
        Iteration = 1:l,
        RMSD = tracking$dRo,
        Parameter = "Ro"
      ))
    }

    # Remove NA rows
    global_df <- global_df[!is.na(global_df$RMSD), ]

    if (nrow(global_df) > 0) {
      p_global <- ggplot2::ggplot(
        global_df,
        ggplot2::aes(x = Iteration, y = RMSD, color = Parameter)
      ) +
        ggplot2::geom_line() +
        ggplot2::geom_point(size = 1)

      if (log_scale) p_global <- p_global + ggplot2::scale_y_log10()
      if (smooth) p_global <- p_global + ggplot2::geom_smooth(se = FALSE, linewidth = 0.5)

      p_global <- p_global +
        scale_color_gedi_discrete(name = "Parameter") +
        ggplot2::labs(
          x = "Iteration", y = "RMSD",
          title = "Global Parameters"
        ) +
        theme_gedi()

      plot_list$global <- p_global
    }
  }

  # Sample-specific parameters (Bi, Qi, oi, si)
  sample_params <- intersect(c("Bi", "Qi", "oi", "si"), params)
  if (length(sample_params) > 0) {
    for (param in sample_params) {
      param_df <- data.frame()
      l <- length(tracking[[paste0("d", param)]][[1]])

      for (i in 1:aux$numSamples) {
        param_df <- rbind(param_df, data.frame(
          Iteration = 1:l,
          RMSD = tracking[[paste0("d", param)]][[i]],
          Sample = model$metadata$sampleIDs[i]
        ))
      }

      param_df <- param_df[!is.na(param_df$RMSD), ]

      if (nrow(param_df) > 0) {
        p <- ggplot2::ggplot(
          param_df,
          ggplot2::aes(x = Iteration, y = RMSD, color = Sample)
        ) +
          ggplot2::geom_line() +
          ggplot2::geom_point(size = 1)

        if (log_scale) p <- p + ggplot2::scale_y_log10()
        if (smooth) p <- p + ggplot2::geom_smooth(se = FALSE, linewidth = 0.5)

        p <- p +
          scale_color_gedi_discrete(name = "Sample") +
          ggplot2::labs(
            x = "Iteration", y = "RMSD",
            title = paste0("Parameter: ", param)
          ) +
          theme_gedi()

        plot_list[[param]] <- p
      }
    }
  }

  # Rk parameters (if L > 0)
  if ("Rk" %in% params && has_Rk) {
    Rk_df <- data.frame()
    l <- length(tracking$dRk[[1]])

    for (k in 1:aux$K) {
      Rk_df <- rbind(Rk_df, data.frame(
        Iteration = 1:l,
        RMSD = tracking$dRk[[k]],
        Factor = paste0("LV", k)
      ))
    }

    # Add Ro
    Rk_df <- rbind(Rk_df, data.frame(
      Iteration = 1:l,
      RMSD = tracking$dRo,
      Factor = "o"
    ))

    Rk_df <- Rk_df[!is.na(Rk_df$RMSD), ]

    if (nrow(Rk_df) > 0) {
      p_Rk <- ggplot2::ggplot(
        Rk_df,
        ggplot2::aes(x = Iteration, y = RMSD, color = Factor)
      ) +
        ggplot2::geom_line() +
        ggplot2::geom_point(size = 1)

      if (log_scale) p_Rk <- p_Rk + ggplot2::scale_y_log10()
      if (smooth) p_Rk <- p_Rk + ggplot2::geom_smooth(se = FALSE, linewidth = 0.5)

      p_Rk <- p_Rk +
        scale_color_gedi_discrete(name = "Factor") +
        ggplot2::labs(
          x = "Iteration", y = "RMSD",
          title = "Covariate Effects (Rk)"
        ) +
        theme_gedi()

      plot_list$Rk <- p_Rk
    }
  }

  # sigma2
  if ("sigma2" %in% params && !all(is.na(tracking$sigma2))) {
    sigma2_df <- data.frame(
      Iteration = 1:length(tracking$sigma2),
      sigma2 = tracking$sigma2
    )

    p_sigma2 <- ggplot2::ggplot(
      sigma2_df,
      ggplot2::aes(x = Iteration, y = sigma2)
    ) +
      ggplot2::geom_line(color = "#377EB8") +
      ggplot2::geom_point(size = 1, color = "#377EB8")

    if (log_scale) p_sigma2 <- p_sigma2 + ggplot2::scale_y_log10()
    if (smooth) p_sigma2 <- p_sigma2 + ggplot2::geom_smooth(se = FALSE, linewidth = 0.5)

    p_sigma2 <- p_sigma2 +
      ggplot2::labs(
        x = "Iteration", y = expression(sigma^2),
        title = "Model Variance"
      ) +
      theme_gedi()

    plot_list$sigma2 <- p_sigma2
  }

  # Return based on layout
  if (layout == "separate") {
    return(plot_list)
  } else if (layout == "faceted") {
    # Combine all into single faceted plot
    combined_df <- data.frame()

    # Add global params
    if (!is.null(plot_list$global)) {
      global_data <- plot_list$global$data
      global_data$Type <- "Global"
      global_data$Group <- global_data$Parameter
      combined_df <- rbind(combined_df, global_data[, c("Iteration", "RMSD", "Parameter", "Type", "Group")])
    }

    # Add sample-specific params
    for (param in sample_params) {
      if (!is.null(plot_list[[param]])) {
        param_data <- plot_list[[param]]$data
        param_data$Parameter <- param
        param_data$Type <- param
        names(param_data)[names(param_data) == "Sample"] <- "Group"
        combined_df <- rbind(combined_df, param_data[, c("Iteration", "RMSD", "Parameter", "Type", "Group")])
      }
    }

    # Add Rk
    if (!is.null(plot_list$Rk)) {
      Rk_data <- plot_list$Rk$data
      Rk_data$Parameter <- "Rk"
      Rk_data$Type <- "Rk"
      names(Rk_data)[names(Rk_data) == "Factor"] <- "Group"
      combined_df <- rbind(combined_df, Rk_data[, c("Iteration", "RMSD", "Parameter", "Type", "Group")])
    }

    # Add sigma2
    if (!is.null(plot_list$sigma2)) {
      sigma2_data <- plot_list$sigma2$data
      sigma2_data$RMSD <- sigma2_data$sigma2
      sigma2_data$Parameter <- "sigma2"
      sigma2_data$Type <- "sigma2"
      sigma2_data$Group <- "sigma2"
      combined_df <- rbind(combined_df, sigma2_data[, c("Iteration", "RMSD", "Parameter", "Type", "Group")])
    }

    if (nrow(combined_df) == 0) {
      stop("No valid tracking data to plot", call. = FALSE)
    }

    p <- ggplot2::ggplot(
      combined_df,
      ggplot2::aes(x = Iteration, y = RMSD, color = Group)
    ) +
      ggplot2::geom_line() +
      ggplot2::geom_point(size = 0.5) +
      ggplot2::facet_wrap(~Type, scales = "free_y")

    if (!log_scale) {
      p <- p + ggplot2::scale_y_continuous()
    } else {
      p <- p + ggplot2::scale_y_log10()
    }

    p <- p +
      scale_color_gedi_discrete(name = "Group") +
      ggplot2::labs(x = "Iteration", y = "RMSD", title = title) +
      theme_gedi()

    return(p)
  } else if (layout == "compact") {
    # Two-panel: global + sample-specific
    stop("Compact layout not yet implemented. Use 'faceted' or 'separate'.",
      call. = FALSE
    )
  }
}

# tests/testthat/test-plot-dispersion.R
# -----------------------------------------------------------
# Tests for plot_dispersion(). Can use a hand-crafted data.frame.
# -----------------------------------------------------------

test_that("plot_dispersion returns a ggplot with valid data.frame", {
  disp_df <- data.frame(
    Expected_Var = runif(100, 0.1, 10),
    Observed_Var = runif(100, 0.1, 10),
    Sample       = rep(c("S1", "S2"), each = 50)
  )
  p <- plot_dispersion(disp_df)
  expect_s3_class(p, "ggplot")
})

test_that("plot_dispersion errors on missing columns", {
  bad_df <- data.frame(x = 1:10, y = 1:10)
  expect_error(plot_dispersion(bad_df))
})

test_that("plot_dispersion show_identity parameter works", {
  disp_df <- data.frame(
    Expected_Var = runif(50, 0.1, 5),
    Observed_Var = runif(50, 0.1, 5),
    Sample       = rep("S1", 50)
  )
  p1 <- plot_dispersion(disp_df, show_identity = TRUE)
  p2 <- plot_dispersion(disp_df, show_identity = FALSE)
  expect_s3_class(p1, "ggplot")
  expect_s3_class(p2, "ggplot")
})

# ==============================================================================
# GEDI Pathway Analysis Functions
# Compute pathway-factor associations in dense and sparse forms
# ==============================================================================

#' Compute Dense Pathway-Factor Associations
#'
#' @description
#' Computes pathway-factor association matrix A in the original pathway space
#' by rotating the learned A matrix from the reduced SVD space back to the
#' original pathway annotations.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#'
#' @return Dense matrix (num_pathways x K) where rows correspond to original
#'   pathways and columns to latent factors. Positive values indicate pathways
#'   that are enriched in a factor, negative values indicate depletion.
#'
#' @details
#' The model internally learns A in a reduced P-dimensional space (where P is
#' determined by SVD of the C matrix). This function maps A back to the original
#' pathway space using the C.rotation matrix:
#' 
#' \deqn{A_{full} = C.rotation \times A_{reduced}}
#' 
#' This provides biological interpretability by showing how each original
#' pathway/gene set relates to each latent factor.
#'
#' @keywords internal
#' @noRd
compute_dense_A <- function(self, private) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if C prior was provided
  if (self$aux$P == 0) {
    stop("Cannot compute pathway associations: no gene-level prior (C) was provided during setup.",
         call. = FALSE)
  }
  
  # Check if aux_static exists
  if (is.null(private$.aux_static)) {
    stop("Missing auxiliary data. Model may not be properly initialized.",
         call. = FALSE)
  }
  
  # Check if inputC exists (need original pathway names)
  if (is.null(private$.aux_static$inputC)) {
    stop("Original C matrix not found. Cannot determine pathway names.",
         call. = FALSE)
  }
  
  if (private$.verbose >= 1) {
    message("Computing dense pathway-factor associations...")
  }
  
  # Compute A in original pathway space: C.rotation %*% A
  # C.rotation is num_pathways x P
  # A is P x K
  # Result is num_pathways x K
  A_full <- private$.aux_static$C.rotation %*% self$params$A
  
  # Add row and column names
  rownames(A_full) <- colnames(private$.aux_static$inputC)
  colnames(A_full) <- paste0("LV", seq_len(self$aux$K))
  
  if (private$.verbose >= 1) {
    message(sprintf("  Computed %d pathways x %d factors", 
                nrow(A_full), ncol(A_full)))
  }
  
  return(A_full)
}


#' Compute Sparse Pathway-Factor Associations
#'
#' @description
#' Re-fits pathway-factor associations using LASSO regression (L1 regularization)
#' to obtain sparse associations. This is useful for interpretation as it
#' identifies the most important pathways for each factor while setting many
#' coefficients to exactly zero.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param C Optional gene x pathway matrix. If NULL (default), uses the original
#'   C matrix provided during model setup. Can provide a different pathway
#'   database for post-hoc interpretation.
#'
#' @return Sparse matrix (num_pathways x K) with many zero entries. Non-zero
#'   values indicate pathways that are strongly associated with each factor.
#'
#' @details
#' For each latent factor k, this function:
#' \enumerate{
#'   \item Extracts the gene expression pattern Z[, k]
#'   \item Performs L1-regularized (LASSO) regression: Z[, k] ~ C
#'   \item Uses cross-validation to select optimal regularization strength
#'   \item Extracts non-zero pathway coefficients
#' }
#' 
#' The result is typically much sparser than the dense A from model training,
#' making it easier to identify key biological processes.
#' 
#' Requires the \code{glmnet} package to be installed.
#'
#' @examples
#' \donttest{
#' # Use original pathways (sparse version)
#' A_sparse <- model$pathway_associations$sparse()
#' 
#' # Test with KEGG pathways post-hoc
#' kegg_C <- create_kegg_matrix(rownames(counts))
#' A_kegg <- model$pathway_associations$sparse(C = kegg_C)
#' }
#'
#' @keywords internal
#' @noRd
compute_sparse_A <- function(self, private, C = NULL) {
  
  # Check if glmnet is available
  if (!requireNamespace("glmnet", quietly = TRUE)) {
    stop("Package 'glmnet' is required for sparse pathway associations.\n",
         "  Install it with: install.packages('glmnet')",
         call. = FALSE)
  }
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Handle C matrix
  if (is.null(C)) {
    # Use original C if available
    if (self$aux$P == 0) {
      stop("No gene-level prior (C) was provided during setup.\n",
           "  You must provide a C matrix to compute sparse associations.",
           call. = FALSE)
    }
    
    if (is.null(private$.aux_static) || is.null(private$.aux_static$inputC)) {
      stop("Original C matrix not found. Cannot compute sparse associations.",
           call. = FALSE)
    }
    
    C <- private$.aux_static$inputC
    
    if (private$.verbose >= 1) {
      message("Computing sparse pathway-factor associations (using original C)...")
    }
    
  } else {
    # User provided custom C
    C <- as.matrix(C)
    
    # Validate dimensions
    if (nrow(C) != self$aux$J) {
      stop(sprintf("Custom C matrix must have J rows (got %d, expected %d)",
                   nrow(C), self$aux$J),
           call. = FALSE)
    }
    
    if (ncol(C) == 0) {
      stop("Custom C matrix must have at least one column (pathway)",
           call. = FALSE)
    }
    
    if (private$.verbose >= 1) {
      message(sprintf("Computing sparse pathway-factor associations (custom C: %d pathways)...",
                  ncol(C)))
    }
  }
  
  # Get Z matrix
  Z <- self$params$Z
  K <- self$aux$K
  num_pathways <- ncol(C)
  
  # Initialize result matrix
  A_sparse <- matrix(NA, nrow = num_pathways, ncol = K)
  
  # For each latent factor, run LASSO regression
  if (private$.verbose >= 1) {
    message(sprintf("  Running LASSO for %d factors...", K))
  }
  
  for (k in 1:K) {
    if (private$.verbose >= 2) {
      message(sprintf("    Factor %d/%d", k, K))
    }
    
    # Run cross-validated LASSO
    # Note: intercept = FALSE because Z columns should be centered
    tryCatch({
      fit <- glmnet::cv.glmnet(
        x = C,
        y = Z[, k],
        intercept = FALSE,
        standardize = TRUE,
        nfolds = 10
      )
      
      # Extract coefficients at lambda.1se (more regularized, sparser)
      # This gives more interpretable results than lambda.min
      coefs <- as.vector(coef(fit, s = "lambda.1se"))
      
      # Remove intercept (first element) even though we set intercept=FALSE
      # glmnet still returns it
      A_sparse[, k] <- coefs[-1]
      
    }, error = function(e) {
      warning(sprintf("LASSO failed for factor %d: %s", k, e$message),
              call. = FALSE)
      A_sparse[, k] <- rep(0, num_pathways)
    })
  }
  
  # Add row and column names
  if (!is.null(colnames(C))) {
    rownames(A_sparse) <- colnames(C)
  } else {
    rownames(A_sparse) <- paste0("Pathway", seq_len(num_pathways))
  }
  colnames(A_sparse) <- paste0("LV", seq_len(K))
  
  # Report sparsity
  if (private$.verbose >= 1) {
    nonzero_per_factor <- colSums(A_sparse != 0)
    total_nonzero <- sum(A_sparse != 0)
    total_possible <- num_pathways * K
    sparsity_pct <- 100 * (1 - total_nonzero / total_possible)
    
    message(sprintf("  Sparsity: %.1f%% (%d / %d nonzero)",
                sparsity_pct, total_nonzero, total_possible))
    message(sprintf("  Nonzero per factor: min=%d, median=%d, max=%d",
                min(nonzero_per_factor),
                median(nonzero_per_factor),
                max(nonzero_per_factor)))
  }
  
  return(A_sparse)
}


#' Create Pathway Associations Accessor Object
#'
#' @description
#' Creates a list-like object with methods to access pathway-factor associations
#' in both dense and sparse forms. This is used internally by the
#' pathway_associations active binding in the GEDI class.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#'
#' @return S3 object of class "gedi_pathway_associations" with two methods:
#'   \itemize{
#'     \item \code{$dense()} - Returns dense pathway associations
#'     \item \code{$sparse(C = NULL)} - Returns sparse pathway associations
#'   }
#'
#' @keywords internal
#' @noRd
create_pathway_associations_accessor <- function(self, private) {
  
  structure(
    list(
      dense = function() {
        compute_dense_A(self, private)
      },
      
      sparse = function(C = NULL) {
        compute_sparse_A(self, private, C)
      }
    ),
    class = "gedi_pathway_associations"
  )
}


#' Print Method for Pathway Associations Accessor
#'
#' @param x Object of class gedi_pathway_associations
#' @param ... Additional arguments (ignored)
#'
#' @return Invisibly returns \code{x}.
#' @method print gedi_pathway_associations
#' @keywords internal
#' @export
print.gedi_pathway_associations <- function(x, ...) {
  cat("<GEDI Pathway Associations>\n")
  cat("\nAvailable methods:\n")
  cat("  $dense()       - Compute dense pathway-factor associations\n")
  cat("                   (from model training, in original pathway space)\n")
  cat("\n")
  cat("  $sparse(C)     - Compute sparse pathway-factor associations\n")
  cat("                   (LASSO regression for interpretability)\n")
  cat("                   C: optional custom pathway matrix (default: uses original C)\n")
  cat("\nExample usage:\n")
  cat("  A_dense <- model$pathway_associations$dense()\n")
  cat("  A_sparse <- model$pathway_associations$sparse()\n")
  cat("  A_kegg <- model$pathway_associations$sparse(C = kegg_pathways)\n")
  
  invisible(x)
}

# R/gedi_plot_utils.R

#' GEDI Plot Theme
#' @return A \code{ggplot2} theme object.
#' @keywords internal
theme_gedi <- function(base_size = 11) {
  ggplot2::theme_bw(base_size = base_size) +
    ggplot2::theme(
      # Legend
      legend.position = "bottom",
      legend.direction = "horizontal",
      legend.box = "horizontal",
      legend.title = ggplot2::element_text(size = base_size * 0.9, face = "plain", hjust = 0.5),
      legend.text = ggplot2::element_text(size = base_size * 0.8),
      legend.key.height = ggplot2::unit(0.4, "cm"),
      legend.key.width = ggplot2::unit(3, "cm"),
      legend.margin = ggplot2::margin(t = 5, b = 5),
      
      # Title
      plot.title = ggplot2::element_text(size = base_size * 1.2, face = "plain", hjust = 0),
      plot.subtitle = ggplot2::element_text(size = base_size * 0.95, hjust = 0),

      # Facets
      strip.text = ggplot2::element_text(size = base_size * 0.95, face = "plain"),
      strip.background = ggplot2::element_rect(fill = "gray90", color = "black", linewidth = 0.3)
    )
}


#' Diverging Color Scale for GEDI Plots
#' @return A \code{ggplot2} continuous color scale.
#' @keywords internal
scale_color_gedi_diverging <- function(limits = NULL, name = "Value", ...) {
  ggplot2::scale_color_gradientn(
    colors = c("blue", "lightgrey", "red"),
    values = scales::rescale(c(-1, 0, 1)),
    limits = limits,
    oob = scales::squish,
    name = name,
    guide = ggplot2::guide_colorbar(
      title.position = "top",
      title.hjust = 0.5,
      barwidth = ggplot2::unit(4, "cm"),
      barheight = ggplot2::unit(0.4, "cm"),
      frame.colour = "black",
      frame.linewidth = 0.5,
      ticks.colour = "black",
      ticks.linewidth = 0.5
    ),
    ...
  )
}


#' Fill Scale for GEDI Plots
#' @return A \code{ggplot2} continuous fill scale.
#' @keywords internal
scale_fill_gedi_diverging <- function(limits = NULL, name = "Value", ...) {
  ggplot2::scale_fill_gradientn(
    colors = c("blue", "lightgrey", "red"),
    values = scales::rescale(c(-1, 0, 1)),
    limits = limits,
    oob = scales::squish,
    name = name,
    guide = ggplot2::guide_colorbar(
      title.position = "top",
      title.hjust = 0.5,
      barwidth = ggplot2::unit(4, "cm"),
      barheight = ggplot2::unit(0.4, "cm"),
      frame.colour = "black",
      frame.linewidth = 0.5,
      ticks.colour = "black",
      ticks.linewidth = 0.5
    ),
    ...
  )
}


#' Discrete Color Palette for GEDI
#' @return A \code{ggplot2} discrete color scale.
#' @keywords internal
scale_color_gedi_discrete <- function(name = "Group", ...) {
  # Extended palette with 24 distinct colors
  # First 9 are from Set1, followed by additional colorblind-friendly colors
  palette_24 <- c(
    "#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "#FF7F00",  # Set1 base
    "#FFFF33", "#A65628", "#F781BF", "#999999",            # Set1 complete
    "#66C2A5", "#FC8D62", "#8DA0CB", "#E78AC3", "#A6D854",  # Set2
    "#FFD92F", "#E5C494", "#B3B3B3",                        # Set2 complete
    "#8DD3C7", "#FFFFB3", "#BEBADA", "#FB8072", "#80B1D3",  # Set3
    "#FDB462", "#B3DE69", "#FCCDE5"                         # Set3 partial
  )

  ggplot2::scale_color_manual(
    values = palette_24,
    name = name,
    ...
  )
}


#' Compute Color Limits from Data
#' @return A numeric vector of length 2 with lower and upper limits.
#' @keywords internal
compute_color_limits <- function(values, symmetric = TRUE, quantile = 0.99) {
  if (symmetric) {
    lim <- stats::quantile(abs(values), quantile, na.rm = TRUE)
    c(-lim, lim)
  } else {
    c(min(values, na.rm = TRUE), max(values, na.rm = TRUE))
  }
}

# tests/testthat/test-plot-utils.R
# -----------------------------------------------------------
# Tests for theme_gedi(), scale_color_gedi_*, scale_fill_gedi_*,
# and compute_color_limits(). No model needed.
# -----------------------------------------------------------

# ===========================================================
# 1. theme_gedi
# ===========================================================
test_that("theme_gedi returns a ggplot2 theme", {
  th <- theme_gedi()
  expect_s3_class(th, "theme")
  expect_s3_class(th, "gg")
})

test_that("theme_gedi respects base_size", {
  th8  <- theme_gedi(base_size = 8)
  th14 <- theme_gedi(base_size = 14)
  expect_s3_class(th8, "theme")
  expect_s3_class(th14, "theme")
})

# ===========================================================
# 2. scale_color_gedi_diverging
# ===========================================================
test_that("scale_color_gedi_diverging returns a ggplot2 scale", {
  sc <- scale_color_gedi_diverging()
  expect_s3_class(sc, "ScaleContinuous")
})

test_that("scale_color_gedi_diverging accepts custom limits", {
  sc <- scale_color_gedi_diverging(limits = c(-5, 5), name = "Test")
  expect_s3_class(sc, "ScaleContinuous")
})

# ===========================================================
# 3. scale_fill_gedi_diverging
# ===========================================================
test_that("scale_fill_gedi_diverging returns a ggplot2 scale", {
  sc <- scale_fill_gedi_diverging()
  expect_s3_class(sc, "ScaleContinuous")
})

# ===========================================================
# 4. scale_color_gedi_discrete
# ===========================================================
test_that("scale_color_gedi_discrete returns a ggplot2 scale", {
  sc <- scale_color_gedi_discrete()
  expect_s3_class(sc, "ScaleDiscrete")
})

# ===========================================================
# 5. compute_color_limits
# ===========================================================
test_that("compute_color_limits returns symmetric limits", {
  vals <- c(-3, -1, 0, 1, 5)
  lim  <- compute_color_limits(vals, symmetric = TRUE, quantile = 1.0)
  expect_length(lim, 2)
  expect_equal(lim[1], -lim[2])  # symmetric about 0
})
 
test_that("compute_color_limits returns asymmetric limits", {
  vals <- c(2, 5, 10, 20)
  lim  <- compute_color_limits(vals, symmetric = FALSE)
  expect_length(lim, 2)
  expect_equal(lim[[1]], 2)
  expect_equal(lim[[2]], 20)
})

test_that("compute_color_limits handles NA values", {
  vals <- c(-1, NA, 3, NA, 5)
  lim  <- compute_color_limits(vals, symmetric = FALSE)
  expect_length(lim, 2)
  expect_false(any(is.na(lim)))
})

# ==============================================================================
# GEDI Dimensionality Reduction - R Wrappers
# Lazy-computed embedding methods with caching
# ==============================================================================

# ==============================================================================
# Embeddings Accessor R6 Class (for lazy evaluation)
# ==============================================================================

EmbeddingsAccessor <- R6Class(
  "EmbeddingsAccessor",
  public = list(
    initialize = function(gedi_self, gedi_private) {
      private$.gedi_self <- gedi_self
      private$.gedi_private <- gedi_private
    },

    compute_umap = function(...) {
      # Custom UMAP with parameters (not cached)
      args <- list(...)
      if (length(args) > 0) {
        return(compute_umap(private$.gedi_self, private$.gedi_private, ...))
      }
      # If no args, return cached default UMAP (same as $umap active binding)
      return(compute_umap_cached(private$.gedi_self, private$.gedi_private))
    },

    print = function() {
      cat("<GEDI Embeddings Accessor>\n")
      cat("\nAvailable embeddings (lazy-computed):\n")
      cat("  $svd   - Factorized SVD\n")
      cat("  $pca   - PCA coordinates\n")
      cat("  $umap  - UMAP embedding (default parameters, cached)\n")
      cat("\nFor custom UMAP parameters, use:\n")
      cat("  $compute_umap(n_neighbors=30, min_dist=0.5, ...)\n")
      cat("\nAccess with: model$embeddings$pca or model$embeddings$umap\n")
      invisible(self)
    }
  ),
  private = list(
    .gedi_self = NULL,
    .gedi_private = NULL
  ),
  active = list(
    svd = function(value) {
      if (!missing(value)) stop("svd is read-only", call. = FALSE)
      compute_svd_factorized(private$.gedi_self, private$.gedi_private)
    },

    pca = function(value) {
      if (!missing(value)) stop("pca is read-only", call. = FALSE)
      compute_pca(private$.gedi_self, private$.gedi_private)
    },

    umap = function(value) {
      if (!missing(value)) stop("umap is read-only", call. = FALSE)
      compute_umap_cached(private$.gedi_self, private$.gedi_private)
    }
  )
)


#' Compute Factorized SVD with Caching
#'
#' @description
#' Computes factorized SVD preserving GEDI structure: SVD(Z) x SVD(middle) x SVD(DB).
#' This approach maintains biological interpretability by respecting the decomposition
#' structure. Results are cached automatically for subsequent access.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return List with three components:
#'   \itemize{
#'     \item d: Singular values (length K vector)
#'     \item u: Left singular vectors (J x K matrix) with gene names
#'     \item v: Right singular vectors (N x K matrix) with cell names
#'   }
#'
#' @keywords internal
#' @noRd
compute_svd_factorized <- function(self, private, force_recompute = FALSE) {
  
  # Check cache
  if (!force_recompute && !is.null(private$.cache$svd)) {
    log_cached("SVD", private$.verbose)
    return(private$.cache$svd)
  }

  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }

  log_start("Factorized SVD", private$.verbose)

  # Call C++ function (silent)
  svd_result <- compute_svd_factorized_cpp(
    Z = self$params$Z,
    D = self$params$D,
    Bi_list = self$params$Bi,
    verbose = 0  # C++ silent
  )

  # Add dimension names
  names(svd_result$d) <- paste0("LV", seq_len(self$aux$K))

  rownames(svd_result$u) <- private$.geneIDs
  colnames(svd_result$u) <- paste0("LV", seq_len(self$aux$K))

  rownames(svd_result$v) <- private$.cellIDs
  colnames(svd_result$v) <- paste0("LV", seq_len(self$aux$K))

  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$svd <- svd_result

  details <- sprintf("%d genes x %d cells x %d factors",
                     nrow(svd_result$u), nrow(svd_result$v), length(svd_result$d))
  log_complete("Factorized SVD", details, private$.verbose)

  return(svd_result)
}


#' Compute PCA Coordinates with Caching
#'
#' @description
#' Computes PCA coordinates as V x diag(d) from the factorized SVD.
#' Uses cached SVD result if available.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return Matrix (N x K) with PCA coordinates, rows = cells, columns = PCs
#'
#' @keywords internal
#' @noRd
compute_pca <- function(self, private, force_recompute = FALSE) {
  
  # Check cache
  if (!force_recompute && !is.null(private$.cache$pca)) {
    log_cached("PCA", private$.verbose)
    return(private$.cache$pca)
  }

  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }

  log_start("PCA coordinates", private$.verbose)

  # Get SVD (uses cache if available)
  svd_result <- compute_svd_factorized(self, private)

  # PCA = V x diag(d)
  pca_coords <- svd_result$v %*% diag(svd_result$d, nrow = length(svd_result$d))

  # Set dimension names
  rownames(pca_coords) <- private$.cellIDs
  colnames(pca_coords) <- paste0("PC", seq_len(ncol(pca_coords)))

  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$pca <- pca_coords

  details <- format_dims(nrow(pca_coords), ncol(pca_coords), "cells", "PCs")
  log_complete("PCA coordinates", details, private$.verbose)

  return(pca_coords)
}


#' Compute UMAP Embedding with Caching (Default Parameters Only)
#'
#' @description
#' Computes UMAP embedding with default parameters and caches the result.
#' Only the default parameter UMAP is cached to save memory.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param force_recompute Logical, if TRUE bypasses cache and recomputes
#'
#' @return Matrix (N x 2) with UMAP coordinates, rows = cells
#'
#' @keywords internal
#' @noRd
compute_umap_cached <- function(self, private, force_recompute = FALSE) {

  # Check cache
  if (!force_recompute && !is.null(private$.cache$umap)) {
    log_cached("UMAP", private$.verbose)
    return(private$.cache$umap)
  }

  # Compute UMAP with default parameters
  umap_result <- compute_umap(self, private,
                               input = "pca",
                               n_neighbors = 15,
                               min_dist = 0.1,
                               n_components = 2,
                               metric = "euclidean",
                               n_threads = 0)

  # Cache result
  if (is.null(private$.cache)) {
    private$.cache <- list()
  }
  private$.cache$umap <- umap_result

  return(umap_result)
}


#' Compute UMAP Embedding
#'
#' @description
#' Computes UMAP embedding using the uwot package. Input can be PCA coordinates,
#' DB projection, or ZDB projection. This function does NOT cache results,
#' allowing for custom parameters. Use compute_umap_cached() for cached default UMAP.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param input Character, input data type: "pca" (default), "db", or "zdb"
#' @param n_neighbors Integer, size of local neighborhood (default: 15)
#' @param min_dist Numeric, minimum distance between points in embedding (default: 0.1)
#' @param n_components Integer, dimensionality of output (default: 2)
#' @param metric Character, distance metric (default: "euclidean")
#' @param n_threads Integer, number of threads (default: 0 = auto)
#' @param ... Additional arguments passed to uwot::umap()
#'
#' @return Matrix (N x n_components) with UMAP coordinates, rows = cells
#'
#' @keywords internal
#' @noRd
compute_umap <- function(self, private, 
                         input = "pca",
                         n_neighbors = 15,
                         min_dist = 0.1,
                         n_components = 2,
                         metric = "euclidean",
                         n_threads = 0,
                         ...) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check uwot package
  if (!requireNamespace("uwot", quietly = TRUE)) {
    stop("Package 'uwot' required for UMAP.\n",
         "  Install with: install.packages('uwot')",
         call. = FALSE)
  }
  
  log_start("UMAP embedding", private$.verbose)

  # Get input data
  if (input == "pca") {
    log_info(sprintf("Input: PCA coordinates"), private$.verbose)
    X <- compute_pca(self, private)

  } else if (input == "db") {
    log_info(sprintf("Input: DB projection"), private$.verbose)
    X <- t(compute_DB(self, private))

  } else if (input == "zdb") {
    log_info(sprintf("Input: ZDB projection"), private$.verbose)
    X <- t(compute_ZDB(self, private))

  } else {
    stop("input must be 'pca', 'db', or 'zdb'", call. = FALSE)
  }

  # Log parameters
  log_info(sprintf("Parameters: n_neighbors=%d, min_dist=%.2f, n_components=%d",
                   n_neighbors, min_dist, n_components), private$.verbose)

  # Run UMAP
  umap_result <- uwot::umap(
    X,
    n_neighbors = n_neighbors,
    min_dist = min_dist,
    n_components = n_components,
    metric = metric,
    n_threads = if (n_threads > 0) n_threads else NULL,
    ret_model = FALSE,
    verbose = private$.verbose > 0,
    ...
  )

  # Set dimension names
  rownames(umap_result) <- private$.cellIDs
  colnames(umap_result) <- paste0("UMAP", seq_len(n_components))

  details <- format_dims(nrow(umap_result), ncol(umap_result), "cells", "dimensions")
  log_complete("UMAP embedding", details, private$.verbose)

  return(umap_result)
}


#' Clear Dimensionality Reduction Cache
#'
#' @description
#' Clears cached embedding results (SVD, PCA, UMAP).
#' Useful after model parameters have been updated.
#'
#' @param private Reference to private environment
#' @param what Character vector specifying which cache entries to clear.
#'   Options: "svd", "pca", "umap", or NULL to clear all (default: NULL)
#'
#' @return NULL (invisible)
#'
#' @keywords internal
#' @noRd
clear_dimred_cache <- function(private, what = NULL) {

  if (is.null(private$.cache)) {
    return(invisible(NULL))
  }

  if (is.null(what)) {
    # Clear all dimred cache
    private$.cache$svd <- NULL
    private$.cache$pca <- NULL
    private$.cache$umap <- NULL
    if (private$.verbose > 0) {
      message("Dimensionality reduction cache cleared")
    }
  } else {
    # Clear specific entries
    valid_entries <- c("svd", "pca", "umap")
    invalid <- setdiff(what, valid_entries)
    if (length(invalid) > 0) {
      warning("Invalid cache entries: ", paste(invalid, collapse = ", "))
    }

    for (entry in intersect(what, valid_entries)) {
      private$.cache[[entry]] <- NULL
      if (private$.verbose > 0) {
        message(sprintf("Cache cleared: %s", entry))
      }
    }
  }

  invisible(NULL)
}


#' Get Dimensionality Reduction Cache Status
#'
#' @description
#' Returns information about which embeddings are currently cached.
#'
#' @param private Reference to private environment
#'
#' @return Named logical vector indicating which embeddings are cached
#'
#' @keywords internal
#' @noRd
get_dimred_cache_status <- function(private) {

  if (is.null(private$.cache)) {
    return(c(svd = FALSE, pca = FALSE, umap = FALSE))
  }

  status <- c(
    svd = !is.null(private$.cache$svd),
    pca = !is.null(private$.cache$pca),
    umap = !is.null(private$.cache$umap)
  )

  return(status)
}


#' Get Dimensionality Reduction Cache Memory Usage
#'
#' @description
#' Estimates memory usage of cached embeddings in MB.
#'
#' @param private Reference to private environment
#'
#' @return Named numeric vector with memory usage in MB for each cached embedding
#'
#' @keywords internal
#' @noRd
get_dimred_cache_memory <- function(private) {

  if (is.null(private$.cache)) {
    return(c(svd = 0, pca = 0, umap = 0, Total = 0))
  }

  memory_mb <- c(
    svd = if (!is.null(private$.cache$svd)) {
      object.size(private$.cache$svd) / 1024^2
    } else 0,
    pca = if (!is.null(private$.cache$pca)) {
      object.size(private$.cache$pca) / 1024^2
    } else 0,
    umap = if (!is.null(private$.cache$umap)) {
      object.size(private$.cache$umap) / 1024^2
    } else 0
  )

  memory_mb["Total"] <- sum(memory_mb)

  return(memory_mb)
}

#' gedi: Gene Expression Data Integration
#'
#' @description
#' A memory-efficient implementation for integrating gene expression data from
#' single-cell RNA sequencing experiments. GEDI v2 uses a high-performance C++
#' backend with thin R wrappers to enable analysis of large-scale single-cell
#' datasets with minimal memory overhead.
#'
#' @details
#' ## Key Features:
#' * **Memory-efficient**: All data lives in C++ backend; R objects are ~1 KB
#' * **Multiple data modalities**: Count matrices (M), paired data (CITE-seq),
#'   binary indicators (X), or pre-processed expression (Y)
#' * **Latent variable model**: Dimensionality reduction with batch effect correction
#' * **High performance**: OpenMP parallelization with optimized C++ backend
#' * **Sparse matrix support**: Efficiently handles sparse single-cell data
#'
#' ## Main Function:
#' The primary interface is \code{\link{CreateGEDIObject}}, which creates a GEDI model
#' from expression data.
#'
#' ## Workflow:
#' \enumerate{
#'   \item Create model: \code{model <- CreateGEDIObject(Samples, M, K)}
#'   \item Train model: \code{model$train(iterations = 50)}
#'   \item Access results: \code{Z <- model$Z}, \code{params <- model$params}
#' }
#'
#' @section Architecture:
#' GEDI v2 implements a three-layer architecture:
#' \itemize{
#'   \item \strong{C++ Core}: Stateful GEDI class with full optimization
#'   \item \strong{R6 Wrapper}: Thin R6 class exposing methods and active bindings
#'   \item \strong{Factory Function}: \code{CreateGEDIObject()} for user-friendly creation
#' }
#'
#' @section Computational Requirements:
#' \itemize{
#'   \item \strong{R}: >= 4.0.0
#'   \item \strong{C++ Compiler}: C++14 or later (default in R >= 4.0)
#'   \item \strong{Eigen}: >= 3.3.0 (linear algebra library)
#'   \item \strong{OpenMP}: Optional, for parallelization
#' }
#'
#' @author
#' Computational and Statistical Genomics Laboratory, McGill University
#'
#' @references
#' Add your publication reference here when available.
#'
#' @seealso
#' \itemize{
#'   \item \code{\link{CreateGEDIObject}}: Create a GEDI model
#' }
#'
#' @examples
#' \donttest{
#' # Load example data
#' pbmc_small <- SeuratObject::pbmc_small
#'
#' # Create GEDI model
#' model <- CreateGEDIObject(
#'     Samples = pbmc_small@meta.data$orig.ident,
#'     M = pbmc_small@assays$RNA@counts,
#'     K = 10,
#'     verbose = 1
#' )
#'
#' # Train model
#' model$train(iterations = 50, track_interval = 5)
#'
#' # Access latent representation
#' Z <- model$Z
#'
#' # View model summary
#' print(model)
#' }
#'
#' @useDynLib gedi2, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @import R6
#' @import Matrix
#' @importFrom methods as is
#' @importFrom stats coef median rnorm runif var
#' @importFrom utils object.size setTxtProgressBar txtProgressBar
#' @keywords internal
"_PACKAGE"

# ==============================================================================
# GEDI Dynamics Analysis
# Vector fields, activity gradients, and trajectory analysis
# ==============================================================================

#' Compute Differential Q in Z-space (JxK)
#'
#' @description
#' Computes sample-variable effects on Qi, returning a J x K matrix.
#' This is the R wrapper for the C++ function getDiffQ_cpp().
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param contrast Contrast vector (length L)
#'
#' @return Matrix J x K showing differential effect in Z-space
#'
#' @keywords internal
#' @noRd
get_diffQ_Z_space <- function(self, private, contrast) {
  
  # Validate contrast
  if (!is.numeric(contrast) || length(contrast) != self$aux$L) {
    stop("contrast must be a numeric vector of length L (", self$aux$L, ")", call. = FALSE)
  }
  
  # Call C++ function (implements the old getDiffQ.gedi logic)
  R <- getDiffQ_cpp(
    Rk_list = self$params$Rk,
    H_rotation = private$.aux_static$H.rotation,
    contrast = contrast,
    verbose = private$.verbose
  )
  
  # Set dimension names
  rownames(R) <- private$.geneIDs
  colnames(R) <- paste0("LV", 1:self$aux$K)
  
  return(R)
}


#' Compute Vector Field SVD (Internal Helper)
#'
#' @description
#' Internal helper to calculate SVD for cells and their vector fields.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param start.cond Contrast vector for start condition (length L)
#' @param end.cond Contrast vector for end condition (length L)
#' @param scale_cond_vector Scaling factor for vector field magnitude (default: 1)
#'
#' @return List with SVD results and indices
#'
#' @keywords internal
#' @noRd
run_vector_field_svd <- function(self, private, start.cond, end.cond, scale_cond_vector = 1) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if H prior was provided
  if (self$aux$L == 0) {
    stop("Cannot compute vector field: no sample-level prior (H) was provided during setup.",
         call. = FALSE)
  }
  
  # Validate contrasts
  if (!is.numeric(start.cond) || length(start.cond) != self$aux$L) {
    stop("start.cond must be a numeric vector of length L (", self$aux$L, ")", call. = FALSE)
  }
  if (!is.numeric(end.cond) || length(end.cond) != self$aux$L) {
    stop("end.cond must be a numeric vector of length L (", self$aux$L, ")", call. = FALSE)
  }
  
  # Get Z matrix
  Z <- self$params$Z
  K <- self$aux$K
  
  # Calculate Z-space differential expression (now calls C++)
  startQ_Z <- get_diffQ_Z_space(self, private, start.cond)
  endQ_Z <- get_diffQ_Z_space(self, private, 
                                     start.cond + (end.cond - start.cond) * scale_cond_vector)
  
  # Calculate coordinate vectors for start and end conditions
  startQ <- startQ_Z + Z
  endQ <- endQ_Z + Z
  
  # Map vectors onto Z coordinate system
  ZtZ <- crossprod(Z)
  rotStartQ <- solve(ZtZ, crossprod(Z, startQ))
  rotEndQ <- solve(ZtZ, crossprod(Z, endQ))
  
  # Get DB projection (uses cache)
  DB <- self$projections$DB
  
  # Concatenate start and end positions: [rotStartQ %*% DB, rotEndQ %*% DB]
  projDB <- cbind(rotStartQ %*% DB, rotEndQ %*% DB)
  
  # Perform factorized SVD (now calls C++)
  result <- run_factorized_svd_cpp(Z, projDB, private$.verbose)
  
  # Set dimension names
  names(result$d) <- paste0("LV", 1:K)
  colnames(result$u) <- paste0("LV", 1:K)
  colnames(result$v) <- paste0("LV", 1:K)
  rownames(result$u) <- private$.geneIDs
  rownames(result$v) <- c(
    paste0(private$.cellIDs, ".start"),
    paste0(private$.cellIDs, ".end")
  )
  
  # Return structured result
  structure(
    list(
      d = result$d,
      u = result$u,
      v = result$v,
      indices = list(
        embedding = 1:self$aux$N,
        vector_field = 1:(2 * self$aux$N)
      ),
      metadata = list(
        method = "vector_field",
        start_condition = start.cond,
        end_condition = end.cond,
        scale_factor = scale_cond_vector
      )
    ),
    class = "gedi_dynamics_svd"
  )
}


#' Compute Activity Gradient SVD (Internal Helper)
#'
#' @description
#' Internal helper to calculate SVD showing pathway activity gradient direction.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param C_index Index of pathway in C matrix
#' @param scale_gradient Scaling factor for gradient magnitude (NA = auto-compute)
#'
#' @return List with SVD results and indices
#'
#' @keywords internal
#' @noRd
run_gradient_svd <- function(self, private, C_index, scale_gradient = NA) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if C prior was provided
  if (self$aux$P == 0) {
    stop("Cannot compute activity gradient: no gene-level prior (C) was provided during setup.",
         call. = FALSE)
  }
  
  # Validate C_index
  if (!is.numeric(C_index) || length(C_index) != 1) {
    stop("C_index must be a single integer", call. = FALSE)
  }
  
  num_pathways <- ncol(private$.aux_static$inputC)
  if (C_index < 1 || C_index > num_pathways) {
    stop(sprintf("C_index must be between 1 and %d", num_pathways), call. = FALSE)
  }
  
  # Get matrices
  Z <- self$params$Z
  K <- self$aux$K
  
  # Compute A in original pathway space and extract gradient
  A_full <- private$.aux_static$C.rotation %*% self$params$A
  gradient <- A_full[C_index, ]
  
  # Get DB projection (uses cache)
  DB <- self$projections$DB
  
  # Auto-compute scale if needed
  if (is.na(scale_gradient)) {
    scale_gradient <- sqrt(var(as.vector(DB)) / mean(gradient^2)) * 0.2
  }
  
  # Concatenate: [DB, DB + gradient]
  projDB <- cbind(DB, DB + gradient * scale_gradient)
  
  # Perform factorized SVD (now calls C++)
  result <- run_factorized_svd_cpp(Z, projDB, private$.verbose)
  
  # Set dimension names
  pathway_name <- colnames(private$.aux_static$inputC)[C_index]
  names(result$d) <- paste0("LV", 1:K)
  colnames(result$u) <- paste0("LV", 1:K)
  colnames(result$v) <- paste0("LV", 1:K)
  rownames(result$u) <- private$.geneIDs
  rownames(result$v) <- c(
    paste0(private$.cellIDs, ".start"),
    paste0(private$.cellIDs, ".gradEnd")
  )
  
  # Return structured result
  structure(
    list(
      d = result$d,
      u = result$u,
      v = result$v,
      indices = list(
        embedding = 1:self$aux$N,
        gradient = 1:(2 * self$aux$N)
      ),
      metadata = list(
        method = "activity_gradient",
        pathway_index = C_index,
        pathway_name = pathway_name,
        scale_factor = scale_gradient
      )
    ),
    class = "gedi_dynamics_svd"
  )
}


#' Compute Joint SVD (Internal Helper)
#'
#' @description
#' Internal helper to create embedding for cells with both condition-based
#' vector field and pathway activity gradient.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#' @param start.cond Contrast vector for start condition
#' @param end.cond Contrast vector for end condition
#' @param C_index Index of pathway in C matrix
#' @param scale_cond_vector Scaling for condition vector field (default: 1)
#' @param scale_gradient Scaling for gradient (NA = auto-compute)
#'
#' @return List with SVD results and indices
#'
#' @keywords internal
#' @noRd
run_joint_svd <- function(self, private, start.cond, end.cond, C_index,
                               scale_cond_vector = 1, scale_gradient = NA) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if both H and C priors were provided
  if (self$aux$L == 0) {
    stop("Cannot compute joint analysis: no sample-level prior (H) was provided during setup.",
         call. = FALSE)
  }
  if (self$aux$P == 0) {
    stop("Cannot compute joint analysis: no gene-level prior (C) was provided during setup.",
         call. = FALSE)
  }
  
  # Validate contrasts
  if (!is.numeric(start.cond) || length(start.cond) != self$aux$L) {
    stop("start.cond must be a numeric vector of length L (", self$aux$L, ")", call. = FALSE)
  }
  if (!is.numeric(end.cond) || length(end.cond) != self$aux$L) {
    stop("end.cond must be a numeric vector of length L (", self$aux$L, ")", call. = FALSE)
  }
  
  # Validate C_index
  num_pathways <- ncol(private$.aux_static$inputC)
  if (C_index < 1 || C_index > num_pathways) {
    stop(sprintf("C_index must be between 1 and %d", num_pathways), call. = FALSE)
  }
  
  # Get matrices
  Z <- self$params$Z
  K <- self$aux$K
  
  # Compute A and extract gradient
  A_full <- private$.aux_static$C.rotation %*% self$params$A
  gradient <- A_full[C_index, ]
  
  # Calculate Z-space differential expression (now calls C++)
  startQ_Z <- get_diffQ_Z_space(self, private, start.cond)
  endQ_Z <- get_diffQ_Z_space(self, private, 
                                     start.cond + (end.cond - start.cond) * scale_cond_vector)
  
  # Calculate coordinate vectors for start and end conditions
  startQ <- startQ_Z + Z
  endQ <- endQ_Z + Z
  
  # Map vectors onto Z coordinate system
  ZtZ <- crossprod(Z)
  rotStartQ <- solve(ZtZ, crossprod(Z, startQ))
  rotEndQ <- solve(ZtZ, crossprod(Z, endQ))
  
  # Get DB projection (uses cache)
  DB <- self$projections$DB
  
  # Auto-compute gradient scale if needed
  if (is.na(scale_gradient)) {
    scale_gradient <- sqrt(var(as.vector(DB)) / mean(gradient^2)) * 0.2
  }
  
  # Concatenate: [rotStartQ*DB, rotEndQ*DB, rotStartQ*(DB+gradient)]
  projDB <- cbind(
    rotStartQ %*% DB,
    rotEndQ %*% DB,
    rotStartQ %*% (DB + gradient * scale_gradient)
  )
  
  # Perform factorized SVD (now calls C++)
  result <- run_factorized_svd_cpp(Z, projDB, private$.verbose)
  
  # Set dimension names
  pathway_name <- colnames(private$.aux_static$inputC)[C_index]
  names(result$d) <- paste0("LV", 1:K)
  colnames(result$u) <- paste0("LV", 1:K)
  colnames(result$v) <- paste0("LV", 1:K)
  rownames(result$u) <- private$.geneIDs
  rownames(result$v) <- c(
    paste0(private$.cellIDs, ".start"),
    paste0(private$.cellIDs, ".condEnd"),
    paste0(private$.cellIDs, ".gradEnd")
  )
  
  # Return structured result
  structure(
    list(
      d = result$d,
      u = result$u,
      v = result$v,
      indices = list(
        embedding = 1:self$aux$N,
        vector_field = 1:(2 * self$aux$N),
        gradient = c(1:self$aux$N, (2 * self$aux$N + 1):(3 * self$aux$N))
      ),
      metadata = list(
        method = "joint_vector_field_gradient",
        start_condition = start.cond,
        end_condition = end.cond,
        pathway_index = C_index,
        pathway_name = pathway_name,
        scale_cond_vector = scale_cond_vector,
        scale_gradient = scale_gradient
      )
    ),
    class = "gedi_dynamics_svd"
  )
}


#' Compute Pathway Gradients in Gene Space (Internal Helper)
#'
#' @description
#' Returns the gradient of all pathways projected into gene expression space.
#' This logic remains in R as it is just two fast matrix multiplications.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#'
#' @return Matrix (J x num_pathways) of gene expression gradients
#'
#' @keywords internal
#' @noRd
get_pathway_gradients <- function(self, private) {
  
  # Validate model state
  if (is.null(private$.lastResult)) {
    stop("No results yet. Run $train() first.", call. = FALSE)
  }
  
  # Check if C prior was provided
  if (self$aux$P == 0) {
    stop("Cannot compute pathway gradients: no gene-level prior (C) was provided during setup.",
         call. = FALSE)
  }
  
  # Compute A in original pathway space
  A_full <- private$.aux_static$C.rotation %*% self$params$A
  
  # Gradient is transpose of A
  gradient <- t(A_full)  # K x num_pathways
  
  # Project to gene space
  Z <- self$params$Z
  gradient <- Z %*% gradient  # J x num_pathways
  
  # Set dimension names
  rownames(gradient) <- private$.geneIDs
  colnames(gradient) <- colnames(private$.aux_static$inputC)
  
  return(gradient)
}


#' Create Dynamics Accessor Object
#'
#' @description
#' Creates accessor object for dynamics analysis with vector field and gradient methods.
#'
#' @param self Reference to GEDI R6 object
#' @param private Reference to private environment
#'
#' @return S3 object of class "gedi_dynamics"
#'
#' @keywords internal
#'@noRd
create_dynamics_accessor <- function(self, private) {
  
  structure(
    list(
      vector_field = function(start.cond, end.cond, scale = 1) {
        run_vector_field_svd(self, private, start.cond, end.cond, scale)
      },
      
      activity_gradient = function(C_index, scale = NA) {
        run_gradient_svd(self, private, C_index, scale)
      },
      
      joint = function(start.cond, end.cond, C_index, scale_cond = 1, scale_grad = NA) {
        run_joint_svd(self, private, start.cond, end.cond, C_index, scale_cond, scale_grad)
      },
      
      pathway_gradients = function() {
        get_pathway_gradients(self, private)
      }
    ),
    class = "gedi_dynamics"
  )
}


#' Print Method for Dynamics Accessor
#'
#' @param x Object of class gedi_dynamics
#' @param ... Additional arguments (ignored)
#'
#' @return Invisibly returns \code{x}.
#' @method print gedi_dynamics
#' @keywords internal
#' @export
print.gedi_dynamics <- function(x, ...) {
  cat("<GEDI Dynamics Analysis>\n")
  cat("\nAvailable methods:\n\n")
  
  cat("Vector Field Analysis:\n")
  cat("  $vector_field(start.cond, end.cond, scale = 1)\n")
  cat("    Compute SVD showing cell state transitions between conditions\n\n")
  
  cat("Pathway Activity:\n")
  cat("  $activity_gradient(C_index, scale = NA)\n")
  cat("    Compute SVD showing pathway activity gradient direction\n\n")
  
  cat("  $pathway_gradients()\n")
  cat("    Get all pathway gradients in gene expression space\n\n")
  
  cat("Joint Analysis:\n")
  cat("  $joint(start.cond, end.cond, C_index, scale_cond = 1, scale_grad = NA)\n")
  cat("    Combine vector field and gradient in single embedding\n\n")
  
  cat("Example usage:\n")
  cat("  vf <- model$dynamics$vector_field(c(0,0), c(1,0))\n")
  cat("  ag <- model$dynamics$activity_gradient(C_index = 5)\n")
  cat("  grads <- model$dynamics$pathway_gradients()\n")
  
  invisible(x)
}


#' Print Method for Dynamics SVD Results
#'
#' @param x Object of class gedi_dynamics_svd
#' @param ... Additional arguments (ignored)
#'
#' @return Invisibly returns \code{x}.
#' @method print gedi_dynamics_svd
#' @keywords internal
#' @export
print.gedi_dynamics_svd <- function(x, ...) {
  cat("<GEDI Dynamics SVD Result>\n\n")
  cat(sprintf("Method: %s\n", x$metadata$method))
  cat(sprintf("Dimensions: %d genes x %d components\n", nrow(x$u), ncol(x$u)))
  cat(sprintf("Cells: %d\n", length(x$indices$embedding)))
  
  if (x$metadata$method == "vector_field") {
    cat(sprintf("\nVector field indices: %d:%d\n", 
                min(x$indices$vector_field), max(x$indices$vector_field)))
  } else if (x$metadata$method == "activity_gradient") {
    cat(sprintf("\nPathway: %s\n", x$metadata$pathway_name))
    cat(sprintf("Gradient scale: %.3f\n", x$metadata$scale_factor))
  } else if (x$metadata$method == "joint_vector_field_gradient") {
    cat(sprintf("\nPathway: %s\n", x$metadata$pathway_name))
    cat(sprintf("Vector field indices: %d:%d\n",
                min(x$indices$vector_field), max(x$indices$vector_field)))
    cat(sprintf("Gradient indices: %s\n", 
                paste(range(x$indices$gradient), collapse = ":")))
  }
  
  cat("\nSingular values:\n")
  print(head(x$d, 5))
  
  cat("\nComponents: $d, $u, $v, $indices, $metadata\n")
  
  invisible(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_wrappers.R
\name{.get_color_vector}
\alias{.get_color_vector}
\title{Get Color Vector from Model}
\usage{
.get_color_vector(model, color_by, projection = "zdb")
}
\arguments{
\item{model}{GEDI model object}

\item{color_by}{Character: "sample", metadata column name, or gene name/index}

\item{projection}{Character: "zdb" or "db" for gene expression}
}
\value{
Vector of length N with color values
}
\description{
Get Color Vector from Model
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi-package.R
\docType{package}
\name{gedi2-package}
\alias{gedi2}
\alias{gedi2-package}
\title{gedi: Gene Expression Data Integration}
\description{
A memory-efficient implementation for integrating gene expression data from
single-cell RNA sequencing experiments. GEDI v2 uses a high-performance C++
backend with thin R wrappers to enable analysis of large-scale single-cell
datasets with minimal memory overhead.
}
\details{
\subsection{Key Features:}{
\itemize{
\item \strong{Memory-efficient}: All data lives in C++ backend; R objects are ~1 KB
\item \strong{Multiple data modalities}: Count matrices (M), paired data (CITE-seq),
binary indicators (X), or pre-processed expression (Y)
\item \strong{Latent variable model}: Dimensionality reduction with batch effect correction
\item \strong{High performance}: OpenMP parallelization with optimized C++ backend
\item \strong{Sparse matrix support}: Efficiently handles sparse single-cell data
}
}

\subsection{Main Function:}{

The primary interface is \code{\link{CreateGEDIObject}}, which creates a GEDI model
from expression data.
}

\subsection{Workflow:}{

\enumerate{
\item Create model: \code{model <- CreateGEDIObject(Samples, M, K)}
\item Train model: \code{model$train(iterations = 50)}
\item Access results: \code{Z <- model$Z}, \code{params <- model$params}
}
}
}
\section{Architecture}{

GEDI v2 implements a three-layer architecture:
\itemize{
\item \strong{C++ Core}: Stateful GEDI class with full optimization
\item \strong{R6 Wrapper}: Thin R6 class exposing methods and active bindings
\item \strong{Factory Function}: \code{CreateGEDIObject()} for user-friendly creation
}
}

\section{Computational Requirements}{

\itemize{
\item \strong{R}: >= 4.0.0
\item \strong{C++ Compiler}: C++14 or later (default in R >= 4.0)
\item \strong{Eigen}: >= 3.3.0 (linear algebra library)
\item \strong{OpenMP}: Optional, for parallelization
}
}

\examples{
\donttest{
# Load example data
pbmc_small <- SeuratObject::pbmc_small

# Create GEDI model
model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M = pbmc_small@assays$RNA@counts,
    K = 10,
    verbose = 1
)

# Train model
model$train(iterations = 50, track_interval = 5)

# Access latent representation
Z <- model$Z

# View model summary
print(model)
}

}
\references{
Add your publication reference here when available.
}
\seealso{
\itemize{
\item \code{\link{CreateGEDIObject}}: Create a GEDI model
}
}
\author{
Computational and Statistical Genomics Laboratory, McGill University
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{.plot_embedding_base}
\alias{.plot_embedding_base}
\title{Plot 2D Embedding (Base Function)}
\usage{
.plot_embedding_base(
  embedding,
  color = NULL,
  color_limits = NULL,
  palette = c("blue", "lightgrey", "red"),
  randomize = TRUE,
  point_size = 0.3,
  alpha = 0.9,
  raster = FALSE,
  xlab = "Dim 1",
  ylab = "Dim 2",
  title = NULL,
  legend_title = NULL
)
}
\arguments{
\item{embedding}{Matrix (N x 2) with x and y coordinates for each cell}

\item{color}{Vector of length N for coloring points, or NULL for uniform color}

\item{color_limits}{Numeric vector c(low, high) for color scale limits,
or NULL to auto-compute from data (uses 99th percentile for symmetric limits)}

\item{palette}{Character vector of colors for continuous scale (length 3 for diverging)}

\item{randomize}{Logical, whether to randomize point order before plotting}

\item{point_size}{Numeric, size of points}

\item{alpha}{Numeric, transparency of points (0-1)}

\item{raster}{Logical, use rasterization for large datasets (>100k points)}

\item{xlab}{Character, x-axis label}

\item{ylab}{Character, y-axis label}

\item{title}{Character, plot title}

\item{legend_title}{Character, legend title (auto-detected if NULL)}
}
\value{
ggplot2 object
}
\description{
Base function for plotting 2D embeddings with customizable coloring.
Supports both continuous and discrete color variables.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h5_readers.R
\name{read_h5ad}
\alias{read_h5ad}
\title{Read H5AD file and convert to sparse matrix}
\usage{
read_h5ad(
  file_path,
  layer = NULL,
  use_raw = FALSE,
  transpose = TRUE,
  return_metadata = FALSE,
  feature_format = "gene_name",
  verbose = FALSE
)
}
\arguments{
\item{file_path}{Character. Path to the H5AD file.}

\item{layer}{Character. The layer to extract from the H5AD file.
Default is NULL, which reads from X (the main expression matrix).
Common alternatives include "counts", "data", "scaled", etc.}

\item{use_raw}{Logical. If TRUE, reads from the raw.X slot instead of X.
Default is FALSE.}

\item{transpose}{Logical. If TRUE, transposes the matrix so genes are rows
and cells are columns (gedi format). Default is TRUE.}

\item{return_metadata}{Logical. If TRUE, returns a list with the expression
matrix, cell metadata (obs), and gene metadata (var). If FALSE, returns
only the expression matrix. Default is FALSE.}

\item{feature_format}{Character. Which feature identifier to use for gene names
when return_metadata = FALSE. Options: "gene_name" (default, uses var rownames)
or "gene_ids" (uses var$gene_ids column). Default is "gene_name".}

\item{verbose}{Logical. If TRUE, prints detailed progress messages. Default is FALSE.}
}
\value{
If return_metadata = FALSE, returns a sparse matrix (dgCMatrix) with
genes as rows and cells as columns. If return_metadata = TRUE, returns a
list with components:
\itemize{
\item X: sparse expression matrix (genes x cells)
\item obs: data.frame of cell metadata
\item var: data.frame of gene metadata
}
}
\description{
Reads an H5AD file (AnnData format) and extracts the expression matrix
as a sparse dgCMatrix suitable for use with gedi R6 object.
}
\details{
This function reads H5AD files, which are the standard format for AnnData
objects in Python. Compatible with gedi R6 class for seamless integration.
}
\examples{
\donttest{
# Round-trip: write a tiny H5AD via write_h5ad(), then read it back.
if (requireNamespace("hdf5r", quietly = TRUE) &&
    requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)

  tmp <- tempfile(fileext = ".h5ad")
  write_h5ad(model, tmp)
  expr_matrix <- read_h5ad(tmp)
  unlink(tmp)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{plot_feature_ratio}
\alias{plot_feature_ratio}
\title{Plot Two-Feature Comparison}
\usage{
plot_feature_ratio(
  model,
  gene1,
  gene2,
  comparison = "difference",
  embedding = "umap",
  projection = "zdb",
  color_limits = NULL,
  randomize = TRUE,
  point_size = 0.3,
  alpha = 0.9,
  title = NULL
)
}
\arguments{
\item{model}{GEDI model object}

\item{gene1}{Character or integer, first gene name or index}

\item{gene2}{Character or integer, second gene name or index}

\item{comparison}{Character, type of comparison ("difference" or "correlation")}

\item{embedding}{Character specifying embedding type ("umap", "pca") or
a custom N x 2 matrix}

\item{projection}{Character, type of projection ("zdb" or "db")}

\item{color_limits}{Numeric vector c(low, high) or NULL for auto-compute}

\item{randomize}{Logical, whether to randomize point order}

\item{point_size}{Numeric, size of points}

\item{alpha}{Numeric, transparency of points}

\item{title}{Character, plot title}
}
\value{
ggplot2 object
}
\description{
Compares two features by computing their difference or correlation in
the projected space. Mathematically grounded for GEDI's log-space representation.
}
\details{
For comparison = "difference":
Computes \verb{(Z[gene1,] - Z[gene2,]) * D * B}, equivalent to \verb{ZDB[gene1,] - ZDB[gene2,]}.
In log-space, this represents log(gene1/gene2) in the original count space.
Positive values indicate gene1 > gene2, negative indicates gene2 > gene1.
}
\examples{
\donttest{
if (requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)
  gene1 <- rownames(pbmc_small)[1]
  gene2 <- rownames(pbmc_small)[2]
  plot_feature_ratio(model, gene1, gene2, comparison = "difference",
                     embedding = "pca")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{plot_vector_field}
\alias{plot_vector_field}
\title{Plot Vector Field from Dynamics Analysis}
\usage{
plot_vector_field(
  dynamics_svd,
  color = NULL,
  alpha = 1,
  n_bins = 50,
  min_per_bin = 10,
  randomize = TRUE,
  arrow_size = 0.5,
  arrow_length = 0.15,
  arrow_color = "black",
  dims = c(1, 2),
  xlab = NULL,
  ylab = NULL,
  title = NULL
)
}
\arguments{
\item{dynamics_svd}{Result from model$dynamics$vector_field() or similar}

\item{color}{Vector of length N for coloring arrows, or NULL}

\item{alpha}{Vector of length N or scalar for arrow transparency}

\item{n_bins}{Integer, number of bins per dimension for aggregation}

\item{min_per_bin}{Integer, minimum observations required per bin}

\item{randomize}{Logical, whether to randomize data order}

\item{arrow_size}{Numeric, size of arrow lines}

\item{arrow_length}{Numeric, length of arrow heads (in cm)}

\item{arrow_color}{Character, color for arrows (if color is NULL)}

\item{dims}{Integer vector of length 2, which dimensions to plot}

\item{xlab}{Character, x-axis label}

\item{ylab}{Character, y-axis label}

\item{title}{Character, plot title}
}
\value{
ggplot2 object
}
\description{
Visualizes vector fields showing cell state transitions. Uses binned
aggregation for cleaner visualization without overplotting.
}
\examples{
\donttest{
# Build a tiny multi-sample fixture with a sample-level prior H so that
# model$dynamics$vector_field() is available.
set.seed(1)
n_genes <- 80; n_cells <- 60; n_samples <- 3
M <- Matrix::Matrix(
  matrix(stats::rpois(n_genes * n_cells, 5), n_genes, n_cells),
  sparse = TRUE
)
rownames(M) <- paste0("G", seq_len(n_genes))
colnames(M) <- paste0("C", seq_len(n_cells))
samples <- factor(rep(paste0("S", seq_len(n_samples)),
                      each = n_cells / n_samples))
H <- matrix(c(1, 0, 0,  0, 1, 0), nrow = 2, byrow = TRUE)
colnames(H) <- paste0("S", seq_len(n_samples))
rownames(H) <- c("cond_a", "cond_b")

model <- CreateGEDIObject(Samples = samples, M = M, K = 3, H = H,
                          verbose = 0)
model$train(iterations = 5)
vf <- model$dynamics$vector_field(start.cond = c(1, 0),
                                  end.cond   = c(0, 1))
plot_vector_field(vf)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_utils.R
\name{compute_color_limits}
\alias{compute_color_limits}
\title{Compute Color Limits from Data}
\usage{
compute_color_limits(values, symmetric = TRUE, quantile = 0.99)
}
\value{
A numeric vector of length 2 with lower and upper limits.
}
\description{
Compute Color Limits from Data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_wrappers.R
\name{plot_embedding}
\alias{plot_embedding}
\title{Plot Embedding with Improved API}
\usage{
plot_embedding(
  model,
  embedding = NULL,
  color_by = NULL,
  color = NULL,
  projection = "zdb",
  color_limits = NULL,
  palette = c("blue", "lightgrey", "red"),
  randomize = TRUE,
  point_size = 0.3,
  alpha = 0.9,
  raster = FALSE,
  xlab = "Dim 1",
  ylab = "Dim 2",
  title = NULL,
  legend_title = NULL,
  verbose = TRUE
)
}
\arguments{
\item{model}{GEDI model object (or embedding matrix for backwards compatibility)}

\item{embedding}{Character ("umap", "pca") or Nx2 matrix}

\item{color_by}{Character: "sample", metadata column, gene name, or NULL}

\item{color}{Vector for manual coloring (overrides color_by)}

\item{projection}{Character: "zdb" or "db" for gene expression projection}

\item{color_limits}{Numeric vector c(low, high) or NULL for auto}

\item{palette}{Character vector of colors for continuous scale}

\item{randomize}{Logical, randomize point order}

\item{point_size}{Numeric, size of points}

\item{alpha}{Numeric, transparency (0-1)}

\item{raster}{Logical, use rasterization for large datasets}

\item{xlab}{Character, x-axis label}

\item{ylab}{Character, y-axis label}

\item{title}{Character, plot title}

\item{legend_title}{Character, legend title}

\item{verbose}{Logical, print computation messages}
}
\value{
ggplot2 object
}
\description{
Simplified interface for plotting embeddings with automatic caching.
Model is the first argument, and color_by handles metadata/genes automatically.
}
\examples{
\donttest{
if (requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)
  plot_embedding(model, embedding = "pca", color_by = "sample")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{check_optional_dependencies}
\alias{check_optional_dependencies}
\title{Check GEDI Optional Dependencies}
\usage{
check_optional_dependencies()
}
\value{
Named logical vector indicating which optional packages are installed
}
\description{
Check which optional dependencies are installed and display their status.
}
\examples{
check_optional_dependencies()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{plot_features}
\alias{plot_features}
\title{Plot Multiple Features on Embedding}
\usage{
plot_features(
  model,
  features,
  embedding = "umap",
  projection = "zdb",
  color_limits = "global",
  ncol = NULL,
  randomize = TRUE,
  point_size = 0.2,
  alpha = 0.9,
  title = NULL
)
}
\arguments{
\item{model}{GEDI model object}

\item{features}{Character vector of gene names or integer indices}

\item{embedding}{Character specifying embedding type ("umap", "pca") or
a custom N x 2 matrix}

\item{projection}{Character, type of projection to compute ("zdb" or "db")}

\item{color_limits}{Character ("global" for shared scale, "individual" for
per-facet scale) or numeric vector c(low, high)}

\item{ncol}{Integer, number of columns in facet layout}

\item{randomize}{Logical, whether to randomize point order}

\item{point_size}{Numeric, size of points}

\item{alpha}{Numeric, transparency of points}

\item{title}{Character, plot title}
}
\value{
ggplot2 object with faceted features
}
\description{
Efficiently plots multiple gene features on a 2D embedding using faceting.
Computes projections on-the-fly without storing full ZDB matrix.
}
\examples{
\donttest{
if (requireNamespace("SeuratObject", quietly = TRUE) &&
    requireNamespace("uwot", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)
  plot_features(model, c(1, 2), embedding = "pca")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_pathway.R
\name{print.gedi_pathway_associations}
\alias{print.gedi_pathway_associations}
\title{Print Method for Pathway Associations Accessor}
\usage{
\method{print}{gedi_pathway_associations}(x, ...)
}
\arguments{
\item{x}{Object of class gedi_pathway_associations}

\item{...}{Additional arguments (ignored)}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print Method for Pathway Associations Accessor
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_class.R
\name{CreateGEDIObject}
\alias{CreateGEDIObject}
\title{Create GEDI Object}
\usage{
CreateGEDIObject(
  Samples,
  M = NULL,
  Y = NULL,
  X = NULL,
  colData = NULL,
  C = NULL,
  H = NULL,
  K = 40,
  mode = "Bsphere",
  adjustD = TRUE,
  orthoZ = TRUE,
  Z_shrinkage = 1,
  A_shrinkage = 1,
  Qi_shrinkage = 1,
  Rk_shrinkage = 1,
  oi_shrinkage = 1,
  o_shrinkage = 1,
  si_shrinkage = 1,
  fixed_si = NA,
  rsvd_p = 10,
  rsvd_sdist = "normal",
  verbose = 1,
  num_threads = 1
)
}
\arguments{
\item{Samples}{Factor or character vector indicating sample of origin for each cell}

\item{M}{Raw count matrix (sparse or dense), or list of two matrices for paired data.
C++ will compute Yi = log(M+1) internally - no Yi copy stored in R!}

\item{Y}{Log-transformed expression matrix (optional if M provided)}

\item{X}{Binary indicator matrix (optional if M or Y provided)}

\item{colData}{Optional data.frame with cell metadata}

\item{C}{Gene-level prior matrix (genes x pathways)}

\item{H}{Sample-level covariate matrix (covariates x samples)}

\item{K}{Number of latent factors (default: 10)}

\item{mode}{Normalization mode: "Bl2" or "Bsphere" (default: "Bl2")}

\item{adjustD}{Whether to adjust D based on B row norms (default: TRUE)}

\item{orthoZ}{Whether Z columns should be orthogonal (default: TRUE)}

\item{Z_shrinkage, A_shrinkage, Qi_shrinkage, Rk_shrinkage, oi_shrinkage, o_shrinkage, si_shrinkage}{Regularization strengths (default: 1)}

\item{fixed_si}{Fix cell library sizes at this value, or NA to optimize (default: NA)}

\item{rsvd_p}{Oversampling parameter for randomized SVD (default: 10)}

\item{rsvd_sdist}{Random distribution for rSVD: "normal", "unif", or "rademacher"
(default: "normal")}

\item{verbose}{Verbosity level: 0 (silent), 1 (info), 2 (debug) (default: 1)}

\item{num_threads}{Number of OpenMP threads (default: 0 = auto)}
}
\value{
GEDI R6 object with memory-efficient architecture
}
\description{
Creates and configures a GEDI (Gene Expression Data Integration) model object.
This implementation uses Option 2 memory optimization: C++ computes Yi from M,
eliminating duplicate Yi storage in R.
}
\examples{
\donttest{
# Load example data
pbmc_small <- SeuratObject::pbmc_small

# Basic usage - memory efficient!
model <- CreateGEDIObject(
  Samples = pbmc_small@meta.data$orig.ident,
  M = pbmc_small@assays$RNA@counts, # Only M stored in R; Yi computed in C++
  K = 10,
  num_threads = 1
)

# Train the model
model$train(iterations = 50)

# Access results via active bindings
Z <- model$Z
params <- model$params

# Access projections (computed on demand, cached)
zdb <- model$projections$ZDB
db <- model$projections$DB

# Access imputed expression
Y_imputed <- model$imputed$Y() # No M needed!
Y_var <- model$imputed$variance(pbmc_small@assays$RNA@counts) # M required
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h5_readers.R
\name{read_h5}
\alias{read_h5}
\title{Read 10X Genomics H5 file}
\usage{
read_h5(
  file_path,
  feature_format = "gene_name",
  unique.features = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{file_path}{Character. Path to the 10X H5 file.}

\item{feature_format}{Character. Which feature identifier to use for gene names.
Options: "gene_name" (default, uses feature names) or "gene_ids" (uses feature IDs).}

\item{unique.features}{Logical. Make feature names unique. Default TRUE.}

\item{verbose}{Logical. Print progress messages. Default FALSE.}
}
\value{
Sparse matrix (dgCMatrix) with genes as rows and cells as columns.
}
\description{
Reads a 10X Genomics HDF5 file (CellRanger v2/v3 format) and converts
it to a sparse dgCMatrix suitable for use with gedi R6 object.
}
\examples{
\donttest{
# read_h5() expects a 10x Genomics filtered_feature_bc_matrix.h5 file.
# The example body runs only when such a file is on disk and hdf5r is
# installed; otherwise it is silently skipped.
h5_file <- "filtered_feature_bc_matrix.h5"
if (file.exists(h5_file) &&
    requireNamespace("hdf5r", quietly = TRUE)) {
  expr_matrix <- read_h5(h5_file)
  expr_matrix_ids <- read_h5(h5_file, feature_format = "gene_ids")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_utils.R
\name{theme_gedi}
\alias{theme_gedi}
\title{GEDI Plot Theme}
\usage{
theme_gedi(base_size = 11)
}
\value{
A \code{ggplot2} theme object.
}
\description{
GEDI Plot Theme
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h5_readers.R
\name{list_h5_structure}
\alias{list_h5_structure}
\title{List structure of H5 or H5AD file}
\usage{
list_h5_structure(file_path, recursive = TRUE)
}
\arguments{
\item{file_path}{Character. Path to the H5 or H5AD file.}

\item{recursive}{Logical. List all nested groups. Default TRUE.}
}
\value{
data.frame with file structure information
}
\description{
Helper function to explore H5/H5AD file structure.
}
\examples{
\donttest{
# Round-trip: write a tiny H5AD via write_h5ad() then list its structure.
if (requireNamespace("hdf5r", quietly = TRUE) &&
    requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)
  tmp <- tempfile(fileext = ".h5ad")
  write_h5ad(model, tmp)
  list_h5_structure(tmp)
  unlink(tmp)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seurat_integration.R
\name{gedi_to_seurat}
\alias{gedi_to_seurat}
\title{Convert GEDI Model to Seurat Object}
\usage{
gedi_to_seurat(
  model,
  M = NULL,
  project = "GEDI",
  assay = "RNA",
  use_imputed = TRUE,
  add_projections = TRUE,
  add_embeddings = TRUE,
  min_cells = 0,
  min_features = 0,
  verbose = TRUE
)
}
\arguments{
\item{model}{GEDI model object (trained)}

\item{M}{Original count matrix (optional). If not provided, will use
back-transformed imputed values as approximate counts.}

\item{project}{Character, project name for Seurat object (default: "GEDI")}

\item{assay}{Character, name for the main assay (default: "RNA")}

\item{use_imputed}{Logical, whether to add imputed data as separate assay
(default: TRUE)}

\item{add_projections}{Logical, whether to add ZDB and DB projections as
separate assays (default: TRUE)}

\item{add_embeddings}{Logical, whether to add UMAP and PCA embeddings if
available (default: TRUE)}

\item{min_cells}{Integer, filter genes with counts in < min_cells (default: 0)}

\item{min_features}{Integer, filter cells with < min_features genes (default: 0)}

\item{verbose}{Logical, whether to print progress messages (default: TRUE)}
}
\value{
Seurat object with:
\itemize{
\item RNA assay: Original or back-transformed counts
\item imputed assay: GEDI imputed expression (if use_imputed = TRUE)
\item ZDB/DB/ADB assays: GEDI projections (if add_projections = TRUE)
\itemize{
\item ZDB: Batch-corrected gene expression (genes x cells)
\item DB: Cell embeddings in latent factor space (K x cells)
\item ADB: Pathway activities (pathways x cells) - only if C matrix provided
}
\item umap/pca reductions: Embeddings (if add_embeddings = TRUE and cached)
\item meta.data: Sample labels and colData from GEDI model
}
}
\description{
Creates a Seurat object from a trained GEDI model, including imputed data,
projections, and embeddings.
}
\examples{
\donttest{
# Load example data
pbmc <- SeuratObject::pbmc_small

# Train GEDI model
gedi_model <- seurat_to_gedi(pbmc, K = 15)
gedi_model$train(iterations = 10)

# Convert back to Seurat
seurat_obj <- gedi_to_seurat(
  gedi_model,
  use_imputed = TRUE,
  add_projections = TRUE,
  add_embeddings = TRUE
)

# Now can use Seurat functions
library(Seurat)
DefaultAssay(seurat_obj) <- "imputed"
seurat_obj <- FindVariableFeatures(seurat_obj)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{plot_convergence}
\alias{plot_convergence}
\title{Plot Training Convergence}
\usage{
plot_convergence(
  model,
  layout = c("faceted", "separate", "compact"),
  params = NULL,
  log_scale = TRUE,
  smooth = FALSE,
  title = "Training Convergence"
)
}
\arguments{
\item{model}{GEDI model object}

\item{layout}{Character, layout style:
\itemize{
\item "faceted": Single plot with facet_wrap (default, best for reports)
\item "separate": List of individual plots (for interactive exploration)
\item "compact": Two-panel plot (global vs sample-specific)
}}

\item{params}{Character vector, which parameters to include. Options:
"Z", "A", "o", "Bi", "Qi", "oi", "si", "Rk", "Ro", "sigma2".
NULL means all available.}

\item{log_scale}{Logical, use log10 scale for y-axis}

\item{smooth}{Logical, add smooth trend line}

\item{title}{Character, plot title}
}
\value{
ggplot2 object (for "faceted" or "compact") or list of ggplot2 objects (for "separate")
}
\description{
Visualizes convergence of model parameters during training.
Supports multiple layout styles for different use cases.
}
\examples{
\donttest{
if (requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 10, track_interval = 2)
  plot_convergence(model)
  plots <- plot_convergence(model, layout = "separate")
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/h5_writers.R
\name{write_h5ad}
\alias{write_h5ad}
\title{Write GEDI model to H5AD file}
\usage{
write_h5ad(
  model,
  file_path,
  X_slot = c("imputed", "projection", "original"),
  M = NULL,
  include_embeddings = TRUE,
  include_raw = FALSE,
  compression = 6,
  verbose = TRUE
)
}
\arguments{
\item{model}{GEDI R6 object (must be trained)}

\item{file_path}{Character. Path where the H5AD file should be written.}

\item{X_slot}{Character. Which expression data to save in the main X slot:
\itemize{
\item "imputed": Imputed expression (default, requires M matrix)
\item "projection": ZDB projection (always available)
\item "original": Original M matrix (requires M parameter)
}}

\item{M}{Optional. Original count matrix to save when X_slot="original" or
include_raw=TRUE. Must match the dimensions used during model training.}

\item{include_embeddings}{Logical. Include PCA/UMAP in obsm if cached. Default TRUE.}

\item{include_raw}{Logical. If TRUE, saves original M in raw.X (requires M parameter).
Default FALSE.}

\item{compression}{Integer. Gzip compression level (0-9). Default 6.}

\item{verbose}{Logical. Print progress messages. Default TRUE.}
}
\value{
Invisibly returns the file path
}
\description{
Exports a trained GEDI model to H5AD (AnnData) format for interoperability
with Python tools like scanpy. The file contains expression data, embeddings,
metadata, and GEDI-specific parameters.
}
\details{
The H5AD file structure contains:
\itemize{
\item X: Main expression matrix (based on X_slot parameter)
\item obs: Cell metadata (sample IDs, colData)
\item var: Gene metadata (gene IDs)
\item obsm: Cell embeddings (X_gedi, X_pca, X_umap)
\item varm: Gene loadings (gedi_Z, gedi_Q_mean)
\item uns: Model parameters and metadata
\item raw.X: Original counts (if include_raw=TRUE)
}

The function handles the technical details of HDF5/AnnData compatibility:
\itemize{
\item Writes sparse matrices in CSR format
\item Transposes dense matrices for Python's row-major layout
\item Creates scalar string attributes (not arrays) for AnnData compatibility
\item Validates M matrix identity using fingerprints
}
}
\examples{
\donttest{
if (requireNamespace("hdf5r", quietly = TRUE) &&
    requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = pbmc_small@assays$RNA@counts,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)

  # Write to a temporary file (CRAN policy: never write to the user's
  # working directory in examples).
  tmp <- tempfile(fileext = ".h5ad")
  write_h5ad(model, tmp)
  write_h5ad(model, tmp, X_slot = "projection")
  unlink(tmp)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_wrappers.R
\name{.get_embedding}
\alias{.get_embedding}
\title{Get Embedding Coordinates with Smart Caching}
\usage{
.get_embedding(model, embedding_type, dims = c(1, 2), verbose = TRUE)
}
\arguments{
\item{model}{GEDI model object}

\item{embedding_type}{Character: "umap", "pca", or a custom Nx2 matrix}

\item{dims}{Integer vector of length 2 for which dimensions to use (default c(1,2))}

\item{verbose}{Logical, print messages about computation}
}
\value{
Nx2 matrix of embedding coordinates
}
\description{
Get Embedding Coordinates with Smart Caching
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seurat_integration.R
\name{seurat_to_gedi}
\alias{seurat_to_gedi}
\title{Convert Seurat Object to GEDI Model}
\usage{
seurat_to_gedi(
  seurat_object,
  assay = "RNA",
  slot = "counts",
  sample_column = "orig.ident",
  subset_samples = NULL,
  K = 10,
  mode = "Bl2",
  C = NULL,
  H = NULL,
  validate_counts = TRUE,
  use_variable_features = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{seurat_object}{Seurat object}

\item{assay}{Character, which assay to use (default: "RNA")}

\item{slot}{Character, which slot/layer to extract (default: "counts").
For Seurat v5 with split layers (e.g., counts.CTRL, counts.STIM),
this will automatically detect and combine all matching layers.}

\item{sample_column}{Character, column name in meta.data for sample labels
(default: "orig.ident")}

\item{subset_samples}{Character vector, subset to specific samples (default: NULL = all)}

\item{K}{Integer, number of latent factors (default: 10)}

\item{mode}{Character, normalization mode: "Bl2" or "Bsphere" (default: "Bl2")}

\item{C}{Gene-level prior matrix (genes x pathways) (default: NULL)}

\item{H}{Sample-level covariate matrix (covariates x samples) (default: NULL)}

\item{validate_counts}{Logical, whether to validate data appears to be counts
(default: TRUE)}

\item{use_variable_features}{Logical, whether to subset to highly variable features
(default: TRUE). If TRUE, uses genes from VariableFeatures(seurat_object).}

\item{verbose}{Logical, whether to print progress messages (default: TRUE)}

\item{...}{Additional arguments passed to CreateGEDIObject()}
}
\value{
GEDI R6 object
}
\description{
Extracts count data from a Seurat object and creates a GEDI model.
Automatically validates that the data contains raw counts (not normalized).
Handles both Seurat v4 and v5, including split layers in v5.
}
\examples{
\donttest{
library(Seurat)

# Load example data
pbmc <- SeuratObject::pbmc_small

# Basic usage
gedi_model <- seurat_to_gedi(
  seurat_object = pbmc,
  sample_column = "orig.ident",
  K = 15
)

# Train the model
gedi_model$train(iterations = 10)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz.R
\name{install_optional_dependencies}
\alias{install_optional_dependencies}
\title{List Optional GEDI Dependencies}
\usage{
install_optional_dependencies(which = "all", verbose = TRUE)
}
\arguments{
\item{which}{Character vector specifying which dependency groups to query.
Options: "h5" (hdf5r), "umap" (uwot), "validation" (digest), or "all" (default).}

\item{verbose}{Logical, whether to print messages (default: TRUE)}
}
\value{
A named logical vector indicating which packages are installed (invisibly).
}
\description{
Reports which optional packages are needed and provides install commands.
Does \strong{not} install anything automatically.
}
\examples{
# Show which optional packages are missing
install_optional_dependencies()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_utils.R
\name{scale_color_gedi_discrete}
\alias{scale_color_gedi_discrete}
\title{Discrete Color Palette for GEDI}
\usage{
scale_color_gedi_discrete(name = "Group", ...)
}
\value{
A \code{ggplot2} discrete color scale.
}
\description{
Discrete Color Palette for GEDI
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plots.R
\name{plot_dispersion}
\alias{plot_dispersion}
\title{Plot Dispersion Analysis}
\usage{
plot_dispersion(
  dispersion_df,
  show_identity = TRUE,
  point_size = 0.1,
  alpha = 0.5,
  title = "Dispersion Analysis"
)
}
\arguments{
\item{dispersion_df}{Data frame from compute_dispersion() with columns:
Expected_Var, Observed_Var, Sample, n, bin}

\item{show_identity}{Logical, whether to show y=x identity line}

\item{point_size}{Numeric, size of points}

\item{alpha}{Numeric, transparency of points}

\item{title}{Character, plot title}
}
\value{
ggplot2 object
}
\description{
Visualizes the relationship between expected and observed variance
for count data. Useful for assessing model fit quality.
}
\examples{
\donttest{
if (requireNamespace("SeuratObject", quietly = TRUE)) {
  pbmc_small <- SeuratObject::pbmc_small
  M <- pbmc_small@assays$RNA@counts
  model <- CreateGEDIObject(
    Samples = pbmc_small@meta.data$orig.ident,
    M       = M,
    K       = 3,
    verbose = 0
  )
  model$train(iterations = 5)
  disp <- model$imputed$dispersion(M)
  plot_dispersion(disp)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gedi_plot_utils.R
\name{scale_color_gedi_diverging}
\alias{scale_color_gedi_diverging}
\title{Diverging Color Scale for GEDI Plots}
\usage{
scale_color_gedi_diverging(limits = NULL, name = "Value", ...)
}
\value{
A \code{ggplot2} continuous color scale.
}
\description{
Diverging Color Scale for GEDI Plots
}
\keyword{internal}
